// my-testbed.cpp : Defines the entry point for the console application.
//

#include <stdio.h>
#include <memory.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "ptf_file.h"
#include "ptf_str.h"
#include "ptf_elevation.h"

// macros to check return code from PTF API functions
#define PTF_FAILED(a) ((a) != peSuccess)
#define PTF_SUCCEEDED(a) ((a) == peSuccess)

// output a text message that corresponds to each PTF API return code
void outputPTFErr( ptf_err err )
{
  switch ( err ) {
    case peFileNotFound:
      fprintf( stderr, "The indicated file was not found.\n" );
      break;
    case peInvalidFile:
      fprintf( stderr, "The indicated path is not a PTF file.\n" );
      break;
    case peIoError:
      fprintf( stderr, "Some error when attempting to read/write the file.\n" );
      break;
    case peInvalidVersion:
      fprintf( stderr, "The indicated path is of the wrong version.\n" );
      break;
    case pePermissionDenied:
      fprintf( stderr, "The indicated path is not accessible, or attempt to write to read-only file.\n" );
      break;
    case peInvalidArguments:
      fprintf( stderr, "The arguments to the function are wrong.\n" );
      break;
    case peTooBigForMachine:
      fprintf( stderr, "The data in question is too big for the current machine (i e, needs 64 bit memory).\n" );
      break;
    case peReadOutOfBounds:
      fprintf( stderr, "Attempt to read past the end of a record.\n" );
      break;
    case peInvalidSequence:
      fprintf( stderr, "The attempted function is not currently valid.\n" );
      break;
    case peReadOnly:
      fprintf( stderr, "The file is not open for writing.\n" );
      break;
    case peException:
      fprintf( stderr, "C++ exception occurred while executing -- probably out of memory.\n" );
      break;
    case peOutOfMemory:
      fprintf( stderr, "One of those rare occasions where out of memory is not an exception.\n" );
      break;
    case peTypeNotFound:
      fprintf( stderr, "Request for a type that doesn't exist.\n" );
      break;
    case peRecordNotFound:
      fprintf( stderr, "Record is out of range, or has been deleted.\n" );
      break;
    default:
      fprintf( stderr, "Operation completed successfully.\n" );
  }
}

static ptf_dem_handle makeTestDem( char const * pName, ptf_16 cols, ptf_16 rows, ptf_8 type, 
                                  float halfWidth, float halfHeight, float xscale, float yscale, float zscale, 
                                  ptf_dem_val nullVal )
{
  // allocate a dem record
  ptf_err ptf_status;
  ptf_dem_handle testDem = NULL;

  fprintf( stdout, "Creating in-memory DEM record '%s'...\n", pName );
  if ( PTF_SUCCEEDED( ptf_status = ptf_dem_create( &testDem, (ptf_char const *) pName ) ) ) {
    // setup meta-data for a small sample record
    fprintf( stdout, "Setting '%s' dimensions (rows, cols, and sample type)...\n", pName );
    if ( PTF_SUCCEEDED( ptf_status = ptf_dem_set_col_row_type( testDem, cols, rows, type ) ) ) {
      fprintf( stdout, "Setting up '%s' header...\n", pName );
      // extents
      if ( PTF_FAILED( ptf_status = ptf_dem_set_extents( testDem, halfWidth, halfHeight ) ) ) {
        outputPTFErr( ptf_status );
        fprintf( stdout, "*** Cannot set extents for '%s'\n", pName );
      }
      
      // null data value
      if ( PTF_FAILED( ptf_status = ptf_dem_set_null_data( testDem, nullVal ) ) ) {
        outputPTFErr( ptf_status );
        fprintf( stdout, "*** Cannot set null data value for '%s'\n", pName );
      }

      // x/y/z scale
      if ( PTF_FAILED( ptf_status = ptf_dem_set_scale( testDem, xscale, yscale, zscale ) ) ) {
        outputPTFErr( ptf_status );
        fprintf( stdout, "*** Cannot set scale for '%s'\n", pName );
      }

      // populate the record
      fprintf( stdout, "Setting post values for '%s'...\n", pName );
      ptf_dem_val post;
      for ( int r = 0; r < rows; r++ ) {
        for ( int c = 0; c < cols; c++ ) {
          switch ( type ) {
            case pvtInt8: post.int8 = c; break;
            case pvtInt16: post.int16 = c; break;
            case pvtFloat32: post.float32 = c; break;
          }

          if ( r != c ) {
            if ( PTF_FAILED( ptf_status = ptf_dem_set_sample_val( testDem, c, r, post ) ) ) {
              outputPTFErr( ptf_status );
              fprintf( stdout, "*** Cannot set post for '%s' - col %d, row %d\n", pName );
            }
          }
          else {
            if ( PTF_FAILED( ptf_status = ptf_dem_set_sample_val( testDem, c, r, nullVal ) ) ) {
              outputPTFErr( ptf_status );
              fprintf( stdout, "*** Cannot set post to NULL for '%s' - col %d, row %d\n", pName, c, r );
            }
          }
        }
      }
    }
    else {
      outputPTFErr( ptf_status );
      fprintf( stdout, "*** Cannot set row/col/sample size for '%s'\n", pName );
      return NULL;
    }
  }
  else {
    outputPTFErr( ptf_status );
    fprintf( stdout, "Cannot create DEM record for '%s'\n", pName );
    return NULL;
  }

  return testDem;
}

static void verifyTestDem( ptf_dem_handle testDem, int index, char const * pName, ptf_16 cols, ptf_16 rows, 
                          ptf_8 type, float halfWidth, float halfHeight, float xscale, float yscale, float zscale, 
                          ptf_dem_val nullVal )
{
  // allocate a dem record
  ptf_err ptf_status;
  fprintf( stdout, "Verifying '%s' header...\n", pName );
  // name
  ptf_name actName;
  if ( PTF_FAILED( ptf_status = ptf_dem_get_name( testDem, actName ) ) ) {
    outputPTFErr( ptf_status );
    fprintf( stdout, "*** Error calling ptf_dem_get_name() for '%s'\n", pName );
  }
  else if ( 0 == ptf_strcasecmp( actName, (ptf_char const *) "XXXXXXXXXXXXXXXXXXXX" ) ) {
    fprintf( stdout, "*** Error: ptf_dem_read_from_file() did not modify dem name!!! \n" );
  }
  else if ( 0 != ptf_strcasecmp( actName, (ptf_char const *) pName ) ) {
    fprintf( stdout, "*** Error: Wrong name: Expected '%s', got '%s'\n", pName, actName );
  }

  // rows and columns
  ptf_16 actCols = 0;
  ptf_16 actRows = 0;
  if ( PTF_FAILED( ptf_status = ptf_dem_get_col_row( testDem, &actCols, &actRows ) ) ) {
    outputPTFErr( ptf_status );
    fprintf( stdout, "*** Error calling ptf_dem_get_col_row() for '%s'\n", pName );
  }
  else if ( !( actCols == cols && actRows == rows ) ) {
    fprintf( stdout, "*** Error: expected %d columns and %d rows; got %d columns and %d rows\n", cols, rows, actCols, actRows );
  }

  // sample type
  ptf_8 actType = pvtInvalid;
  if ( PTF_FAILED( ptf_status = ptf_dem_get_sample_type( testDem, &actType ) ) ) {
    outputPTFErr( ptf_status );
    fprintf( stdout, "*** Error calling ptf_dem_get_sample_type() for '%s'\n", pName );
  }
  else if ( type != actType ) {
    fprintf( stdout, "*** Error: expected sample type %d, got %d\n", type, actType );
  }

  // extents
  float actWidth = 0.f;
  float actHeight = 0.f;
  if ( PTF_FAILED( ptf_status = ptf_dem_get_extents( testDem, &actWidth, &actHeight ) ) ) {
    outputPTFErr( ptf_status );
    fprintf( stdout, "*** Error calling ptf_dem_get_extents() for '%s'\n", pName );
  }
  else if ( !( halfWidth == actWidth && halfHeight == actHeight ) ) {
    fprintf( stdout, "*** Error: expected width = %f, height = %f; got width = %f, height = %f\n", halfWidth, halfHeight, actWidth, actHeight );
  }
  
  // null data value
  ptf_dem_val actNull;
  bool result = false;
  if ( PTF_FAILED( ptf_status = ptf_dem_get_null_data( testDem, &result, &actNull ) ) ) {
    outputPTFErr( ptf_status );
    fprintf( stdout, "*** Error calling ptf_dem_get_null_data() for '%s'\n", pName );
  }
  else if ( !result ) {
    fprintf( stdout, "*** Error: This file does not have an established Null Value\n" );
  }
  else if ( 0 != memcmp( &nullVal, &actNull, sizeof( ptf_dem_val ) ) ) {
    ptf_32 expected, actual;
    memcpy( &expected, &nullVal, sizeof( ptf_dem_val ) );
    memcpy( &actual, &actNull, sizeof( ptf_dem_val ) );
    fprintf( stdout, "*** Error: expected NULL value = %x, got %x\n", expected, actual );
  }

  // x/y/z scale
  float actX = 0.f;
  float actY = 0.f;
  float actZ = 0.f;
  if ( PTF_FAILED( ptf_status = ptf_dem_get_scale( testDem, &actX, &actY, &actZ ) ) ) {
    outputPTFErr( ptf_status );
    fprintf( stdout, "*** Error calling ptf_dem_get_scale() for '%s'\n", pName );
  }
  else if ( !( actX == xscale && actY == yscale && actZ == zscale ) ) {
    fprintf( stdout, "*** Error: expected scales: X = %f, Y = %f, Z = %f; got X = %f; Y = %f, Z = %f\n", xscale, yscale, zscale, actX, actY, actZ );
  }

  // populate the record
  fprintf( stdout, "Verifying post values for '%s'...\n", pName );
  ptf_dem_val post;
  for ( int r = 0; r < rows; r++ ) {
    for ( int c = 0; c < cols; c++ ) {
      if ( PTF_FAILED( ptf_status = ptf_dem_get_sample_val( testDem, c, r, &post ) ) ) {
        outputPTFErr( ptf_status );
        fprintf( stdout, "*** [ %d, %d ] Failed in ptf_dem_get_sample_val()\n", c, r );
      }
      else if ( r == c ) {
        if ( 0 != memcmp( &nullVal, &post, sizeof( ptf_dem_val ) ) ) {
          ptf_32 expected, actual;
          memcpy( &expected, &nullVal, sizeof( ptf_dem_val ) );
          memcpy( &actual, &post, sizeof( ptf_dem_val ) );
          fprintf( stdout, "*** [ %d, %d ] Expected null post = %x; got %x\n", c, r, expected, actual );
        }
      }
      else if ( post.int8 != c ) {
        ptf_32 actual;
        memcpy( &actual, &post, sizeof( ptf_dem_val ) );
        fprintf( stdout, "*** [ %d, %d ] Expected post %x; got %x\n", c, r, c, actual );
      }
    }
  }

  fprintf( stdout, "Verify complete for '%s'...\n", pName );
}

int main(int argc, char * argv[])
{
  int status = 0;
  ptf_err ptf_status;

  // setup the filename for use with PTF API
  char const *srcname = "ptf_elev_test_file.ptf";

  // "Tiny DEM" parameters
  const char *pTinyName = "Tiny DEM";
  ptf_16 tinyCols = 12;
  ptf_16 tinyRows = 10;
  ptf_8 tinyType = pvtInt8;
  float tinyWidth = 120.123f;
  float tinyHeight = 100.567f;
  ptf_dem_val tinyNull;
  memset( &tinyNull, 0, sizeof( tinyNull ) );
  tinyNull.int8 = (ptf_8) -1;
  float tinyScaleX = 1.1f;
  float tinyScaleY = 1.2f;
  float tinyScaleZ = 10.0f;

  // "Level 1 DEM" parameters
  const char *pLev1Name = "Level 1 DEM";
  ptf_16 lev1Cols = 121;
  ptf_16 lev1Rows = 101;
  ptf_8 lev1Type = pvtInt16;
  float lev1Width = 1200.123f;
  float lev1Height = 1000.567f;
  ptf_dem_val lev1Null;
  lev1Null.int8 = -1;
  memset( &lev1Null, 0, sizeof( lev1Null ) );
  float lev1ScaleX = 1.9f;
  float lev1ScaleY = 1.8f;
  float lev1ScaleZ = 11.1f;

  // *** test write-appending a DEM record ***

  // first off, make a new file
  ptf_file_handle outFile;
  fprintf( stdout, "Opening '%s' for write (%s)...\n", srcname, pTinyName );
  if ( PTF_SUCCEEDED( ptf_status = ptf_file_open( (ptf_char const *) srcname, ptf_open_clobber, &outFile ) ) ) {
    
    // create a small test DEM
    ptf_dem_handle testDem;
    if ( testDem = makeTestDem( pTinyName, tinyCols, tinyRows, tinyType, tinyWidth, tinyHeight, 
      tinyScaleX, tinyScaleY, tinyScaleZ, tinyNull ) ) {

      // write it out
      fprintf( stdout, "Writing 'Tiny DEM' to disk...\n" );
      if ( PTF_FAILED( ptf_status = ptf_dem_append_to_file( testDem, outFile, NULL ) ) ) {
        outputPTFErr( ptf_status );
        fprintf( stdout, "*** Cannot write-append '%s' to output file\n", pTinyName );
      }

      // delete the small test DEM
      if ( PTF_FAILED( ptf_status = ptf_dem_destroy( testDem ) ) ) {
        outputPTFErr( ptf_status );
        fprintf( stdout, "*** Failure when destroying '%s'\n", pTinyName );
      }
    }

    // clost the PTF file
    if ( PTF_FAILED( ptf_status = ptf_file_close( outFile ) ) ) {
      outputPTFErr( ptf_status );
      fprintf( stdout, "*** Error closing file '%s'\n", srcname );
    }
  }
  else {
    outputPTFErr( ptf_status );
    fprintf( stdout, "*** Cannot open file %s for UPDATE\n", srcname );
    status = 1;
  }

  // *** test reading a DEM record ***

  // reopen the file - read-only this time
  fprintf( stdout, "\nOpening '%s' for read (%s)...\n", srcname, pTinyName );
  if ( PTF_SUCCEEDED( ptf_status = ptf_file_open( (ptf_char const *) srcname, ptf_open_read, &outFile ) ) ) {
    
    // create an empty DEM record
    ptf_dem_handle testDem;
    fprintf( stdout, "Creating in-memory DEM record...\n" );
    if ( PTF_SUCCEEDED( ptf_status = ptf_dem_create( &testDem, (ptf_char const *) "XXXXXXXXXXXXXXXXXXXX" ) ) ) {

      // read back the small test DEM into the empty DEM record
      fprintf( stdout, "Read '%s' from disk...\n", pTinyName );
      ptf_fixed_ref ref;
      ref.index = 0;
      if ( PTF_SUCCEEDED( ptf_status = ptf_dem_read_from_file( testDem, outFile, ref ) ) ) {

        // verify that we got back what we wrote
        fprintf( stdout, "Validate '%s'...\n", pTinyName );
        verifyTestDem( testDem, 0, pTinyName, tinyCols, tinyRows, tinyType, tinyWidth, tinyHeight, 
          tinyScaleX, tinyScaleY, tinyScaleZ, tinyNull );

        // verify again, using all bogus values, thus testing the verify function (and to some extent, the API)
        ptf_dem_val bogusNull;
        bogusNull.int16 = 666;
        fprintf( stdout, "\nRead and validate against bogus values (%s)...\n", pTinyName );
        verifyTestDem( testDem, 0, "Bogus DEM", 13, 11, pvtInt16, 12.234f, 45.678f, 
          1.0f, 1.0f, 1.0f, bogusNull );
      }
      else {
        outputPTFErr( ptf_status );
        fprintf( stdout, "Cannot read DEM record for '%s'\n", pTinyName );
      }

      // delete the DEM record
      if ( PTF_FAILED( ptf_status = ptf_dem_destroy( testDem ) ) ) {
        outputPTFErr( ptf_status );
        fprintf( stdout, "*** Failure when destroying '%s'\n", pTinyName );
      }
    }
    else {
      outputPTFErr( ptf_status );
      fprintf( stdout, "Cannot create DEM record ('%s' read/verify)\n", pTinyName );
    }

    // close the PTF file
    if ( PTF_FAILED( ptf_status = ptf_file_close( outFile ) ) ) {
      outputPTFErr( ptf_status );
      fprintf( stdout, "*** Error closing file '%s'\n", srcname );
    }
  }
  else {
    outputPTFErr( ptf_status );
    fprintf( stdout, "*** Cannot open file %s for READ_ONLY\n", srcname );
    status = 1;
  }

  // *** test write-updating a DEM record

  // open the file for update
  fprintf( stdout, "\nOpening '%s' for write (%s)...\n", srcname, pTinyName );
  if ( PTF_SUCCEEDED( ptf_status = ptf_file_open( (ptf_char const *) srcname, ptf_open_update, &outFile ) ) ) {

    // Make a new test DEM with different parameters
    ptf_dem_handle testDem;
    if ( testDem = makeTestDem( pLev1Name, lev1Cols, lev1Rows, lev1Type, lev1Width, lev1Height, 
      lev1ScaleX, lev1ScaleY, lev1ScaleZ, lev1Null ) ) {
  
      // write the new DEM out, replacing the DEM from the write-append test
      fprintf( stdout, "Writing '%s' to disk, replacing '%s'\n", pLev1Name, pTinyName );

      ptf_fixed_ref ref;
      ref.index = 0;
      if ( PTF_FAILED( ptf_status = ptf_dem_update_to_file( testDem, outFile, ref ) ) ) {
        outputPTFErr( ptf_status );
        fprintf( stdout, "*** Cannot write-update '%s' to output file\n", pLev1Name );
      }

      // delete the dem record
      if ( PTF_FAILED( ptf_status = ptf_dem_destroy( testDem ) ) ) {
        outputPTFErr( ptf_status );
        fprintf( stdout, "*** Failure when destroying '%s'\n", pLev1Name );
      }
    }

    // read and verify the record we just updated
    if ( PTF_SUCCEEDED( ptf_status = ptf_dem_create( &testDem, (ptf_char const *) "XXXXXXXXXXXXXXXXXXXX" ) ) ) {
      fprintf( stdout, "Read '%s' from disk...\n", pTinyName );
      ptf_fixed_ref ref;
      ref.index = 0;
      if ( PTF_SUCCEEDED( ptf_status = ptf_dem_read_from_file( testDem, outFile, ref ) ) ) {

        // verify that we got back what we wrote
        fprintf( stdout, "Validate '%s'...\n", pLev1Name );
        verifyTestDem( testDem, 0, pLev1Name, lev1Cols, lev1Rows, lev1Type, lev1Width, lev1Height, 
          lev1ScaleX, lev1ScaleY, lev1ScaleZ, lev1Null );
      }

      // delete the dem record
      if ( PTF_FAILED( ptf_status = ptf_dem_destroy( testDem ) ) ) {
        outputPTFErr( ptf_status );
        fprintf( stdout, "*** Failure when destroying '%s'\n", pLev1Name );
      }
    }

    // close the PFT file one last time
    if ( PTF_FAILED( ptf_status = ptf_file_close( outFile ) ) ) {
      outputPTFErr( ptf_status );
      fprintf( stdout, "*** Error closing file '%s'\n", srcname );
    }
  }
  else {
    outputPTFErr( ptf_status );
    fprintf( stdout, "*** Cannot open file %s for UPDATE\n", srcname );
    status = 1;
  }

  return status;
}

