
#include "ptf.h"
#include "ptf_bookmark.h"
#include "ptf_file.h"
#include "ptf_str.h"
#include "ptf_date.h"
#include "ptf_types.h"
#include "ptf_octree.h"
#include "ptf_content.h"
#include "ptf_mesh.h"
#include "ptf_util.h"
#include "ptf_material.h"
#include "ptf_texture.h"
#include "ptf_lod.h"
#include "ptf_collection.h"
#include "ptf_spheroid.h"

#include <assert.h>
#include <math.h>
#include <string.h>
#include <time.h>
#include <deque>

#ifdef  NDEBUG
#define CHECK(a,o,b) \
  if( !(a o b) ) { char *__temp = 0; *__temp = 1; }
#else
#define CHECK(a,o,b) \
  assert( a o b )
#endif

#define CALL(e,f,r) \
  e = f ; assert( e == r )

#define CLOSE_ENOUGH(a,b) \
  fabs(a-b) < 10e-6

#if defined(_WIN32) || defined(_WIN64)

#include <windows.h>

#define MAIN __stdcall WinMain( \
    HINSTANCE hInstance, \
    HINSTANCE hPrevInstance, \
    LPSTR lpCmdLine, \
    int nCmdShow \
)

#define unlink _wunlink
#define File(x) x
#define WPATHSEP '\\'

#else

#define MAIN main(int, char**)

class File {
  public:
    File(wchar_t const *str) {
      //assert(!wcsncmp(str, L"c:\\", 3));
      wstr = new wchar_t[wcslen(str)+10];
      wchar_t *oot = wstr;
      if( !wcsncmp(str, L"c:\\", 3) ) {
        wcsncpy(wstr, L"/tmp/", 6);
        oot = wstr+5;
        str += 3;
      }
      while (*str) {
        if (*str == '\\') {
          *oot = '/';
        }
        else {
          *oot = *str;
        }
        ++str;
        ++oot;
      }
      *oot = 0;
      buf = 0;
    }
    ~File() {
      delete[] buf;
      delete[] wstr;
    }
    operator wchar_t const *() {
      return wstr;
    }
    operator char const *() {
      if (!buf) {
        size_t s = wcslen(wstr)*4+1;
        buf = new char[s];
        wcstombs(buf, wstr, s);
        buf[s-1] = 0;
      }
      return buf;
    }
    wchar_t *wstr;
    char *buf;
};
#define WPATHSEP '/'

#endif



ptf_geopos random_coord(double min, double max)
{
  static long long seed;
  if (!seed) {
    seed = 0xc34aa59b347593f7ULL;
  }
  ptf_geopos ret;
  seed = seed * 0x8acd888a8f8887ULL + 0xc34985739487adfULL;
  ret.x = ((double)((seed >> 10) & 0x7fffffffUL) + double(seed & 0x3ff) * 1.0/1024.0) * 1.0 / 0x80000000;
  ret.x = ret.x * (max-min) + min;
  seed = seed * 0x8acd888a8f8887ULL + 0xc34985739487adfULL;
  ret.y = ((double)((seed >> 10) & 0x7fffffffUL) + double(seed & 0x3ff) * 1.0/1024.0) * 1.0 / 0x80000000;
  ret.y = ret.y * (max-min) + min;
  seed = seed * 0x8acd888a8f8887ULL + 0xc34985739487adfULL;
  ret.z = ((double)((seed >> 10) & 0x7fffffffUL) + double(seed & 0x3ff) * 1.0/1024.0) * 1.0 / 0x80000000;
  ret.z = ret.z * (max-min) + min;

  return ret;
}

void test_ptf_str()
{
  char const * str1 = "str 1\n";
  char const * str3 = "stR 1\n";
  wchar_t const * str2 = L"str 2\n";
  ptf_char buf[20];
  size_t siz;
  CALL( siz, ptf_str_from_char(str1, buf, 10), 7 ); 
  assert(!strcmp((char const *)buf, str1));
  CALL( siz, ptf_str_from_wchar(str2, buf, 10), 7 );
  memset(buf, 4, 10);
  CALL( siz, ptf_str_from_wchar(str2, buf, 3), 7 );
  assert(buf[3] == 4);
  assert(buf[2] == 0);
  ptf_char path[20];
  char cstr[30];
  wchar_t cwstr[30];
  CALL( siz, ptf_str_from_char("../foo/filename.xtn", path, 20), 20 );
  CALL( siz, ptf_str_to_char(path, cstr, 30, '?'), 20 );
  assert(!strcmp(cstr, "../foo/filename.xtn"));
  CALL( siz, ptf_str_to_wchar(path, cwstr, 30, '?'), 20 );
  assert(!wcscmp(cwstr, L"../foo/filename.xtn"));
  CALL( siz, ptf_str_basename(path, buf, 10), 13 );
  CHECK( strncmp((char const *)buf, "filename.", 10), ==, 0 );
  CALL( siz, ptf_str_basename(path, buf, 20), 13 );
  CHECK( strncmp((char const *)buf, "filename.xtn", 20), ==, 0 );
  CALL( siz, ptf_str_from_char(str1, buf, 20), 7 );
  CALL( siz, ptf_str_from_char(str3, path, 20), 7 );
  CALL( siz, ptf_strcasecmp(buf, buf), 0 );
  CALL( siz, ptf_strcasecmp(buf, path), 0 );
  CALL( siz, ptf_strncpy(buf, 7, path), 0 );
  CALL( siz, ptf_strncpy(buf, 3, path), -1 );
}

void test_ptf_date()
{
  struct tm testtime;
  testtime.tm_sec = 1;
  testtime.tm_min = 2;
  testtime.tm_hour = 3;
  testtime.tm_mday = 4;
  testtime.tm_mon = 5;
  testtime.tm_year = 2006 - 1900;
  testtime.tm_wday = 5;
  testtime.tm_yday = 4;
  testtime.tm_isdst = 1;
  ptf_date testdate;
  ptf_date_from_tm(&testtime, &testdate);
  CHECK( testdate.year, ==, 2006 );
  CHECK( testdate.month, ==, 6 );
  CHECK( testdate.day, ==, 4 );
  CHECK( testdate.hour, ==, 3 );
  CHECK( testdate.minute, ==, 2 );
  CHECK( testdate.second, ==, 1 );
  CHECK( testdate.hundredth, ==, 0 );
}

void test_ptf_unpack()
{
  struct unpacktest {
    short a[4];
    char b[4];
    float c[2];
  } packed[] = {
    {
      { -32767, 32767, 0, 18 },
      { -127, -64, 0, 127 },
      { -3.125f, 5.5f },
    },
  };
  ptf_err err;
  float out[10], *ptr = out;

  void const *start = &packed[0];
  CALL( err, ptf_type_unpack(pvtInt16, 4, 4, &start, ptr), peSuccess );
  CALL( err, ptf_type_unpack(pvtInt8, 4, 4, &start, ptr+4), peSuccess );
  CALL( err, ptf_type_unpack(pvtFloat32, 2, 4, &start, ptr+8), peSuccess );
  CHECK( out[0], ==, -1.0f );
  CHECK( out[1], ==, 1.0f );
  CHECK( out[2], ==, 0.0f );
  CHECK( CLOSE_ENOUGH( out[3], 18.0f / 32767.0f ), ==, true );    //CHECK( out[3], ==, 18.0f / 32767.0f );
  CHECK( out[4], ==, -1.0f );
  CHECK( CLOSE_ENOUGH( out[5], ( -64.0f / 127.0f ) ), ==, true ); //CHECK( out[5], ==, -64.0f / 127.0f );
  CHECK( out[6], ==, 0.0f );
  CHECK( out[7], ==, 1.0f );
  CHECK( out[8], ==, -3.125 );
  CHECK( out[9], ==, 5.5f );

  start = &packed[0];
  CALL( err, ptf_type_unpack(pvtUInt16, 4, 4, &start, ptr), peSuccess );
  CALL( err, ptf_type_unpack(pvtUInt8, 4, 4, &start, ptr+4), peSuccess );
  CALL( err, ptf_type_unpack(pvtFloat32, 2, 4, &start, ptr+8), peSuccess );
  CHECK( CLOSE_ENOUGH( out[0], 32769.0f / 65535.0f ), ==, true );   //CHECK( out[0], ==, 32769.0f / 65535.0f );
  CHECK( CLOSE_ENOUGH( out[1], 32767.0f / 65535.0f ), ==, true );   //CHECK( out[1], ==, 32767.0f / 65535.0f );
  CHECK( out[2], ==, 0.0f );
  CHECK( CLOSE_ENOUGH( out[3], 18.0f / 65535 ), ==, true );   //CHECK( out[3], ==, 18.0f / 65535);
  CHECK( CLOSE_ENOUGH( out[4], 129.0f / 255.0f ), ==, true ); //CHECK( out[4], ==, 129.0f / 255.0f );
  CHECK( CLOSE_ENOUGH( out[5], 192.0f / 255.0f ), ==, true ); //CHECK( out[5], ==, 192.0f / 255.0f );
  CHECK( out[6], ==, 0.0f );
  CHECK( CLOSE_ENOUGH( out[7], 127.0f / 255.0f ), ==, true ); //CHECK( out[7], ==, 127.0f / 255.0f );
  CHECK( out[8], ==, -3.125 );
  CHECK( out[9], ==, 5.5f );

  start = &packed[0];
  CALL( err, ptf_type_unpack(pvtInt16_4, 4, 4, &start, ptr), peSuccess );
  CALL( err, ptf_type_unpack(pvtInt8_4, 4, 4, &start, ptr+4), peSuccess );
  CALL( err, ptf_type_unpack(pvtFloat32, 2, 4, &start, ptr+8), peSuccess );
  CHECK( CLOSE_ENOUGH( out[0], 8.0f * -32767.0f / 32768.0f ), ==, true ); //CHECK( out[0], ==, 8.0f * -32767.0f / 32768.0f );
  CHECK( CLOSE_ENOUGH( out[1], 8.0f * 32767.0f / 32768.0f ), ==, true );  //CHECK( out[1], ==, 8.0f * 32767.0f / 32768.0f );
  CHECK( out[2], ==, 0.0f );
  CHECK( CLOSE_ENOUGH( out[3], 8.0f * 18.0f / 32768.0f ), ==, true );     //CHECK( out[3], ==, 8.0f * 18.0f / 32768.0f );
  CHECK( CLOSE_ENOUGH( out[4], 8.0f * -127.0f / 128.0f ), ==, true );     //CHECK( out[4], ==, 8.0f * -127.0f / 128.0f );
  CHECK( CLOSE_ENOUGH( out[5], 8.0f * -64.0f / 128.0f ), ==, true );      //CHECK( out[5], ==, 8.0f * -64.0f / 128.0f );
  CHECK( out[6], ==, 0.0f );
  CHECK( CLOSE_ENOUGH( out[7], 8.0f * 127.0f / 128.0f ), ==, true );      //CHECK( out[7], ==, 8.0f * 127.0f / 128.0f );
  CHECK( out[8], ==, -3.125 );
  CHECK( out[9], ==, 5.5f );

  start = &packed[0];
  CALL( err, ptf_type_unpack(pvtInt16_8, 4, 4, &start, ptr), peSuccess );
  CALL( err, ptf_type_unpack(pvtInt8_8, 4, 4, &start, ptr+4), peSuccess );
  CALL( err, ptf_type_unpack(pvtFloat32, 2, 4, &start, ptr+8), peSuccess );
  CHECK( CLOSE_ENOUGH( out[0], 128.0f * -32767.0f / 32768.0f ), ==, true ); //CHECK( out[0], ==, 128.0f * -32767.0f / 32768.0f );
  CHECK( CLOSE_ENOUGH( out[1], 128.0f * 32767.0f / 32768.0f ), ==, true );  //CHECK( out[1], ==, 128.0f * 32767.0f / 32768.0f );
  CHECK( out[2], ==, 0.0f );
  CHECK( CLOSE_ENOUGH( out[3], 128.0f * 18.0f / 32768.0f ), ==, true );     //CHECK( out[3], ==, 128.0f * 18.0f / 32768.0f );
  CHECK( CLOSE_ENOUGH( out[4], 128.0f * -127.0f / 128.0f ), ==, true );     //CHECK( out[4], ==, 128.0f * -127.0f / 128.0f );
  CHECK( CLOSE_ENOUGH( out[5], 128.0f * -64.0f / 128.0f ), ==, true );      //CHECK( out[5], ==, 128.0f * -64.0f / 128.0f );
  CHECK( out[6], ==, 0.0f );
  CHECK( CLOSE_ENOUGH( out[7], 128.0f * 127.0f / 128.0f ), ==, true );      //CHECK( out[7], ==, 128.0f * 127.0f / 128.0f );
  CHECK( out[8], ==, -3.125 );
  CHECK( out[9], ==, 5.5f );

  start = &packed[0];
  CALL( err, ptf_type_unpack(pvtInt16_12, 4, 4, &start, ptr), peSuccess );
  CALL( err, ptf_type_unpack(pvtInt8_12, 4, 4, &start, ptr+4), peSuccess );
  CALL( err, ptf_type_unpack(pvtFloat32, 2, 4, &start, ptr+8), peSuccess );
  CHECK( CLOSE_ENOUGH( out[0], 2048.0f * -32767.0f / 32768.0f ), ==, true ); //CHECK( out[0], ==, 2048.0f * -32767.0f / 32768.0f );
  CHECK( CLOSE_ENOUGH( out[1], 2048.0f * 32767.0f / 32768.0f ), ==, true );  //CHECK( out[1], ==, 2048.0f * 32767.0f / 32768.0f );
  CHECK( out[2], ==, 0.0f );
  CHECK( CLOSE_ENOUGH( out[3], 2048.0f * 18.0f / 32768.0f ), ==, true );     //CHECK( out[3], ==, 2048.0f * 18.0f / 32768.0f );
  CHECK( CLOSE_ENOUGH( out[4], 2048.0f * -127.0f / 128.0f ), ==, true );     //CHECK( out[4], ==, 2048.0f * -127.0f / 128.0f );
  CHECK( CLOSE_ENOUGH( out[5], 2048.0f * -64.0f / 128.0f ), ==, true );      //CHECK( out[5], ==, 2048.0f * -64.0f / 128.0f );
  CHECK( out[6], ==, 0.0f );
  CHECK( CLOSE_ENOUGH( out[7], 2048.0f * 127.0f / 128.0f ), ==, true );      //CHECK( out[7], ==, 2048.0f * 127.0f / 128.0f );
  CHECK( out[8], ==, -3.125 );
  CHECK( out[9], ==, 5.5f );

  start = &packed[0];
  CALL( err, ptf_type_unpack(pvtInt16_16, 4, 4, &start, ptr), peSuccess );
  CALL( err, ptf_type_unpack(pvtInt8_16, 4, 4, &start, ptr+4), peSuccess );
  CALL( err, ptf_type_unpack(pvtFloat32, 2, 4, &start, ptr+8), peSuccess );
  CHECK( CLOSE_ENOUGH( out[0], 32768.0f * -32767.0f / 32768.0f ), ==, true ); //CHECK( out[0], ==, 32768.0f * -32767.0f / 32768.0f );
  CHECK( CLOSE_ENOUGH( out[1], 32768.0f * 32767.0f / 32768.0f ), ==, true );  //CHECK( out[1], ==, 32768.0f * 32767.0f / 32768.0f );
  CHECK( out[2], ==, 0.0f );
  CHECK( CLOSE_ENOUGH( out[3], 32768.0f * 18.0f / 32768.0f ), ==, true );     //CHECK( out[3], ==, 32768.0f * 18.0f / 32768.0f );
  CHECK( CLOSE_ENOUGH( out[4], 32768.0f * -127.0f / 128.0f ), ==, true );     //CHECK( out[4], ==, 32768.0f * -127.0f / 128.0f );
  CHECK( CLOSE_ENOUGH( out[5], 32768.0f * -64.0f / 128.0f ), ==, true );      //CHECK( out[5], ==, 32768.0f * -64.0f / 128.0f );
  CHECK( out[6], ==, 0.0f );
  CHECK( CLOSE_ENOUGH( out[7], 32768.0f * 127.0f / 128.0f ), ==, true );      //CHECK( out[7], ==, 32768.0f * 127.0f / 128.0f );
  CHECK( out[8], ==, -3.125 );
  CHECK( out[9], ==, 5.5f );

  start = &packed[0];
  CALL( err, ptf_type_unpack(pvtUInt16_4, 4, 4, &start, ptr), peSuccess );
  CALL( err, ptf_type_unpack(pvtUInt8_4, 4, 4, &start, ptr+4), peSuccess );
  CALL( err, ptf_type_unpack(pvtFloat32, 2, 4, &start, ptr+8), peSuccess );
  CHECK( CLOSE_ENOUGH( out[0], 16.0f * 32769.0f / 65536.0f ), ==, true ); //CHECK( out[0], ==, 16.0f * 32769.0f / 65536.0f );
  CHECK( CLOSE_ENOUGH( out[1], 16.0f * 32767.0f / 65536.0f ), ==, true ); //CHECK( out[1], ==, 16.0f * 32767.0f / 65536.0f );
  CHECK( out[2], ==, 0.0f );
  CHECK( CLOSE_ENOUGH( out[3], 16.0f * 18.0f / 65536.0f ), ==, true );    //CHECK( out[3], ==, 16.0f * 18.0f / 65536.0f );
  CHECK( CLOSE_ENOUGH( out[4], 16.0f * 129.0f / 256.0f ), ==, true );     //CHECK( out[4], ==, 16.0f * 129.0f / 256.0f );
  CHECK( CLOSE_ENOUGH( out[5], 16.0f * 192.0f / 256.0f ), ==, true );     //CHECK( out[5], ==, 16.0f * 192.0f / 256.0f );
  CHECK( out[6], ==, 0.0f );
  CHECK( CLOSE_ENOUGH( out[7], 16.0f * 127.0f / 256.0f ), ==, true );     //CHECK( out[7], ==, 16.0f * 127.0f / 256.0f );
  CHECK( out[8], ==, -3.125 );
  CHECK( out[9], ==, 5.5f );
 
  start = &packed[0];
  CALL( err, ptf_type_unpack(pvtUInt16_8, 4, 4, &start, ptr), peSuccess );
  CALL( err, ptf_type_unpack(pvtUInt8_8, 4, 4, &start, ptr+4), peSuccess );
  CALL( err, ptf_type_unpack(pvtFloat32, 2, 4, &start, ptr+8), peSuccess );
  CHECK( CLOSE_ENOUGH( out[0], 256.0f * 32769.0f / 65536.0f ), ==, true ); //CHECK( out[0], ==, 256.0f * 32769.0f / 65536.0f );
  CHECK( CLOSE_ENOUGH( out[1], 256.0f * 32767.0f / 65536.0f ), ==, true ); //CHECK( out[1], ==, 256.0f * 32767.0f / 65536.0f );
  CHECK( out[2], ==, 0.0f );
  CHECK( CLOSE_ENOUGH( out[3], 256.0f * 18.0f / 65536.0f ), ==, true );    //CHECK( out[3], ==, 256.0f * 18.0f / 65536.0f );
  CHECK( CLOSE_ENOUGH( out[4], 256.0f * 129.0f / 256.0f ), ==, true );     //CHECK( out[4], ==, 256.0f * 129.0f / 256.0f );
  CHECK( CLOSE_ENOUGH( out[5], 256.0f * 192.0f / 256.0f ), ==, true );     //CHECK( out[5], ==, 256.0f * 192.0f / 256.0f );
  CHECK( out[6], ==, 0.0f );
  CHECK( CLOSE_ENOUGH( out[7], 256.0f * 127.0f / 256.0f ), ==, true );     //CHECK( out[7], ==, 256.0f * 127.0f / 256.0f );
  CHECK( out[8], ==, -3.125 );
  CHECK( out[9], ==, 5.5f );

  start = &packed[0];
  CALL( err, ptf_type_unpack(pvtUInt16_12, 4, 4, &start, ptr), peSuccess );
  CALL( err, ptf_type_unpack(pvtUInt8_12, 4, 4, &start, ptr+4), peSuccess );
  CALL( err, ptf_type_unpack(pvtFloat32, 2, 4, &start, ptr+8), peSuccess );
  CHECK( CLOSE_ENOUGH( out[0], 4096.0f * 32769.0f / 65536.0f ), ==, true ); //CHECK( out[0], ==, 4096.0f * 32769.0f / 65536.0f );
  CHECK( CLOSE_ENOUGH( out[1], 4096.0f * 32767.0f / 65536.0f ), ==, true ); //CHECK( out[1], ==, 4096.0f * 32767.0f / 65536.0f );
  CHECK( out[2], ==, 0.0f );
  CHECK( CLOSE_ENOUGH( out[3], 4096.0f * 18.0f / 65536.0f ), ==, true );    //CHECK( out[3], ==, 4096.0f * 18.0f / 65536.0f );
  CHECK( CLOSE_ENOUGH( out[4], 4096.0f * 129.0f / 256.0f ), ==, true );     //CHECK( out[4], ==, 4096.0f * 129.0f / 256.0f );
  CHECK( CLOSE_ENOUGH( out[5], 4096.0f * 192.0f / 256.0f ), ==, true );     //CHECK( out[5], ==, 4096.0f * 192.0f / 256.0f );
  CHECK( out[6], ==, 0.0f );
  CHECK( CLOSE_ENOUGH( out[7], 4096.0f * 127.0f / 256.0f ), ==, true );     //CHECK( out[7], ==, 4096.0f * 127.0f / 256.0f );
  CHECK( out[8], ==, -3.125 );
  CHECK( out[9], ==, 5.5f );

  start = &packed[0];
  CALL( err, ptf_type_unpack(pvtUInt16_16, 4, 4, &start, ptr), peSuccess );
  CALL( err, ptf_type_unpack(pvtUInt8_16, 4, 4, &start, ptr+4), peSuccess );
  CALL( err, ptf_type_unpack(pvtFloat32, 2, 4, &start, ptr+8), peSuccess );
  CHECK( CLOSE_ENOUGH( out[0], 65536.0f * 32769.0f / 65536.0f ), ==, true ); //CHECK( out[0], ==, 65536.0f * 32769.0f / 65536.0f );
  CHECK( CLOSE_ENOUGH( out[1], 65536.0f * 32767.0f / 65536.0f ), ==, true ); //CHECK( out[1], ==, 65536.0f * 32767.0f / 65536.0f );
  CHECK( out[2], ==, 0.0f );
  CHECK( CLOSE_ENOUGH( out[3], 65536.0f * 18.0f / 65536.0f ), ==, true );    //CHECK( out[3], ==, 65536.0f * 18.0f / 65536.0f );
  CHECK( CLOSE_ENOUGH( out[4], 65536.0f * 129.0f / 256.0f ), ==, true );     //CHECK( out[4], ==, 65536.0f * 129.0f / 256.0f );
  CHECK( CLOSE_ENOUGH( out[5], 65536.0f * 192.0f / 256.0f ), ==, true );     //CHECK( out[5], ==, 65536.0f * 192.0f / 256.0f );
  CHECK( out[6], ==, 0.0f );
  CHECK( CLOSE_ENOUGH( out[7], 65536.0f * 127.0f / 256.0f ), ==, true );     //CHECK( out[7], ==, 65536.0f * 127.0f / 256.0f );
  CHECK( out[8], ==, -3.125 );
  CHECK( out[9], ==, 5.5f );
}

void test_ptf_file(wchar_t const * filename1, wchar_t const * filename2)
{
  ptf_char buf[100];
  ptf_err err;
  ptf_file_handle file;

  CHECK( ptf_str_from_wchar(filename1, buf, 100), >, 0 );
  CALL( err, ptf_file_open(buf, ptf_open_read, &file), peFileNotFound );
  CHECK( file, ==, 0 );
  CALL( err, ptf_file_open_accessors(buf, ptf_open_update, ptf_file_platform_accessors(), NULL, &file), peFileNotFound );
  CHECK( file, ==, 0 );
  CALL( err, ptf_file_open_accessors(buf, ptf_open_create, ptf_file_platform_accessors(), NULL, &file), peSuccess );
  CHECK( file, !=, 0 );
  CALL( err, ptf_file_close(file), peSuccess );

  CALL( err, ptf_file_open_accessors(buf, ptf_open_read, ptf_file_platform_accessors(), NULL, &file), peSuccess );
  CHECK( file, !=, 0 );
  ptf_record_ref ref;
  CALL( err, ptf_file_new_complete_record(file, ptTestType1, "", 1, &ref), peReadOnly );
  CALL( err, ptf_file_close(file), peSuccess );

  CALL( err, ptf_file_open_accessors(buf, ptf_open_update, ptf_file_platform_accessors(), NULL, &file), peSuccess );
  CHECK( file, !=, 0 );
  CALL( err, ptf_file_new_complete_record(file, ptTestType1, "12345", 6, &ref), peSuccess);
  CHECK( ref.index, ==, 1 );  //  record 0 is never returned
  ptf_32 cnt = 0;
  CALL( err, ptf_file_count_types(file, &cnt), peSuccess );
  CHECK( cnt, ==, 2 );
  ptf_type type1, type2;
  CALL( err, ptf_file_get_type_at(file, 0, &type1), peSuccess );
  CHECK( type1, ==, ptTestType1 );
  CALL( err, ptf_file_close(file), peSuccess );

  CALL( err, ptf_file_open_accessors(buf, ptf_open_read, ptf_file_platform_accessors(), NULL, &file), peSuccess );
  CHECK( file, !=, 0 );
  CALL( err, ptf_file_count_types(file, &cnt), peSuccess );
  CHECK( cnt, ==, 2 );      //  "peEmptyRecord" type is also added
  CALL( err, ptf_file_get_type_at(file, 0, &type1), peSuccess );
  if (type1 != ptTestType1) {
    CALL( err, ptf_file_get_type_at(file, 1, &type2), peSuccess );
    CHECK( type2, ==, ptTestType1 );
  }
  CALL( err, ptf_file_count_type_records(file, ptTestType1, &cnt), peSuccess );
  CHECK( cnt, ==, 2 );    //  record 0 is never returned
  CALL( err, ptf_file_count_type_records(file, ptTestType2, &cnt), peTypeNotFound );
  CALL( err, ptf_file_close(file), peSuccess );

  file = 0;
  CALL( err, ptf_file_open_accessors(buf, ptf_open_create, ptf_file_platform_accessors(), NULL, &file), peSuccess );
  CHECK( file, !=, 0 );
  CALL( err, ptf_file_count_types(file, &cnt), peSuccess );
  CHECK( cnt, ==, 2 );      //  "peEmptyRecord" type is also added
  CALL( err, ptf_file_count_type_records(file, ptTestType1, &cnt), peSuccess );
  CHECK( cnt, ==, 2 );
  ptf_record_ref rref;
  memset(&rref, 0, sizeof(rref));
  rref.index = 1;
  rref.type = ptTestType1;
  char str[10];
  //  reading too much should return an error
  CALL( err, ptf_file_read_record(file, &rref, 0, str, 10), peInvalidArguments );
  //  reading just right should return the data
  CALL( err, ptf_file_read_record(file, &rref, 0, str, 6), peSuccess );
  CHECK( strcmp(str, "12345"), ==, 0 );
  CALL( err, ptf_file_flush(file), peSuccess ); //  nothing's changed
  CALL( err, ptf_file_new_record(file, ptTestType2), peSuccess );
  char bbuf[256];
  for (int i = 0; i < 256; ++i) {
    bbuf[i] = 255-i;
  }
  CALL( err, ptf_file_write_record(file, bbuf, 127), peSuccess );
  CALL( err, ptf_file_write_record(file, &bbuf[127], 129), peSuccess );
  memset(&rref, -1, sizeof(rref));
  CALL( err, ptf_file_commit_record(file, &rref), peSuccess );
  CHECK( rref.file, ==, 0 );
  CHECK( rref.index, ==, 1 );
  CHECK( rref.type, ==, ptTestType2 );
  CALL( err, ptf_file_count_type_records(file, ptTestType2, &cnt), peSuccess );
  CHECK( cnt, ==, 2 );
  CALL( err, ptf_file_read_record(file, &rref, 128, bbuf, 64), peSuccess );
  for (int i = 0; i < 64; ++i) {
    CHECK( bbuf[i], ==, 127-i );
  }

  memset(&ref, 0, sizeof(ref));
  ref.type = ptTestType1;
  CALL( err, ptf_file_replace_record(file, &ref), peInvalidArguments );

  CHECK( file, !=, 0 );
  CALL( err, ptf_file_new_complete_record(file, ptTestType1, "23456", 6, &ref), peSuccess);
  CALL( err, ptf_file_count_types(file, &cnt), peSuccess );
  CHECK( cnt, ==, 3 );
  CALL( err, ptf_file_count_type_records(file, ptTestType2, &cnt), peSuccess );
  CHECK( cnt, ==, 2 );
  CALL( err, ptf_file_read_record(file, &rref, 128, bbuf, 64), peSuccess );
  for (int i = 0; i < 64; ++i) {
    CHECK( bbuf[i], ==, 127-i );
  }
  rref.type = ptTestType1;
  rref.index = 0;
  CALL( err, ptf_file_remove_record(file, &rref), peInvalidArguments );
  rref.index = 1;
  CALL( err, ptf_file_remove_record(file, &rref), peSuccess );
  ptf_64 size = 0;
  CALL( err, ptf_file_get_record_size(file, &rref, &size), peRecordNotFound );
  rref.index = 2;
  CALL( err, ptf_file_get_record_size(file, &rref, &size), peSuccess );
  CHECK( size, ==, 6 );
  rref.index = 3;
  CALL( err, ptf_file_get_record_size(file, &rref, &size), peRecordNotFound );
  rref.index = 0;
  CALL( err, ptf_file_replace_record(file, &rref), peInvalidArguments );
  rref.index = 1;
  CALL( err, ptf_file_replace_record(file, &rref), peSuccess );
  CALL( err, ptf_file_write_record(file, bbuf, 65), peSuccess );
  CALL( err, ptf_file_commit_record(file, &rref), peSuccess );
  CHECK( rref.type, ==, ptTestType1 );
  CHECK( rref.index, ==, 1 );
  CALL( err, ptf_file_close(file), peSuccess );

  CALL( err, ptf_file_open_accessors(buf, ptf_open_update, ptf_file_platform_accessors(), NULL, &file), peSuccess );
  ptf_file_header hdr;
  CALL( err, ptf_file_get_header(file, &hdr), peSuccess);
  CHECK( hdr.range.x_min, ==, 0 );
  CHECK( hdr.range.x_max, ==, 0 );
  CHECK( hdr.range.y_min, ==, 0 );
  CHECK( hdr.range.y_max, ==, 0 );
  CHECK( hdr.range.z_min, ==, 0 );
  CHECK( hdr.range.z_max, ==, 0 );
  hdr.range.x_min = 0x70000000;
  hdr.range.x_max = 0x90000000;
  hdr.range.y_min = 0x07000000;
  hdr.range.y_max = 0x09000000;
  hdr.range.z_min = 0x00700000;
  hdr.range.z_max = 0x00900000;
  CALL( err, ptf_file_update_header(file, &hdr), peSuccess );

  CALL( err, ptf_file_allocate_records(file, ptTestType3, 25, &rref), peSuccess );
  CHECK( rref.index, ==, 1 );
  CALL( err, ptf_file_replace_record(file, &rref), peSuccess );
  CALL( err, ptf_file_commit_record(file, 0), peSuccess );
  rref.index = 25;
  CALL( err, ptf_file_replace_record(file, &rref), peSuccess );
  CALL( err, ptf_file_commit_record(file, 0), peSuccess );
  rref.index = 26;
  CALL( err, ptf_file_replace_record(file, &rref), peRecordNotFound );

  CALL( err, ptf_file_close(file), peSuccess );
}

void test_ptf_bookmarks(wchar_t const * filename1)
{
  ptf_char buf[100];
  ptf_err err;
  ptf_file_handle file;

  CHECK( ptf_str_from_wchar(filename1, buf, 100), >, 0 );

  // bookmarks
  CALL( err, ptf_file_open(buf, ptf_open_clobber, &file), peSuccess );

  ptf_bookmark bk1;
  ptf_str_from_char("Bookmark one", bk1.name, PTF_NAMESIZE);
  bk1.pos.x = 600000.0;
  bk1.pos.y = 500000.0;
  bk1.pos.z = 700000.0;
  bk1.radius = 123.4567;
  bk1.ori.x = 0.1f;
  bk1.ori.y = 0.2f;
  bk1.ori.z = 0.3f;
  bk1.ori.w = 0.4f;

  ptf_bookmark bk2;
  ptf_str_from_char("Bookmark two", bk2.name, PTF_NAMESIZE);
  bk2.pos.x = 123456.0;
  bk2.pos.y = 246810.0;
  bk2.pos.z = 135791.1;
  bk2.radius = 987.65432;
  bk2.ori.x = 0.12345f;
  bk2.ori.y = 0.98765f;
  bk2.ori.z = 0.666f;
  bk2.ori.w = 0.999f;

  ptf_bookmark bk3;
  ptf_str_from_char("Bookmark three", bk3.name, PTF_NAMESIZE);
  bk3.pos.x = 600001.0;
  bk3.pos.y = 600002.0;
  bk3.pos.z = 600003.0;
  bk3.radius = 100.0;
  bk3.ori.x = 111.f;
  bk3.ori.y = 222.f;
  bk3.ori.z = 333.f;
  bk3.ori.w = 444.f;

  ptf_fixed_ref ref1;
  ref1.index = 0;
  CALL( err, ptf_file_create_bookmark( file, &bk1, &ref1 ), peSuccess );
  ptf_fixed_ref ref2;
  ref2.index = 0;
  CALL( err, ptf_file_create_bookmark( file, &bk2, &ref2 ), peSuccess );
  ptf_fixed_ref ref3;
  ref3.index = 0;
  CALL( err, ptf_file_create_bookmark( file, &bk3, &ref3 ), peSuccess );

  CALL( err, ptf_file_close(file), peSuccess );
  CALL( err, ptf_file_open(buf, ptf_open_update, &file), peSuccess );

  ptf_bookmark read1, read2, read3;
  memset( &read1, 0, sizeof( read1 ) );
  memset( &read2, 0, sizeof( read2 ) );
  memset( &read3, 0, sizeof( read3 ) );
  CALL( err, ptf_file_get_bookmark( file, ref1, &read1 ), peSuccess );
  int status = memcmp( &bk1, &read1, sizeof( ptBookmark ) );
  CHECK( status, ==, 0 );
  CALL( err, ptf_file_get_bookmark( file, ref2, &read2 ), peSuccess );
  status = memcmp( &bk2, &read2, sizeof( ptBookmark ) );
  CHECK( status, ==, 0 );
  CALL( err, ptf_file_get_bookmark( file, ref3, &read3 ), peSuccess );
  status = memcmp( &bk3, &read3, sizeof( ptBookmark ) );
  CHECK( status, ==, 0 );

  ptf_bookmark mod2;
  memcpy( &mod2, &bk2, sizeof( ptf_bookmark ) );
  ptf_str_from_char("Modified two", mod2.name, PTF_NAMESIZE);
  CALL( err, ptf_file_set_bookmark( file, ref2, &mod2 ), peSuccess );

  ptf_bookmark mod2Read;
  memset( &mod2Read, 0, sizeof( mod2Read ) );
  CALL( err, ptf_file_get_bookmark( file, ref2, &mod2Read ), peSuccess );
  status = memcmp( &mod2Read, &mod2, sizeof( ptBookmark ) );
  CHECK( status, ==, 0 );

  CALL( err, ptf_file_remove_bookmark( file, ref1 ), peSuccess );

  CALL( err, ptf_file_close(file), peSuccess );
  CALL( err, ptf_file_open(buf, ptf_open_update, &file), peSuccess );
  
  ptf_32 o_count = 0;
  CALL( err, ptf_file_count_type_records(file, ptBookmark, &o_count), peSuccess );
  // TRICKY: expect 4 here because count includes deleted records (which includes the one we removed plus
  // that got replaced with we did 'set_bookmark')
  CHECK( o_count, ==, 4 );

  memset( &read1, 0, sizeof( read1 ) );
  CALL( err, ptf_file_get_bookmark( file, ref1, &read1 ), peRecordNotFound );
  memset( &read2, 0, sizeof( read2 ) );
  CALL( err, ptf_file_get_bookmark( file, ref2, &read2 ), peSuccess );
  memset( &read3, 0, sizeof( read3 ) );
  CALL( err, ptf_file_get_bookmark( file, ref3, &read3 ), peSuccess );

  CALL( err, ptf_file_close(file), peSuccess );
}

void test_ptf_octree(wchar_t const * filename1)
{
  ptf_char buf[100];
  ptf_err err;
  ptf_file_handle file;
  ptf_file_header hdr;
  ptf_geopos savePos;
  ptf_32 side, level;

  // test get/set methods on ptf_cube
  ptf_ocube accTest;
  ptf_geopos accMin, accMax;
  accMin.x = 0.0; accMin.y = 0.0; accMin.z = 5999995.0;
  accMax.x = 3000.0; accMax.y = 3000.0; accMax.z = 6000005.0;
  CALL( err, ptf_octree_box_set_coords(accMin, accMax, &accTest), peSuccess );
  // check that box was set correctly
  double minBox = -PTF_MIN_GEOPOS * 16.0;
  CHECK( accTest.x_min, ==, minBox );
  CHECK( accTest.y_min, ==, minBox );
  CHECK( accTest.z_min, ==, ( minBox + ( 5999995.0 * 16.0 ) ) );
  CHECK( accTest.x_max, ==, ( minBox + ( 3000.0 * 16.0 ) ) );
  CHECK( accTest.y_max, ==, ( minBox + ( 3000.0 * 16.0 ) ) );
  double tempe = minBox + ( 6000005.0 * 16.0 );
  CHECK( accTest.z_max, ==, ( minBox + ( 6000005.0 * 16.0 ) ) );
  ptf_geopos accMinResult, accMaxResult;
  CALL( err, ptf_octree_box_get_coords(accTest, &accMinResult, &accMaxResult), peSuccess );
  // check that set/get yields what we started with
  CHECK( accMin.x, ==, accMinResult.x );
  CHECK( accMin.y, ==, accMinResult.y );
  CHECK( accMin.z, ==, accMinResult.z );
  CHECK( accMax.x, ==, accMaxResult.x );
  CHECK( accMax.y, ==, accMaxResult.y );
  CHECK( accMax.z, ==, accMaxResult.z );

  // Set up data
  CHECK( ptf_str_from_wchar(filename1, buf, 100), >, 0 );
  {
    CALL( err, ptf_file_open(buf, ptf_open_update, &file), peSuccess );
    ptf_file_header hdr;
    CALL( err, ptf_file_get_header(file, &hdr), peSuccess);
    CHECK( hdr.range.x_min, ==, 0 );
    CHECK( hdr.range.x_max, ==, 0 );
    CHECK( hdr.range.y_min, ==, 0 );
    CHECK( hdr.range.y_max, ==, 0 );
    CHECK( hdr.range.z_min, ==, 0 );
    CHECK( hdr.range.z_max, ==, 0 );
    hdr.range.x_min = 0x70000000;
    hdr.range.x_max = 0x90000000;
    hdr.range.y_min = 0x07000000;
    hdr.range.y_max = 0x09000000;
    hdr.range.z_min = 0x00700000;
    hdr.range.z_max = 0x00900000;
    CALL( err, ptf_file_update_header(file, &hdr), peSuccess );
    CALL( err, ptf_file_close(file), peSuccess );
  }

  // Start test
  CALL( err, ptf_box_side_to_level( 0x80000000, &level ), peSuccess );
  CHECK( level, == , 0 );
  CALL( err, ptf_box_side_to_level( 0x40000000, &level ), peSuccess );
  CHECK( level, == , 1 );
  CALL( err, ptf_box_side_to_level( 2, &level ), peSuccess );
  CHECK( level, ==, 30 );

  CALL( err, ptf_level_to_box_side( 0, &side ), peSuccess );
  CHECK( side, ==, 0x80000000 );
  CALL( err, ptf_level_to_box_side( 1, &side ), peSuccess );
  CHECK( side, ==, 0x40000000 );
  CALL( err, ptf_level_to_box_side( 30, &side ), peSuccess );
  CHECK( side, ==, 2 );

  CALL( err, ptf_file_open(buf, ptf_open_update, &file), peSuccess );
  CALL( err, ptf_file_get_header(file, &hdr), peSuccess);
  CHECK( hdr.range.x_min, ==, 0x70000000 );
  CHECK( hdr.range.x_max, ==, 0x90000000 );
  CHECK( hdr.range.y_min, ==, 0x07000000 );
  CHECK( hdr.range.y_max, ==, 0x09000000 );
  CHECK( hdr.range.z_min, ==, 0x00700000 );
  CHECK( hdr.range.z_max, ==, 0x00900000 );
  //  create an actual octree
  hdr.range.x_min = hdr.range.y_min = hdr.range.z_min = 0x0;
  hdr.range.x_max = hdr.range.y_max = hdr.range.z_max = 0x80000000;
  ptf_octree_handle oh;
  CALL( err, ptf_octree_create(file, &hdr.range, &oh, 0, 0), peSuccess );
  ptf_onode_ref parent, child;
  CALL( err, ptf_octree_get_root(oh, &parent), peSuccess );
  CHECK( (parent.extent.x_max - parent.extent.x_min), ==, 0x80000000 );

  ptf_geopos lower, upper;
  CALL( err, ptf_octree_box_get_coords(parent.extent, &lower, &upper), peSuccess );
  CHECK( upper.x, ==, 67108864.0 );
  CHECK( upper.y, ==, 67108864.0 );
  CHECK( upper.z, ==, 67108864.0 );
  CHECK( lower.x, ==, -67108864.0 );
  CHECK( lower.y, ==, -67108864.0 );
  CHECK( lower.z, ==, -67108864.0 );

  ptf_32 ptf_meters = 0xf0f0f0f0;
  CALL( err, ptf_meters_to_octreecoord( -67108864.0, &ptf_meters ), peSuccess );
  CHECK( ptf_meters, ==, 0 );
  ptf_meters = 0xf0f0f0f0;
  CALL( err, ptf_meters_to_octreecoord( 67108864.0, &ptf_meters ), peSuccess );
  CHECK( ptf_meters, ==, 0x80000000 );

  double meters = 0.;
  CALL( err, ptf_octreecoord_to_meters( 0, &meters ), peSuccess );
  CHECK( meters, ==, -67108864.0 );
  meters = 0.;
  CALL( err, ptf_octreecoord_to_meters( 0x80000000, &meters ), peSuccess );
  CHECK( meters, ==, 67108864.0 );

  ptf_geopos gp;
  gp.x = -100000;
  gp.y = 0;
  gp.z = 5000000;
  CALL( err, ptf_octree_split(oh, gp, parent, &child), peSuccess );
  CHECK( child.node_level, ==, 1 );
  parent = child;
  CALL( err, ptf_octree_split(oh, gp, parent, &child), peSuccess );
  CHECK( child.node_level, ==, 2 );
  parent = child;
  CALL( err, ptf_octree_split(oh, gp, parent, &child), peSuccess );
  CHECK( child.node_level, ==, 3 );
  CHECK( (child.extent.x_max - child.extent.x_min), ==, (0x80000000 >> 3) );

  ptf_ocube cube_check;
  CALL( err, ptf_octree_coord_get_box(gp, 0, &cube_check), peSuccess );
  CHECK( (cube_check.x_max - cube_check.x_min), ==, 0x80000000 );
  CALL( err, ptf_octree_coord_get_box(gp, 3, &cube_check), peSuccess );
  CHECK( (cube_check.x_max - cube_check.x_min), ==, (0x80000000 >> 3) );
  ptf_geopos up;
  gp.x = 0;
  gp.y = 0;
  gp.z = -(PTF_MAX_GEOPOS * 0.25);
  up.x = gp.x + 10;
  up.y = gp.x + 100;
  up.z = -gp.z;
  CALL( err, ptf_octree_coord_get_box_natural( gp, up, &cube_check ), peSuccess );
  CHECK( cube_check.x_min, ==, 0x40000000 );
  CHECK( cube_check.y_min, ==, 0x40000000 );
  CHECK( cube_check.z_min, ==, 0x20000000 );
  CHECK( cube_check.z_max, ==, 0x60000000 );

  CALL( err, ptf_octree_close(oh), peSuccess );
  CALL( err, ptf_file_close(file), peSuccess );

  CALL( err, ptf_file_open_accessors(buf, ptf_open_update, ptf_file_platform_accessors(), NULL, &file), peSuccess );

  CALL( err, ptf_octree_create(file, &hdr.range, &oh, 0, 0), peSuccess );

  CALL( err, ptf_octree_get_root(oh, &parent), peSuccess );
  gp.x = -100000;
  gp.y = 0;
  gp.z = 5000000;
  CALL( err, ptf_octree_split(oh, gp, parent, &child), peSuccess );
  CHECK( child.node_level, ==, 1 );
  CHECK( child.node_index, ==, 1 );
  CALL( err, ptf_octree_resolve(oh, gp, &parent), peSuccess );
  CHECK( parent.node_level, ==, 3 );
  CHECK( parent.node_index, ==, 3 );

  ptf_onode_ref parent1;
  CALL( err, ptf_octree_resolve_to_level(oh, gp, 2, &parent1), peSuccess );
  CHECK( parent1.node_level, ==, 2 );
  CHECK( parent1.node_index, ==, 2 );
  // highest existing level is 3
  CALL( err, ptf_octree_resolve_to_level(oh, gp, 12, &parent1), peSuccess );
  CHECK( parent1.node_level, ==, 3 );
  CHECK( parent1.node_index, ==, 3 );

  savePos = gp;
  gp.z = -10000000;
  CALL( err, ptf_octree_get_root(oh, &parent), peSuccess );
  CALL( err, ptf_octree_split(oh, gp, parent, &child), peSuccess );
  parent = child;
  CALL( err, ptf_octree_split(oh, gp, parent, &child), peSuccess );
  parent = child;
  CALL( err, ptf_octree_split(oh, gp, parent, &child), peSuccess );
  parent = child;
  CALL( err, ptf_octree_split(oh, gp, parent, &child), peSuccess );
  CHECK( child.node_level, ==, 4 );
  CHECK( child.node_index, ==, 7 );
  memset(&parent, 0, sizeof(parent));
  CALL( err, ptf_octree_resolve(oh, gp, &parent), peSuccess );
  CHECK( parent.node_level, ==, 4 );
  CHECK( parent.node_index, ==, 7 );
  CALL( err, ptf_octree_close(oh), peSuccess );

  //  do some stress testing
  hdr.range.x_min = hdr.range.y_min = hdr.range.z_min = 0x0;
  hdr.range.x_max = hdr.range.y_max = hdr.range.z_max = 0x80000000;
  CALL( err, ptf_octree_create(file, &hdr.range, &oh, 0, 0), peSuccess );

#define TEST2_LIMIT 10000
  //  This test stresses depth (down to 25).
  //  it simulates a very fine-grained scene in a small area
  //  of about 2 km.
  for (int i = 0; i < TEST2_LIMIT; ++i) {
    //  two kilometer side cube
    gp = random_coord(-1000, 1000);
    //  put it around Z-ville
    gp.x -= 550;
    gp.y += 36050;
    gp.z += 6370000.0;
    CALL( err, ptf_octree_get_root(oh, &parent), peSuccess );
    int depth = (rand() & 0xf) + 10;
    for (int d = 0; d < depth; ++d) {
      CALL( err, ptf_octree_split(oh, gp, parent, &child), peSuccess );
      parent = child;
    }
  }

#define TEST1_LIMIT 1000      //  I set it up to 1000000 and it worked
  //  This test stresses a bunch of different areas. It 
  //  simulates putting a large number of rocks all around 
  //  the earth and space -- it's as worst case as you can 
  //  get for an octree.
  for (int i = 0; i < TEST1_LIMIT; ++i) {
    gp = random_coord(PTF_MIN_GEOPOS, PTF_MAX_GEOPOS);
    CALL( err, ptf_octree_get_root(oh, &parent), peSuccess );
    int depth = (rand() & 0xf) + 5;
    for (int d = 0; d < depth; ++d) {
      CALL( err, ptf_octree_split(oh, gp, parent, &child), peSuccess );
      parent = child;
    }
  }

  CALL( err, ptf_octree_resolve(oh, savePos, &parent), peSuccess );
  parent.content_index = 2;
  CALL( err, ptf_octree_update_content_index(oh, parent), peSuccess );
  child = parent;

  CALL( err, ptf_octree_close(oh), peSuccess );
  CALL( err, ptf_file_close(file), peSuccess );

  CALL( err, ptf_file_open_accessors(buf, ptf_open_read, ptf_file_platform_accessors(), NULL, &file), peSuccess );
  CALL( err, ptf_octree_create(file, &hdr.range, &oh, 0, 0), peSuccess );

  CALL( err, ptf_octree_resolve(oh, savePos, &parent), peSuccess );
  CHECK( parent.octree_index, ==, child.octree_index );
  CHECK( parent.node_index, ==, child.node_index );
  CHECK( parent.content_index, ==, child.content_index );
  CHECK( parent.content_index, ==, 2 );

  CALL( err, ptf_octree_close(oh), peSuccess );
  CALL( err, ptf_file_close(file), peSuccess );

  CALL( err, ptf_file_open(buf, ptf_open_clobber, &file), peSuccess );
  hdr.range.x_min = 0x03800000;
  hdr.range.x_max = 0x04000000;
  hdr.range.y_min = 0x20000000;
  hdr.range.y_max = 0x20800000;
  hdr.range.z_min = 0x00800000;
  hdr.range.z_max = 0x01000000;
  //  This creates a box at level 8
  CALL( err, ptf_octree_create(file, &hdr.range, &oh, 0, 0), peSuccess );

  CALL( err, ptf_octree_get_root(oh, &parent), peSuccess );
  CALL( err, ptf_octree_box_get_coords(parent.extent, &lower, &upper), peSuccess );

  CALL( err, ptf_octree_resolve_to_level(oh, lower, 9, &child), peSuccess );
  CHECK( child.node_level, ==, 0 );
  CHECK( child.octree_index, ==, 1 );

  upper.x -= 1;
  upper.y -= 1;
  upper.z -= 1;
  CALL( err, ptf_octree_resolve_to_level(oh, upper, 9, &child), peSuccess );
  CHECK( child.node_level, ==, 0 );
  CHECK( child.octree_index, ==, 1 );

  CALL( err, ptf_octree_close(oh), peSuccess );
  CALL( err, ptf_file_close(file), peSuccess );
}

void test_ptf_content(wchar_t const * filename1)
{
  ptf_char buf[100];
  ptf_err err;
  ptf_file_handle file;
  ptf_file_header hdr;
  ptf_content_handle cont1, cont2;
  ptf_32 cnt = 1;
  void const * ptr;
  size_t size;
  ptf_type type;
  ptf_fixed_ref fr;
  ptf_geopos gp1, gp2, gp;
  ptf_ori ori = { 0, 0, 0, 1 }, ori1;

  CHECK( ptf_str_from_wchar(filename1, buf, 100), >, 0 );
  CALL( err, ptf_file_open_accessors(buf, ptf_open_update, ptf_file_platform_accessors(), NULL, &file), peSuccess );
  CALL( err, ptf_file_get_header(file, &hdr), peSuccess);

  CALL( err, ptf_content_create(&cont1), peSuccess );
  CALL( err, ptf_content_create(&cont2), peSuccess );

  CALL( err, ptf_content_append_to_file(cont2, file, &fr), peSuccess );
  CHECK( fr.index, ==, 1 );

  CALL( err, ptf_content_set_annotation(cont1, "hello, world!\n", 15, 0), peSuccess );
  CALL( err, ptf_content_get_annotation(cont1, &ptr, &size, &type), peSuccess );
  CHECK( strcmp((char const *)ptr, "hello, world!\n"), ==, 0 );
  CALL( err, ptf_content_count_references(cont1, &cnt), peSuccess );
  CHECK( cnt, ==, 0 );
  CALL( err, ptf_content_count_reference_instances(cont1, 1, &cnt), peInvalidArguments );

  ptf_record_ref rref;
  memset(&rref, 0, sizeof(rref));
  rref.type = 0xf0000000;
  rref.index = 1;
  CALL( err, ptf_content_add_reference(cont1, rref, &cnt), peSuccess );
  CHECK( cnt, ==, 0 );
  rref.index = 2;
  CALL( err, ptf_content_add_reference(cont1, rref, &cnt), peSuccess );
  CHECK( cnt, ==, 1 );
  rref.index = 3;
  CALL( err, ptf_content_add_reference(cont1, rref, &cnt), peSuccess );
  CHECK( cnt, ==, 2 );
  CALL( err, ptf_content_remove_reference(cont1, 3), peInvalidArguments );
  CALL( err, ptf_content_remove_reference(cont1, 2), peSuccess );
  CALL(err, ptf_content_count_references (cont1, &cnt), peSuccess );
  CHECK( cnt, ==, 2 );
  rref.index = 4;
  CALL( err, ptf_content_add_reference(cont2, rref, &cnt), peSuccess );
  CHECK( cnt, ==, 0 );
  CALL( err, ptf_content_set_reference_annotation(cont1, 0, "ref X", 6, 0), peSuccess );
  CALL( err, ptf_content_set_reference_annotation(cont1, 1, "ref 2", 6, 0), peSuccess );
  CALL( err, ptf_content_get_reference_annotation(cont1, 0, &ptr, &size, &type), peSuccess );
  CHECK( strcmp((char const *)ptr, "ref X"), ==, 0 );
  CALL( err, ptf_content_set_reference_annotation(cont1, 0, "ref 1", 6, 0), peSuccess );
  CALL( err, ptf_content_get_reference_annotation(cont1, 0, &ptr, &size, &type), peSuccess );
  CHECK( strcmp((char const *)ptr, "ref 1"), ==, 0 );

  gp1 = random_coord(-7000000, 7000000);
  CALL( err, ptf_content_add_reference_instance(cont1, 0, gp1, ori, &cnt), peSuccess );
  CHECK( cnt, ==, 0 );
  gp2 = random_coord(-7000000, 7000000);
  CALL( err, ptf_content_add_reference_instance(cont1, 0, gp2, ori, &cnt), peSuccess );
  CHECK( cnt, ==, 1 );

  gp2 = random_coord(-7000000, 7000000);
  CALL( err, ptf_content_update_reference_instance(cont1, 0, cnt, gp2, ori), peSuccess );

  CALL( err, ptf_content_update_to_file(cont2, file, fr), peSuccess );
  CHECK( fr.index, ==, 1 );
  CALL( err, ptf_content_append_to_file(cont1, file, &fr), peSuccess );
  CHECK( fr.index, ==, 2 );

  CALL( err, ptf_content_destroy(cont1), peSuccess );
  CALL( err, ptf_content_destroy(cont2), peSuccess );

  CALL( err, ptf_content_create(&cont1), peSuccess );

  fr.index = 1;
  CALL( err, ptf_content_read_from_file(cont1, file, fr), peSuccess );
  CALL( err, ptf_content_count_references(cont1, &cnt), peSuccess );
  CHECK( cnt, ==, 1 );
  CALL( err, ptf_content_count_reference_instances(cont1, 0, &cnt), peSuccess );
  CHECK( cnt, ==, 0 );
  fr.index = 2;
  CALL( err, ptf_content_read_from_file(cont1, file, fr), peSuccess );
  CALL( err, ptf_content_count_references(cont1, &cnt), peSuccess );
  CHECK( cnt, ==, 2 );
  CALL( err, ptf_content_count_reference_instances(cont1, 0, &cnt), peSuccess );
  CHECK( cnt, ==, 2 );
  CALL( err, ptf_content_count_reference_instances(cont1, 1, &cnt), peSuccess );
  CHECK( cnt, ==, 0 );
  CALL( err, ptf_content_get_reference_instance(cont1, 0, 0, &gp, &ori1), peSuccess );
  CHECK( memcmp(&gp, &gp1, sizeof(gp)), ==, 0 );
  CHECK( memcmp(&ori1, &ori, sizeof(ori)), ==, 0 );
  CALL( err, ptf_content_get_reference_instance(cont1, 0, 1, &gp, &ori1), peSuccess );
  CHECK( memcmp(&gp, &gp2, sizeof(gp)), ==, 0 );
  CHECK( memcmp(&ori1, &ori, sizeof(ori)), ==, 0 );

  CALL( err, ptf_content_destroy(cont1), peSuccess );

  CALL( err, ptf_file_close(file), peSuccess );
}

void test_ptf_mesh(wchar_t const * filename)
{
  ptf_char buf[100];
  ptf_err err;
  ptf_file_handle file;
  ptf_file_header hdr;
  ptf_32 cnt = 1;
  ptf_fixed_ref fr;
  ptf_mesh_handle mesh1, mesh2;

  static unsigned short ind1a[] = {
    0, 1, 3, 3, 1, 4, 1, 2, 4, 4, 2, 5,
  };
  static unsigned short ind1b[] = {
    3, 4, 6, 6, 4, 7, 4, 5, 7, 7, 5, 8,
  };
  static float vert1[][3] = {
    { -1, -1, 0 }, { 0, -1, 0.1f }, { 1, -1, 0.5f },
    { -1, 0, 0.5f }, { 0, 0, 1.0f }, { 1, 0, 0.5f },
    { -1, 1, 0.3f }, { 0, 1, 0.5f }, { 1, 1, 0.2f },
  };
  static ptf_vertex_component_description vcomp1[] = {
    { pvuPosition, 0, pvtFloat32, 3 },
  };
  const ptf_vertex_component_description *vcomp1check;
  ptf_16 vcomp_count;
  ptf_16 vcomp_alignment;
  ptf_mesh_subset_description desc1;
  memset(&desc1, 0, sizeof(desc1));
  desc1.index_type = pvtUInt16;
  desc1.max_index = 5;
  desc1.min_index = 0;
  desc1.num_indices = 12;
  desc1.primitive_type = 3;
  ptf_mesh_subset_description desc2 = desc1;
  desc2.min_index = 3;
  desc2.max_index = 8;
  void const * ptr_check;

  CHECK( ptf_str_from_wchar(filename, buf, 100), >, 0 );
  CALL( err, ptf_file_open_accessors(buf, ptf_open_update, ptf_file_platform_accessors(), NULL, &file), peSuccess );
  CALL( err, ptf_file_get_header(file, &hdr), peSuccess);
  
  CALL( err, ptf_mesh_create(&mesh1, (ptf_char const *)"mesh_name"), peSuccess );

  ptf_subset_information const *v_info;
  ptf_16 v_info_count;
  CALL( err, ptf_mesh_get_subset_information(mesh1, 0, &v_info, &v_info_count), peInvalidArguments );

  CALL( err, ptf_mesh_add_mesh_subset(mesh1, desc1, ind1a), peSuccess );
  CALL( err, ptf_mesh_set_vertex_components(mesh1, vcomp1, 1, 4), peSuccess );
  CALL( err, ptf_mesh_get_vertex_components(mesh1, &vcomp1check, &vcomp_count, &vcomp_alignment), peSuccess );
  CHECK( vcomp_count, ==, 1 );
  CHECK( vcomp_alignment, ==, 4 );
  CHECK( vcomp1check[0].type, ==, pvtFloat32 );
  CHECK( vcomp1check[0].type_count, ==, 3 );
  CHECK( vcomp1check[0].usage, ==, pvuPosition );
  CHECK( vcomp1check[0].usage_index, ==, 0 );
  
  CALL( err, ptf_mesh_add_mesh_subset(mesh1, desc2, ind1b), peSuccess );
  ptf_mesh_subset_description desc_check;
  CALL( err, ptf_mesh_get_mesh_subset(mesh1, 1, &desc_check, &ptr_check), peSuccess );
  CHECK( desc_check.index_type, ==, pvtUInt16 );
  CHECK( desc_check.max_index, ==, 8 );
  CHECK( desc_check.min_index, ==, 3 );
  CHECK( desc_check.num_indices, ==, 12 );
  CHECK( desc_check.primitive_type, ==, 3 );
  CHECK( ((unsigned short*)ptr_check)[0], ==, 3 );
  CHECK( ((unsigned short*)ptr_check)[1], ==, 4 );
  CHECK( ((unsigned short*)ptr_check)[2], ==, 6 );

  CALL( err, ptf_mesh_add_mesh_subset(mesh1, desc2, ind1b), peSuccess );
  CALL( err, ptf_mesh_remove_mesh_subset(mesh1, 2), peSuccess );

  CALL( err, ptf_mesh_check_consistency(mesh1), peInvalidArguments );
  CALL( err, ptf_mesh_set_vertex_data(mesh1, vert1, 12, 9), peSuccess );

  CALL( err, ptf_mesh_get_subset_information(mesh1, 0, &v_info, &v_info_count), peSuccess );
  CHECK( v_info_count, ==, 2 );
  CHECK( v_info[0].type, ==, pviSubsetInfoMin );
  CHECK( v_info[1].type, ==, pviSubsetInfoMax );
  float const *fcp = (float const *)v_info[0].data;
  CHECK( fcp[0], ==, -1.0f );
  CHECK( fcp[1], ==, -1.0f );
  CHECK( fcp[2], ==, 0 );
  fcp = (float const *)v_info[1].data;
  CHECK( fcp[0], ==, 1.0f );
  CHECK( fcp[1], ==, 0.0f );  //  not all vertices are examined in this subset
  CHECK( fcp[2], ==, 1.0f );

  CALL( err, ptf_mesh_get_subset_information(mesh1, 1, &v_info, &v_info_count), peSuccess );
  CHECK( v_info_count, ==, 2 );
  CHECK( v_info[0].type, ==, pviSubsetInfoMin );
  CHECK( v_info[1].type, ==, pviSubsetInfoMax );
  fcp = (float const *)v_info[0].data;
  CHECK( fcp[0], ==, -1.0f );
  CHECK( fcp[1], ==, 0.0f );
  CHECK( fcp[2], ==, 0.2f );
  fcp = (float const *)v_info[1].data;
  CHECK( fcp[0], ==, 1.0f );
  CHECK( fcp[1], ==, 1.0f );  //  not all vertices are examined in this subset
  CHECK( fcp[2], ==, 1.0f );

  ptf_char const *str;
  CALL( err, ptf_mesh_get_name(mesh1, &str), peSuccess );
  CHECK( strcmp((char const *)str, "mesh_name"), ==, 0 );
  CALL( err, ptf_mesh_set_name(mesh1, (ptf_char const *)"some other name"), peSuccess );
  CALL( err, ptf_mesh_get_name(mesh1, &str), peSuccess );
  CHECK( strcmp((char const *)str, "some other name"), ==, 0 );

  ptf_16 vertex_size_check;
  ptf_32 vertex_count_check;
  CALL( err, ptf_mesh_get_vertex_data(mesh1, &ptr_check, &vertex_size_check, &vertex_count_check), peSuccess );
  CHECK( ((float*)ptr_check)[0], ==, -1.0 );
  CHECK( ((float*)ptr_check)[1], ==, -1.0 );
  CHECK( ((float*)ptr_check)[2], ==, 0.0 );
  CHECK( vertex_size_check, ==, 12 );
  CHECK( vertex_count_check, ==, 9 );

  // test edge list stuff
  // -- left boundary
  CALL( err, ptf_mesh_set_edge_plane( mesh1, MeshLeftPlane, 100.5 ), peSuccess );
  ptf_mesh_edge edge;
  edge.edge_id = 1;
  edge.start_vtx = 0;
  edge.end_vtx = 3;
  edge.material_index = 42;
  CALL( err, ptf_mesh_add_edge( mesh1, MeshLeftPlane, &edge ), peSuccess );
  edge.edge_id = 2;
  edge.start_vtx = 3;
  edge.end_vtx = 6;
  edge.material_index = 43;
  CALL( err, ptf_mesh_add_edge( mesh1, MeshLeftPlane, &edge ), peSuccess );

  double plane_d = 0.;
  CALL( err, ptf_mesh_get_edge_plane( mesh1, MeshLeftPlane, &plane_d ), peSuccess );
  CHECK( plane_d, ==, 100.5 );

  ptf_32 edge_count = 0;
  CALL( err, ptf_mesh_count_edges( mesh1, MeshLeftPlane, &edge_count ), peSuccess );
  CHECK( edge_count, ==, 2 );
  // right plane s/b undefined
  CALL( err, ptf_mesh_count_edges( mesh1, MeshRightPlane, &edge_count ), peTypeNotFound );

  ptf_mesh_edge read_edge;
  CALL( err, ptf_mesh_get_edge( mesh1, MeshLeftPlane, 1, &read_edge ), peSuccess );
  CHECK( read_edge.edge_id, ==, 2 );
  CHECK( read_edge.start_vtx, ==, 3 );
  CHECK( read_edge.end_vtx, ==, 6 );
  CHECK( read_edge.material_index, ==, 43 );
  CALL( err, ptf_mesh_get_edge( mesh1, MeshLeftPlane, 0, &read_edge ), peSuccess );
  CHECK( read_edge.edge_id, ==, 1 );
  CHECK( read_edge.start_vtx, ==, 0 );
  CHECK( read_edge.end_vtx, ==, 3 );
  CHECK( read_edge.material_index, ==, 42 );

  // --- top boundary
  CALL( err, ptf_mesh_set_edge_plane( mesh1, MeshTopPlane, 200.25 ), peSuccess );
  edge.edge_id = 10;
  edge.start_vtx = 6;
  edge.end_vtx = 7;
  edge.material_index = 33;
  CALL( err, ptf_mesh_add_edge( mesh1, MeshTopPlane, &edge ), peSuccess );
  edge.edge_id = 11;
  edge.start_vtx = 7;
  edge.end_vtx = 8;
  edge.material_index = 36;
  CALL( err, ptf_mesh_add_edge( mesh1, MeshTopPlane, &edge ), peSuccess );

  // --- hither boundary
  CALL( err, ptf_mesh_set_edge_plane( mesh1, MeshHitherPlane, 300.125 ), peSuccess );
  edge.edge_id = 42;
  edge.start_vtx = 3;
  edge.end_vtx = 5;
  edge.material_index = 9;
  CALL( err, ptf_mesh_add_edge( mesh1, MeshHitherPlane, &edge ), peSuccess );

  ptf_subset_information psi;
  psi.data = "hi";
  psi.size = 3;
  psi.type = 0x80004711UL;  //  for testing
  CALL( err, ptf_mesh_set_subset_information(mesh1, 1, &psi, 1), peSuccess );

  CALL( err, ptf_mesh_append_to_file(mesh1, file, &fr), peSuccess );
  CALL( err, ptf_mesh_update_to_file(mesh1, file, fr), peSuccess );

  CALL( err, ptf_mesh_create(&mesh2, (ptf_char const *)"another_name"), peSuccess );
  CALL( err, ptf_mesh_read_from_file(mesh2, file, fr), peSuccess );
  CALL( err, ptf_mesh_count_mesh_subsets(mesh2, &cnt), peSuccess );
  CHECK( cnt, ==, 2 );
  CALL( err, ptf_mesh_get_subset_information(mesh2, 0, &v_info, &v_info_count), peSuccess );
  CHECK( v_info_count, ==, 2 );
  CHECK( v_info[0].type, ==, pviSubsetInfoMin );
  CHECK( v_info[1].type, ==, pviSubsetInfoMax );
  CALL( err, ptf_mesh_get_subset_information(mesh2, 1, &v_info, &v_info_count), peSuccess );
  CHECK( v_info_count, ==, 1 );
  CHECK( v_info[0].type, ==, 0x80004711UL );
  CHECK( v_info[0].size, ==, 3 );
  CHECK( strcmp((char const *)v_info[0].data, "hi"), ==, 0 );

  // test write/read of edge lists
  CALL( err, ptf_mesh_get_edge_plane( mesh2, MeshLeftPlane, &plane_d ), peSuccess );
  CHECK( plane_d, ==, 100.5 );
  CALL( err, ptf_mesh_count_edges( mesh1, MeshLeftPlane, &edge_count ), peSuccess );
  CHECK( edge_count, ==, 2 );
  CALL( err, ptf_mesh_get_edge( mesh1, MeshLeftPlane, 1, &read_edge ), peSuccess );
  CHECK( read_edge.edge_id, ==, 2 );
  CHECK( read_edge.start_vtx, ==, 3 );
  CHECK( read_edge.end_vtx, ==, 6 );
  CHECK( read_edge.material_index, ==, 43 );
  CALL( err, ptf_mesh_get_edge( mesh1, MeshLeftPlane, 0, &read_edge ), peSuccess );
  CHECK( read_edge.edge_id, ==, 1 );
  CHECK( read_edge.start_vtx, ==, 0 );
  CHECK( read_edge.end_vtx, ==, 3 );
  CHECK( read_edge.material_index, ==, 42 );

  CALL( err, ptf_mesh_get_edge_plane( mesh2, MeshTopPlane, &plane_d ), peSuccess );
  CHECK( plane_d, ==, 200.25 );
  CALL( err, ptf_mesh_count_edges( mesh1, MeshTopPlane, &edge_count ), peSuccess );
  CHECK( edge_count, ==, 2 );
  CALL( err, ptf_mesh_get_edge( mesh1, MeshTopPlane, 1, &read_edge ), peSuccess );
  CHECK( read_edge.edge_id, ==, 11 );
  CHECK( read_edge.start_vtx, ==, 7 );
  CHECK( read_edge.end_vtx, ==, 8 );
  CHECK( read_edge.material_index, ==, 36 );
  CALL( err, ptf_mesh_get_edge( mesh1, MeshTopPlane, 0, &read_edge ), peSuccess );
  CHECK( read_edge.edge_id, ==, 10 );
  CHECK( read_edge.start_vtx, ==, 6 );
  CHECK( read_edge.end_vtx, ==, 7 );
  CHECK( read_edge.material_index, ==, 33 );

  CALL( err, ptf_mesh_get_edge_plane( mesh2, MeshHitherPlane, &plane_d ), peSuccess );
  CHECK( plane_d, ==, 300.125 );
  CALL( err, ptf_mesh_count_edges( mesh1, MeshHitherPlane, &edge_count ), peSuccess );
  CHECK( edge_count, ==, 1 );
  CALL( err, ptf_mesh_get_edge( mesh1, MeshHitherPlane, 0, &read_edge ), peSuccess );
  CHECK( read_edge.edge_id, ==, 42 );
  CHECK( read_edge.start_vtx, ==, 3 );
  CHECK( read_edge.end_vtx, ==, 5 );
  CHECK( read_edge.material_index, ==, 9 );

  CALL( err, ptf_mesh_destroy(mesh1), peSuccess );
  CALL( err, ptf_mesh_destroy(mesh2), peSuccess );
  CALL( err, ptf_file_close(file), peSuccess );
}


static void test_ptf_block()
{
  impl::block_t b;
  bool r, s = false;
  int q = 0;
  CALL( r, b.put((int)3), true );
  CHECK( *(int const*)b.data(), ==, 3 );
  CALL( r, b.put((unsigned char)0xff), true );
  CALL( r, b.seek(0, 0), true );
  CALL( r, b.get(q), true );
  ptf_char buf[1024];
  CALL( r, b.get(buf, 1024), false );
  CALL( r, b.get(s), true );
  CHECK( q, ==, 3 );
  CHECK( s, !=, false );
  impl::block_t b2(b);
  size_t pos = b2.pos();
  b.put((ptf_char const *)"hello, world!");
  impl::block_t b3; b3 = b;
  CHECK( b2.size(), <, b3.size() );
  CHECK( b.size(), ==, b3.size() );
  CALL( r, b.seek( pos, 0 ), true );
  CALL( r, b.get( buf, 10 ), false );
  CALL( r, b.get( buf, 1024 ), true );
  CALL( r, b2.get( buf, 1024 ), false );
  CALL( r, b2.put( (ptf_char const *)"foobar" ), true );
  CALL( r, b2.seek( pos, 0 ), true );
  CALL( r, b2.get( buf, 1024 ), true );
  CHECK( strcmp((char const *)buf, "foobar"), ==, 0 );
  CALL( r, b3.seek( pos, 0 ), true );
  CALL( r, b3.get( buf, 1024 ), true );
  CHECK( strcmp((char const *)buf, "hello, world!"), ==, 0 );
}

void test_ptf_material(wchar_t const * filename)
{
  ptf_char buf[100];
  ptf_err err = (ptf_err)1;
  ptf_file_handle file;
  ptf_file_header hdr;
  ptf_32 cnt = 1;
  ptf_16 version = 0;
  ptf_fixed_ref fr;
  ptf_material_archive arc = 0;
  ptf_material_handle mat = 0;
  ptf_material_handle fmat = 0;
  ptf_material_parameter texmap = 0;
  ptf_material_parameter ftexmap = 0;
  ptf_material_parameter param = 0;
  ptf_material_parameter fparam = 0;
  float fv = 1.0f;
  ptf_char const * o_name;
  ptf_material_type o_type;
  size_t o_size;

  CALL( err, ptf_material_archive_create(&arc), peSuccess );
  CHECK( arc, !=, 0 );
  CALL( err, ptf_material_archive_create_material(arc, (ptf_char const *)"deleteme", &mat), peSuccess );
  CHECK( mat, !=, 0 );
  CALL( err, ptf_material_archive_count_materials(arc, &cnt), peSuccess );
  CHECK( cnt, ==, 1 );
  CALL( err, ptf_material_remove(mat), peSuccess );
  mat = 0;
  CALL( err, ptf_material_archive_count_materials(arc, &cnt), peSuccess );
  CHECK( cnt, ==, 0 );

  CALL( err, ptf_material_archive_create_material(arc, (ptf_char const *)"concrete", &mat), peSuccess );
  CHECK( mat, !=, 0 );
  CALL( err, ptf_material_archive_find_material(arc, (ptf_char const *)"concrete", &fmat), peSuccess );
  CALL( err, ptf_material_add_parameter(mat, (ptf_char const *)"deletemetoo", pmtTextureMapExplicit, &texmap), peSuccess );
  CHECK( texmap, !=, 0 );
  CALL( err, ptf_material_find_parameter(mat, (ptf_char const *)"deletemetoo", &ftexmap), peSuccess );
  CALL( err, ptf_parameter_remove(ftexmap), peSuccess );
  CALL( err, ptf_material_add_parameter(mat, (ptf_char const *)"diffuse", pmtTextureMapExplicit, &texmap), peSuccess );
  CHECK( texmap, !=, 0 );

  // Check that zero-length and large data do not corrupt file
  ptf_material_parameter zero = 0;
  ptf_material_parameter toobig = 0;
  CALL( err, ptf_material_add_parameter(mat, (ptf_char const *)"zero", pmtString, &zero), peSuccess );
  CHECK( zero, !=, 0 );
  CALL( err, ptf_material_add_parameter(mat, (ptf_char const *)"toobig", pmtString, &toobig), peSuccess );
  CHECK( toobig, !=, 0 );
  char bigdata[1024*30];
  CALL( err, ptf_parameter_set_value(toobig, &bigdata[0], sizeof(bigdata)), peSuccess );

  CALL( err, ptf_parameter_add_parameter(texmap, (ptf_char const *)"texture_data", pmtRecordRef, &param), peSuccess );
  CHECK( param, !=, 0 );
  ptf_record_ref rref;
  memset(&rref, 0, sizeof(rref));
  rref.type = ptTextureData;
  rref.index = 1;
  CALL( err, ptf_parameter_set_value(param, &rref, sizeof(rref)), peSuccess );
  fv = 12.0f;
  CALL( err, ptf_parameter_add_parameter(texmap, (ptf_char const *)"physical_scale_u", pmtFloatTuple, &param), peSuccess );
  CALL( err, ptf_parameter_set_value(param, &fv, sizeof(fv)), peSuccess );
  fv = 18.0f;
  CALL( err, ptf_parameter_add_parameter(texmap, (ptf_char const *)"physical_scale_v", pmtFloatTuple, &param), peSuccess );
  CALL( err, ptf_parameter_set_value(param, &fv, sizeof(fv)), peSuccess );
  CALL( err, ptf_parameter_add_parameter(texmap, (ptf_char const *)"physical_scale_v", pmtFloatTuple, &param), peDuplicateName );
  CALL( err, ptf_parameter_count_parameters(texmap, &cnt), peSuccess );
  CHECK( cnt, ==, 3 );
  CALL( err, ptf_parameter_find_parameter(texmap,(ptf_char const *)"physical_scale_v", &fparam), peSuccess );
  CALL( err, ptf_material_count_parameters(mat, &cnt), peSuccess );
  CHECK( cnt, ==, 3 );
  CALL( err, ptf_material_archive_count_materials(arc, &cnt), peSuccess );
  CHECK( cnt, ==, 1 );

  CHECK( ptf_str_from_wchar(filename, buf, 100), >, 0 );
  CALL( err, ptf_file_open_accessors(buf, ptf_open_clobber, ptf_file_platform_accessors(), NULL, &file), peSuccess );
  CALL( err, ptf_file_get_header(file, &hdr), peSuccess );
  CALL( err, ptf_material_archive_append_to_file(arc, file, &fr), peSuccess );
  CALL( err, ptf_material_archive_destroy(arc), peSuccess );
  CALL( err, ptf_file_close(file), peSuccess );

  CALL( err, ptf_file_open_accessors(buf, ptf_open_update, ptf_file_platform_accessors(), NULL, &file), peSuccess );
  CALL( err, ptf_material_archive_create(&arc), peSuccess);
  CALL( err, ptf_material_archive_read_from_file(arc, file, fr, &version), peSuccess );
  CALL( err, ptf_material_archive_count_materials(arc, &cnt), peSuccess );
  CHECK( cnt, ==, 1 );
  mat = 0;
  CALL( err, ptf_material_archive_get_material(arc, 0, &mat), peSuccess );
  CALL( err, ptf_material_count_parameters(mat, &cnt), peSuccess );
  CHECK( cnt, ==, 3 );
  texmap = 0;
  CALL( err, ptf_material_get_parameter(mat, 0, &texmap), peSuccess );
  CALL( err, ptf_parameter_count_parameters(texmap, &cnt), peSuccess );
  CHECK( cnt, ==, 3 );
  param = 0;
  CALL( err, ptf_parameter_get_parameter(texmap, 2, &param), peSuccess );
  CALL( err, ptf_parameter_get_info(param, &o_name, &o_type, &o_size), peSuccess );
  CHECK( strcmp((char const *)o_name, "physical_scale_v"), ==, 0 );
  CHECK( o_type, ==, pmtFloatTuple );
  CHECK( o_size, ==, sizeof(float) );
  fv = 0;
  CALL( err, ptf_parameter_get_value(param, &fv, &o_size), peSuccess );
  CHECK( fv, ==, 18.0f );
  CHECK( o_size, ==, sizeof(float) );
  CALL( err, ptf_parameter_get_parameter(texmap, 1, &param), peSuccess );
  CALL( err, ptf_parameter_get_value(param, &fv, &o_size), peSuccess );
  CHECK( fv, ==, 12.0f );
  //  now, add another material, with two parameters
  CALL( err, ptf_material_archive_create_material(arc, (ptf_char const *)"stone", &mat), peSuccess );
  CALL( err, ptf_material_add_parameter(mat, (ptf_char const *)"name", pmtString, &param), peSuccess );
  CALL( err, ptf_parameter_set_value(param, "foo", 3), peSuccess );
  CALL( err, ptf_material_add_parameter(mat, (ptf_char const *)"date", pmtString, &param), peSuccess );
  CALL( err, ptf_parameter_set_value(param, "12/25/06", 8), peSuccess );
  CALL( err, ptf_material_archive_update_to_file(arc, file, fr), peSuccess );

  CALL( err, ptf_file_close(file), peSuccess );

  CALL( err, ptf_material_archive_destroy(arc), peSuccess );
}

static void test_ptf_texture(wchar_t const *filename)
{
  ptf_char buf[100];
  ptf_err err = (ptf_err)1;
  ptf_file_handle file;
  ptf_fixed_ref fr;
  ptf_texture_handle tex = 0;
  ptf_texture_data header;
  void const * ptr;
  size_t siz;

  CALL( err, ptf_texture_create(&tex), peSuccess );
  CHECK( tex, !=, 0 );
  ptf_32 *pp = (ptf_32 *)&header, *pe = pp + sizeof(header)/4;
  while (pp < pe) {
    *pp++ = 0xd;
  }
  CALL( err, ptf_texture_resize(tex, &header), peInvalidArguments );
  memset(&header, 0, sizeof(header));
  header.alignment = 8;
  header.border = 1;
  header.depth = 1;
  header.dimensions = 2;
  header.format = ptdRawBgra8;
  header.height = 18;
  header.num_mipmaps = 5;
  header.num_surfaces = 1;
  header.width = 10;
  CALL( err, ptf_texture_resize(tex, &header), peSuccess );
  CALL( err, ptf_texture_set_data(tex, 1, 0, 0, 0), peInvalidArguments );
  char data[18][10][4];
  for (int i = 0; i < 18; ++i) {
    for (int j = 0; j < 10; ++j) {
      for (int k = 0; k < 4; ++k) {
        data[i][j][k] = (i * 10 + j) * (k + 1);
      }
    }
  }
  CALL( err, ptf_texture_set_data(tex, 0, 0, data, sizeof(data)), peSuccess );
  CALL( err, ptf_texture_set_data(tex, 0, 1, data, 10*6*4), peSuccess );
  CALL( err, ptf_texture_set_data(tex, 0, 2, data, 6*4*4), peSuccess );
  CALL( err, ptf_texture_set_data(tex, 0, 3, data, 4*(3+1)*4), peSuccess );
  CALL( err, ptf_texture_set_data(tex, 0, 4, data, 3*(3+1)*4), peSuccess );
  CALL( err, ptf_texture_set_data(tex, 0, 5, data, sizeof(data)), peInvalidArguments );

  CHECK( ptf_str_from_wchar(filename, buf, 100), >, 0 );
  CALL( err, ptf_file_open_accessors(buf, ptf_open_clobber, ptf_file_platform_accessors(), NULL, &file), peSuccess );
  CALL( err, ptf_texture_append_to_file(tex, file, &fr), peSuccess );
  CALL( err, ptf_texture_update_to_file(tex, file, fr), peSuccess );
  CALL( err, ptf_texture_destroy(tex), peSuccess );
  CALL( err, ptf_file_close(file), peSuccess );

  file = 0;
  CALL( err, ptf_file_open_accessors(buf, ptf_open_update, ptf_file_platform_accessors(), NULL, &file), peSuccess );
  CALL( err, ptf_texture_create(&tex), peSuccess );
  CALL( err, ptf_texture_resize(tex, &header), peSuccess );
  CALL( err, ptf_texture_read_from_file(tex, file, fr), peSuccess );
  memset(&header, 0xff, sizeof(header));
  CALL( err, ptf_texture_get_info(tex, &header), peSuccess );
  CHECK( header.format, ==, ptdRawBgra8 );
  CHECK( header.width, ==, 10 );
  CALL( err, ptf_texture_get_data(tex, 0, 4, &ptr, &siz), peSuccess );
  CHECK( siz, ==, 3*(3+1)*4 );
  CHECK( memcmp(data, ptr, siz), ==, 0 );
  CALL( err, ptf_texture_destroy(tex), peSuccess );
  CALL( err, ptf_file_close(file), peSuccess );
}


void test_ptf_reference(wchar_t const * filename, wchar_t const * filenamefixed, wchar_t const * filename2, wchar_t const * filenameRoot)
{
  ptf_char buf[100];
  ptf_err err = (ptf_err)1;
  ptf_file_handle file;
  ptf_file_header hdr;
  ptf_file_definition_record fdr, fdr_ret;
  ptf_32 fdr_idx;

  char name[ 2048 ];
  char const * rel = "";
  char fixed[ 2048 ];

  wcstombs( name, filename2, 2047 );
  wcstombs( fixed, filenamefixed, 2047 );

  CHECK( ptf_str_from_wchar(filename, buf, 100), >, 0 );
  CALL( err, ptf_file_open_accessors(buf, ptf_open_update, ptf_file_platform_accessors(), NULL, &file), peSuccess );
  CALL( err, ptf_file_get_header(file, &hdr), peSuccess);
  
  strcpy( (char*)fdr.name, "" );
  CALL( err, ptf_file_make_reference(file, &fdr, &fdr_idx), peInvalidArguments );

  strcpy( (char*)fdr.name, name );
  strcpy( (char*)fdr.rel_path, rel );
  strcpy( (char*)fdr.fixed_path, fixed );
  CALL( err, ptf_file_make_reference(file, &fdr, &fdr_idx), peSuccess );

  CALL( err, ptf_file_get_reference(file, fdr_idx, &fdr_ret), peSuccess );
  CHECK( strcmp( (char*)fdr.name, (char*)fdr_ret.name ), ==, 0 );
  CHECK( strcmp( (char*)fdr.rel_path, (char*)fdr_ret.rel_path ), ==, 0 );
  CHECK( strcmp( (char*)fdr.fixed_path, (char*)fdr_ret.fixed_path ), ==, 0 );

  CALL( err, ptf_reference_resolve_file_accessors(&fdr_ret, NULL, ptf_file_platform_accessors(), 0), peSuccess );

  fdr_ret.fixed_path[0] = 0;
  ptf_char searchPath[ 1024 ];
  CHECK( ptf_str_from_wchar(filenameRoot, searchPath, 1023), >, 0 );
  // strip off trailing '/' from search path if present
  if ( *searchPath ) {
    int last = wcslen( filenameRoot ) - 1;
    if ( searchPath[last] == '\\' || searchPath[last] == '/' ) {
      searchPath[last] = 0;
    }
  }
  ptf_char * search[ 2 ] = { searchPath, 0 };
  CALL( err, ptf_reference_resolve_file_accessors(&fdr_ret, search, ptf_file_platform_accessors(), 0), peSuccess );
  CHECK( strcmp( (char*)fdr.fixed_path, (char*)fdr_ret.fixed_path ), ==, 0 );

  CALL( err, ptf_file_close(file), peSuccess );
}

void test_ptf_lod(wchar_t const * filename)
{
  ptf_char buf[100];
  ptf_err err = (ptf_err)1;
  ptf_file_handle file;
  ptf_file_header hdr;

  ptf_lod_handle lod;
  ptf_32 count = 0xffffffff;

  CALL( err, ptf_lod_create(&lod, 0), peSuccess );
  CALL( err, ptf_lod_count_semantics(lod, &count), peSuccess );
  CHECK( count, ==, 0 );
  CALL( err, ptf_lod_destroy(lod), peSuccess );

  CALL( err, ptf_lod_create(&lod, (ptf_char const *)"some chunk"), peSuccess );
  CALL( err, ptf_lod_count_semantics(lod, &count), peSuccess );
  CHECK( count, ==, 1 );
  ptf_32 index = 0xffffffff;
  ptf_16 c16 = 0xffff;
  ptf_16 sz16 = 0xffff;
  ptf_32 dz32 = 0xffffffff;
  CALL( err, ptf_lod_get_semantic(lod, plsGroupName, &index, &c16, &sz16, &dz32), peSuccess );
  CHECK( index, ==, 0 );
  CHECK( c16, ==, 1 );
  CHECK( sz16, ==, 0 );
  CHECK( dz32, ==, 16 );
  ptf_32 semantic;
  c16 = 10;
  sz16 = 10;
  dz32 = 10;
  CALL( err, ptf_lod_get_semantic_at(lod, 0, &semantic, &c16, &sz16, &dz32), peSuccess );
  CHECK( semantic, ==, plsGroupName );
  CHECK( c16, ==, 1 );
  CHECK( sz16, ==, 0 );
  CHECK( dz32, ==, 16 );
  strcmp((char *)buf, "xxxx");
  CALL( err, ptf_lod_get_data(lod, 0, buf, 16), peSuccess );
  buf[16] = 0;
  CHECK( strcmp((char *)buf, "some chunk"), ==, 0 );

  strcpy((char *)buf, "Hello, world!\n");
  CALL( err, ptf_lod_set_semantic_blob(lod, plsExperimental, 16, buf), peSuccess );
  CALL( err, ptf_lod_count_semantics(lod, &count), peSuccess );
  CHECK( count, ==, 2 );
  ptf_lod_table_entry lte[4];
  lte[0].feature_size = 10;
  lte[0].mesh.index = 10;
  lte[1].feature_size = 5;
  lte[1].mesh.index = 20;
  lte[2].feature_size = 2.5f;
  lte[2].mesh.index = 40;
  lte[3].feature_size = 1.0f;
  lte[3].mesh.index = 80;
  CALL( err, ptf_lod_set_semantic_items(lod, plsLodTable, (ptf_16)sizeof(lte[0]), 4, lte), peSuccess );
  CALL( err, ptf_lod_count_semantics(lod, &count), peSuccess );
  CHECK( count, ==, 3 );
  CALL( err, ptf_lod_get_semantic_at(lod, 2, &semantic, &c16, &sz16, &dz32), peSuccess );
  CHECK( semantic, ==, plsLodTable );
  CHECK( c16, ==, 4 );
  CHECK( sz16, ==, sizeof(ptf_lod_table_entry) );
  CHECK( dz32, ==, 8*4 );
  CALL( err, ptf_lod_get_data(lod, 1, buf, 16), peSuccess );
  CHECK( strcmp((char const *)buf, "Hello, world!\n"), ==, 0 );
  CALL( err, ptf_lod_get_data(lod, 2, buf, 32), peSuccess );
  CHECK( memcmp(buf, lte, 32), ==, 0 );

  CHECK( ptf_str_from_wchar(filename, buf, 100), >, 0 );
  CALL( err, ptf_file_open_accessors(buf, ptf_open_update, ptf_file_platform_accessors(), NULL, &file), peSuccess );
  CALL( err, ptf_file_get_header(file, &hdr), peSuccess);
  ptf_fixed_ref fr;
  fr.index = 99;
  CALL( err, ptf_lod_append_to_file(lod, file, &fr), peSuccess );
  CALL( err, ptf_file_close(file), peSuccess );
  CHECK( fr.index, ==, 1 );
  CALL( err, ptf_lod_destroy(lod), peSuccess );
  lod = 0;

  CALL( err, ptf_lod_create(&lod, 0), peSuccess );
  CALL( err, ptf_file_open_accessors(buf, ptf_open_update, ptf_file_platform_accessors(), NULL, &file), peSuccess );
  CALL( err, ptf_lod_read_from_file(lod, file, fr), peSuccess );
  CALL( err, ptf_lod_count_semantics(lod, &count), peSuccess );
  CHECK( count, ==, 3 );
  CALL( err, ptf_lod_get_semantic_at(lod, 2, &semantic, &c16, &sz16, &dz32), peSuccess );
  CHECK( semantic, ==, plsLodTable );
  CHECK( c16, ==, 4 );
  CHECK( sz16, ==, sizeof(ptf_lod_table_entry) );
  CHECK( dz32, ==, 8*4 );
  CALL( err, ptf_lod_get_data(lod, 1, buf, 16), peSuccess );
  CHECK( strcmp((char const *)buf, "Hello, world!\n"), ==, 0 );
  CALL( err, ptf_lod_get_data(lod, 2, buf, 32), peSuccess );
  CHECK( memcmp(buf, lte, 32), ==, 0 );
  //  the success of this call isn't actually checked...
  CALL( err, ptf_lod_update_to_file(lod, file, fr), peSuccess );

  CALL( err, ptf_lod_destroy(lod), peSuccess );
  CALL( err, ptf_file_close(file), peSuccess );
}

void test_ptf_collection(wchar_t const * filename)
{
  ptf_char buf[100];
  ptf_err err = (ptf_err)1;
  ptf_file_handle file = 0;
  ptf_collection_handle coll;
  ptf_fixed_ref ref;
  ptf_remote_ref mesh;
  ptf_ocube bounds;
  ptf_32 id;

  CALL( err, ptf_collection_create(&coll, NULL, ptMeshData), peSuccess );

  mesh.file = 0;
  mesh.index = 1;

  // Check misaligned box
  bounds.x_min = bounds.y_min = bounds.z_min = 129;
  bounds.x_max = bounds.y_max = bounds.z_max = bounds.x_min+128;
  CALL( err, ptf_collection_set_root(coll, 13.0, mesh, bounds, &id), peInvalidArguments );

  // Check badly-sized box
  bounds.x_min = bounds.y_min = bounds.z_min = 128;
  bounds.x_max = bounds.y_max = bounds.z_max = bounds.x_min+129;
  CALL( err, ptf_collection_set_root(coll, 13.0, mesh, bounds, &id), peInvalidArguments );

  // Check misaligned and badly-sized box
  bounds.x_min = bounds.y_min = bounds.z_min = 354;
  bounds.x_max = bounds.y_max = bounds.z_max = 736;
  CALL( err, ptf_collection_set_root(coll, 13.0, mesh, bounds, &id), peInvalidArguments );

  // Check adding a good box
  bounds.x_min = bounds.y_min = bounds.z_min = 128;
  bounds.x_max = bounds.y_max = bounds.z_max = bounds.x_min+128;
  CALL( err, ptf_collection_set_root(coll, 10.0, mesh, bounds, &id), peSuccess );
  CHECK( id, ==, 1 );

  // Add a box that's the right size, but too far away to be contained by an octree node one level higher
  bounds.x_min = bounds.y_min = bounds.z_min = 1024;
  bounds.x_max = bounds.y_max = bounds.z_max = bounds.x_min+128;
  CALL( err, ptf_collection_set_root(coll, 13.0, mesh, bounds, &id), peInvalidArguments );

  // Add another good box
  bounds.x_min = 256;
  bounds.y_min = bounds.z_min = 128;
  bounds.x_max = bounds.x_min+128;
  bounds.y_max = bounds.z_max = bounds.y_min+128;
  CALL( err, ptf_collection_set_root(coll, 12.0, mesh, bounds, &id), peInvalidArguments );

  // Check adding a box that extends the bounding cube too far in one dimension
  bounds.x_min = 256+128;
  bounds.y_min = bounds.z_min = 128;
  bounds.x_max = bounds.y_max = bounds.z_max = bounds.x_min+128;
  CALL( err, ptf_collection_set_root(coll, 12.0, mesh, bounds, &id), peInvalidArguments );

  // Add another good box
  bounds.x_min = bounds.y_min = bounds.z_min = 256;
  bounds.x_max = bounds.y_max = bounds.z_max = bounds.x_min+128;
  CALL( err, ptf_collection_set_root(coll, 12.0, mesh, bounds, &id), peInvalidArguments );

  // Check adding a duplicate box
  bounds.x_min = bounds.y_min = bounds.z_min = 256;
  bounds.x_max = bounds.y_max = bounds.z_max = bounds.x_min+128;
  CALL( err, ptf_collection_set_root(coll, 13.0, mesh, bounds, &id), peInvalidArguments );

  // Check adding a box with an invalid parent
  CALL( err, ptf_collection_node_add(coll, 77, 13.0, mesh, 3, &id), peInvalidArguments );

  // Check adding a box with an invalid child index
  CALL( err, ptf_collection_node_add(coll, 1, 13.0, mesh, 33, &id), peInvalidArguments );

  // Add some boxes in a randomish order
  CALL( err, ptf_collection_node_add(coll, 1, 23.0, mesh, 4, &id), peSuccess );
  CHECK( id, ==, 2 );
  CALL( err, ptf_collection_node_add(coll, 2, 23.0, mesh, 1, &id), peSuccess );
  CHECK( id, ==, 3 );
  CALL( err, ptf_collection_node_add(coll, 2, 23.0, mesh, 5, &id), peSuccess );
  CHECK( id, ==, 4 );
  CALL( err, ptf_collection_node_add(coll, 1, 23.0, mesh, 0, &id), peSuccess );
  CHECK( id, ==, 5 );
  CALL( err, ptf_collection_node_add(coll, 4, 23.0, mesh, 2, &id), peSuccess );
  CHECK( id, ==, 6 );
  CALL( err, ptf_collection_node_add(coll, 4, 23.0, mesh, 7, &id), peSuccess );
  CHECK( id, ==, 7 );
  CALL( err, ptf_collection_node_add(coll, 3, 23.0, mesh, 0, &id), peSuccess );
  CHECK( id, ==, 8 );
  CALL( err, ptf_collection_node_add(coll, 1, 23.0, mesh, 7, &id), peSuccess );
  CHECK( id, ==, 9 );

  // Check adding a box with a duplicate child index
  CALL( err, ptf_collection_node_add(coll, 1, 13.0, mesh, 4, &id), peInvalidArguments );

  CHECK( ptf_str_from_wchar(filename, buf, 100), >, 0 );
  CALL( err, ptf_file_open_accessors(buf, ptf_open_clobber, ptf_file_platform_accessors(), NULL, &file), peSuccess );
  CALL( err, ptf_collection_append_to_file(coll, file, &ref), peSuccess );
  CALL( err, ptf_file_close(file), peSuccess );
  CALL( err, ptf_collection_destroy(coll), peSuccess );

  ptf_crec_t *recs;
  ptf_32 num_recs;
  CALL( err, ptf_collection_create(&coll, NULL, ptMeshData), peSuccess );
  CALL( err, ptf_file_open_accessors(buf, ptf_open_read, ptf_file_platform_accessors(), NULL, &file), peSuccess );
  CALL( err, ptf_collection_read_from_file(coll, file, ref), peSuccess );
  CALL( err, ptf_collection_get_records(coll, &recs, &num_recs), peSuccess );
  CHECK( num_recs, ==, 9 );

  CHECK( recs[0].bottom, ==, 128 );
  CHECK( recs[0].feature_size, ==, 10.0 );

  CHECK( recs[1].yon, ==, 128 );
  CHECK( recs[1].feature_size, ==, 23.0 );

  // Check the lineage...
  // Count children
  CHECK( recs[0].num_children, ==, 3 );
  CHECK( recs[1].num_children, ==, 2 );
  CHECK( recs[2].num_children, ==, 0 );
  CHECK( recs[3].num_children, ==, 0 );
  CHECK( recs[4].num_children, ==, 1 );
  CHECK( recs[5].num_children, ==, 2 );
  CHECK( recs[6].num_children, ==, 0 );
  CHECK( recs[7].num_children, ==, 0 );
  CHECK( recs[8].num_children, ==, 0 );

  // Check child start markers
  CHECK( recs[0].children, ==, &recs[1] );
  CHECK( recs[1].children, ==, &recs[4] );
  CHECK( recs[2].children, ==, NULL );
  CHECK( recs[3].children, ==, NULL );
  CHECK( recs[4].children, ==, &recs[6] );
  CHECK( recs[5].children, ==, &recs[7] );
  CHECK( recs[6].children, ==, NULL );
  CHECK( recs[7].children, ==, NULL );
  CHECK( recs[8].children, ==, NULL );

  // Check parents
  CHECK( recs[0].parent, ==, NULL );
  CHECK( recs[1].parent, ==, &recs[0] );
  CHECK( recs[2].parent, ==, &recs[0] );
  CHECK( recs[3].parent, ==, &recs[0] );
  CHECK( recs[4].parent, ==, &recs[1] );
  CHECK( recs[5].parent, ==, &recs[1] );
  CHECK( recs[6].parent, ==, &recs[4] );
  CHECK( recs[7].parent, ==, &recs[5] );
  CHECK( recs[8].parent, ==, &recs[5] );

  CALL( err, ptf_file_close(file), peSuccess );
  CALL( err, ptf_collection_destroy(coll), peSuccess );
}

void test_ptf_delegation(wchar_t const *filename1, wchar_t const *filename2)
{
  ptf_char buf[200];
  ptf_err err = (ptf_err)1;
  ptf_file_handle file1 = 0, file2 = 0;
  ptf_file_header hdr1, hdr2;
  ptf_octree_handle otree1 = 0, otree2 = 0;
  ptf_onode_ref oref, ocld[8];
  ptf_32 index = 0;

  //  Create the second file with a given sub-cube, off-center.
  //  This file will later be added to the first file.
  ptf_str_from_wchar(filename2, buf, sizeof(buf));
  CALL( err, ptf_file_open(buf, ptf_open_clobber, &file2), peSuccess );
  CALL( err, ptf_file_get_header(file2, &hdr2), peSuccess );
  hdr2.range.x_min = 0x00380000;
  hdr2.range.x_max = 0x00480000;
  hdr2.range.y_min = 0x00800000;
  hdr2.range.y_max = 0x00900000;
  hdr2.range.z_min = 0x00100000;
  hdr2.range.z_max = 0x00200000;
  CALL( err, ptf_file_update_header(file2, &hdr2), peSuccess );

  CALL( err, ptf_octree_create(file2, &hdr2.range, &otree2, 0, 0), peSuccess );
  CALL( err, ptf_octree_get_root(otree2, &oref), peSuccess );
  CALL( err, ptf_onode_get_children(otree2, oref, ocld), peSuccess );
  CHECK( oref.octree_index, ==, 1 );
  CHECK( ocld[0].octree_index, ==, 2 );

  //  create the root PTF file
  ptf_str_from_wchar(filename1, buf, sizeof(buf));
  CALL( err, ptf_file_open(buf, ptf_open_clobber, &file1), peSuccess );
  CALL( err, ptf_file_get_header(file1, &hdr1), peSuccess );
  hdr1.range.x_min = 0;
  hdr1.range.x_max = 0x80000000;
  hdr1.range.y_min = 0;
  hdr1.range.y_max = 0x80000000;
  hdr1.range.z_min = 0;
  hdr1.range.z_max = 0x80000000;
  CALL( err, ptf_file_update_header(file1, &hdr1), peSuccess );

  CALL( err, ptf_octree_create(file1, &hdr1.range, &otree1, 0, 0), peSuccess );
  CALL( err, ptf_octree_get_root(otree1, &oref), peSuccess );
  CALL( err, ptf_onode_get_children(otree1, oref, ocld), peSuccess );
  CHECK( oref.octree_index, ==, 1 );
  //  no children, as the cube was aligned
  CHECK( ocld[0].node_level, ==, 0 );

  //  add an external reference
  //  split down to the appropriate level
  //  add an octree reference to the external file
  CALL( err, ptf_octree_set_reference(otree1, file2), peSuccess );

  //  flush
  CALL( err, ptf_file_flush(file1), peSuccess );

  CALL( err, ptf_octree_close(otree2), peSuccess );
  CALL( err, ptf_file_close(file2), peSuccess );
  CALL( err, ptf_octree_close(otree1), peSuccess );
  CALL( err, ptf_file_close(file1), peSuccess );

  //  verify that the data works as intended

  //  open root file
  ptf_str_from_wchar(filename1, buf, sizeof(buf));
  CALL( err, ptf_file_open(buf, ptf_open_read, &file1), peSuccess );
  memset(&hdr1, 0, sizeof(hdr1));
  CALL( err, ptf_file_get_header(file1, &hdr1), peSuccess );
  //  walk until a reference to a new file is found
  otree1 = 0;
  CALL( err, ptf_octree_create(file1, &hdr1.range, &otree1, 0, 0), peSuccess);
  ptf_onode_ref parent, children[8];
  std::deque<ptf_onode_ref> toProcess;
  CALL( err, ptf_octree_get_root(otree1, &parent), peSuccess );
  toProcess.push_back(parent);
  while (toProcess.size() > 0) {
    parent = toProcess.front();
    toProcess.pop_front();
    CALL( err, ptf_onode_get_children(otree1, parent, children), peSuccess );
    for (int i = 0; i < 8; ++i) {
      if (children[i].node_level == 0) {
        continue;
      }
      if (children[i].file_index != 0) {
        parent = children[i];
        goto found_it;
      }
      //  breadth-first search
      toProcess.push_back(children[i]);
    }
  }
  //  If I get here, it means that the queue was exhausted, 
  //  so this assert will always fail -- which is intended, as I 
  //  should find a remote file ref before the queue runs dry
  CHECK( parent.file_index, !=, 0 );
found_it:
  //  resolve that reference
  ptf_64 size = 0;
  ptf_record_ref rref;
  memset(&rref, 0, sizeof(rref));
  rref.type = ptFileDefinition;
  rref.index = parent.file_index;
  CALL( err, ptf_file_get_record_size(file1, &rref, &size), peSuccess );
  CHECK( size, ==, sizeof(ptf_file_definition_record) );
  ptf_file_definition_record fdr;
  CALL( err, ptf_file_read_record(file1, &rref, 0, &fdr, sizeof(fdr)), peSuccess );
  CALL( err, ptf_reference_resolve_file(&fdr, 0), peSuccess );
  //  open the file
  CALL( err, ptf_file_open(fdr.fixed_path, ptf_open_read, &file2), peSuccess );
  CALL( err, ptf_file_get_header(file2, &hdr2), peSuccess );
  //  declare success!
  CALL( err, ptf_octree_close(otree1), peSuccess );
  CALL( err, ptf_file_close(file1), peSuccess );
  CALL( err, ptf_file_close(file2), peSuccess );
}

void test_ptf_spheroid(wchar_t const *filename)
{
  ptf_char buf[100];
  ptf_err err = (ptf_err)1;
  ptf_file_handle file;
  ptf_fixed_ref fr;
  ptf_spheroid_handle spheroid = 0;

  double radius = 2.36;
  double squash = 298.257;
  float texture_u = 0.0;
  float texture_v = 0.0;
  ptf_16 material_id = 1;
  ptf_fixed_ref mat_arc;
  mat_arc.index = 23;

  ptf_spheroid_data data;
  memset( &data, 0, sizeof( ptf_spheroid_data ) );

  CALL( err, ptf_spheroid_create( &spheroid, 
    (ptf_char const *)"AssigningSpheroidNameThatIsTooLongSoCreateSpherodWillFail" ), peInvalidArguments );
  CHECK( spheroid, ==, 0 );
  CALL( err, ptf_spheroid_create( &spheroid, (ptf_char const *)"spheroid1" ), peSuccess );
  CHECK( spheroid, !=, 0 );

  CALL( err, ptf_spheroid_get_data(spheroid, &data), peSuccess );
  assert( strcmp( (const char * )data.name, "spheroid1" ) == 0 );
  data.radius       = radius;
  data.squash       = squash;
  data.material_id  = material_id;
  data.material_arc = mat_arc;
  data.surface      = 1;

  CALL( err, ptf_spheroid_set_data(spheroid, data), peSuccess );
  memset( &data, 0, sizeof( ptf_spheroid_data ) );
  CALL( err, ptf_spheroid_get_data(spheroid, &data), peSuccess );
  assert( strcmp( (const char * )data.name, "spheroid1" ) == 0 );
  CHECK( data.radius, ==, radius );
  CHECK( data.squash, ==, squash );
  CHECK( data.material_id, ==, material_id );
  CHECK( data.material_arc.index, ==, mat_arc.index );
  CHECK( data.surface, ==, 1 );

  CHECK( ptf_str_from_wchar(filename, buf, 100), >, 0 );
  CALL( err, ptf_file_open(buf, ptf_open_clobber, &file), peSuccess );
  CALL( err, ptf_spheroid_append_to_file(spheroid, file, &fr), peSuccess );
  CALL( err, ptf_spheroid_update_to_file(spheroid, file, fr), peSuccess );
  CALL( err, ptf_spheroid_destroy(spheroid), peSuccess );
  CALL( err, ptf_file_close(file), peSuccess );

  file = 0;
  memset( &data, 0, sizeof( ptf_spheroid_data ) );

  CALL( err, ptf_file_open(buf, ptf_open_update, &file), peSuccess );
  CALL( err, ptf_spheroid_create( &spheroid, (ptf_char const *)"spheroid2" ), peSuccess );
  CALL( err, ptf_spheroid_read_from_file(spheroid, file, fr), peSuccess );
  CALL( err, ptf_spheroid_get_data(spheroid, &data), peSuccess );
  assert( strcmp( (const char * )data.name, "spheroid1" ) == 0 );
  CHECK( data.radius, ==, radius );
  CHECK( data.squash, ==, squash );
  CHECK( data.material_id, ==, material_id );
  CHECK( data.material_arc.index, ==, mat_arc.index );
  CHECK( data.surface, ==, 1 );
  CALL( err, ptf_spheroid_destroy(spheroid), peSuccess );
  CALL( err, ptf_file_close(file), peSuccess );
}

void test_ptf_alignment()
{
  // Check for any alignment or data structure changes or problems
  assert( sizeof( ptf_8 ) == 1 );
  assert( sizeof( ptf_char ) == 1 );
  assert( sizeof( ptf_16 ) == 2 );
  assert( sizeof( ptf_32 ) == 4 );
  assert( sizeof( ptf_type ) == 4 );
  assert( sizeof( ptf_64 ) == 8 );
  assert( sizeof( ptf_date ) == 8 );
  assert( sizeof( ptf_name ) == 48 );

  ptf_toc_entry toc;
  unsigned char * pToc = (unsigned char*)&toc;
  assert( sizeof( toc.offset ) == 8 );
  assert( (unsigned char*)&toc.offset == pToc );
  assert( sizeof( toc.hash ) == 8 );
  assert( (unsigned char*)&toc.hash == pToc+8 );
  assert( sizeof( toc.size_disk ) == 4 );
  assert( (unsigned char*)&toc.size_disk == pToc+16 );
  assert( sizeof( toc.size_mem ) == 4 );
  assert( (unsigned char*)&toc.size_mem == pToc+20 );
  assert( sizeof( toc.encryption ) == 2 );
  assert( (unsigned char*)&toc.encryption == pToc+24 );
  assert( sizeof( toc.compression ) == 2 );
  assert( (unsigned char*)&toc.compression == pToc+26 );
  assert( sizeof( toc.reserved ) == 4 );
  assert( (unsigned char*)&toc.reserved == pToc+28 );
  assert( sizeof( toc.type ) == 4 );
  assert( (unsigned char*)&toc.type == pToc+28 );
  assert( sizeof( toc ) == 32 );

  ptf_ocube cube;
  unsigned char * pCube = (unsigned char*)&cube;
  assert( sizeof( cube.x_min ) == 4 );
  assert( (unsigned char*)&cube.x_min == pCube );
  assert( sizeof( cube.x_max ) == 4 );
  assert( (unsigned char*)&cube.x_max == pCube+4 );
  assert( sizeof( cube.y_min ) == 4 );
  assert( (unsigned char*)&cube.y_min == pCube+8 );
  assert( sizeof( cube.y_max ) == 4 );
  assert( (unsigned char*)&cube.y_max == pCube+12 );
  assert( sizeof( cube.z_min ) == 4 );
  assert( (unsigned char*)&cube.z_min == pCube+16 );
  assert( sizeof( cube.z_max ) == 4 );
  assert( (unsigned char*)&cube.z_max == pCube+20 );
  assert( sizeof( cube ) == 24 );

  ptf_file_header hdr;
  unsigned char * pHdr = (unsigned char*)&hdr;
  assert( sizeof( hdr.magic ) == 8 );
  assert( (unsigned char*)&hdr.magic == pHdr );
  assert( sizeof( hdr.file_name ) == 48 );
  assert( (unsigned char*)&hdr.file_name == pHdr+8 );
  assert( sizeof( hdr.spec_version ) == 4 );
  assert( (unsigned char*)&hdr.spec_version == pHdr+56 );
  assert( sizeof( hdr.min_version ) == 4 );
  assert( (unsigned char*)&hdr.min_version == pHdr+60 );
  assert( sizeof( hdr.cr_date ) == 8 );
  assert( (unsigned char*)&hdr.cr_date == pHdr+64 );
  assert( sizeof( hdr.mod_date ) == 8 );
  assert( (unsigned char*)&hdr.mod_date == pHdr+72 );
  assert( sizeof( hdr.alignment ) == 4 );
  assert( (unsigned char*)&hdr.alignment == pHdr+80 );
  assert( sizeof( hdr.spare1 ) == 4 );
  assert( (unsigned char*)&hdr.spare1 == pHdr+84 );
  assert( sizeof( hdr.meta_toc ) == 32 );
  assert( (unsigned char*)&hdr.meta_toc == pHdr+88 );
  assert( sizeof( hdr.range ) == 24 );
  assert( (unsigned char*)&hdr.range == pHdr+120 );
  assert( sizeof( hdr.size ) == 8 );
  assert( (unsigned char*)&hdr.size == pHdr+144 );
  assert( sizeof( hdr.schema ) == 32 );
  assert( (unsigned char*)&hdr.schema == pHdr+152 );
  assert( sizeof( hdr.spare2 ) == 72 );
  assert( (unsigned char*)&hdr.spare2 == pHdr+184 );
  assert( sizeof( hdr ) == 256 );
}

int MAIN
{
  test_ptf_alignment();
  test_ptf_block();
  test_ptf_str();
  test_ptf_date();
  test_ptf_unpack();
  unlink(File(L"c:\\file1.ptf"));
  unlink(File(L"c:\\file2.ptf"));
  // clean up after files accidently created due to a bug
  unlink(File(L".\\file1.ptf"));
  unlink(File(L".\\file2.ptf"));
  test_ptf_material(File(L"c:\\file2.ptf"));
  test_ptf_texture(File(L"c:\\file2.ptf"));
  test_ptf_file(File(L"c:\\file1.ptf"), File(L"c:\\file2.ptf"));
  test_ptf_bookmarks(File(L"c:\\file1.ptf"));
  test_ptf_octree(File(L"c:\\file1.ptf"));
  test_ptf_content(File(L"c:\\file1.ptf"));
  test_ptf_mesh(File(L"c:\\file1.ptf"));
  test_ptf_reference(File(L"c:\\file1.ptf"), File(L"c:\\file2.ptf"), File(L"file2.ptf"), File(L"c:\\"));
  test_ptf_lod(File(L"c:\\file1.ptf"));
  test_ptf_collection(File(L"c:\\file1.ptf"));
  test_ptf_delegation(File(L"c:\\file1.ptf"), File(L"c:\\file2.ptf"));
  test_ptf_spheroid(File(L"c:\\file1.ptf"));
  return 0;
}

