// ptf_copy.cpp : Defines the entry point for the console application.
//

#include <stdio.h>
#include <memory.h>
#include "ptf_file.h"
#include "ptf_str.h"

#define RECBUFSIZE 65536

// macros to check return code from PTF API functions
#define PTF_FAILED(a) ((a) != peSuccess)
#define PTF_SUCCEEDED(a) ((a) == peSuccess)

void printUsage()
{
  fprintf( stdout, "Usage: ptf_copy sourcefile destfile\n" );
}

// output a text message that corresponds to each PTF API return code
void outputPTFErr( ptf_err err )
{
  switch ( err ) {
    case peFileNotFound:
      fprintf( stderr, "The indicated file was not found.\n" );
      break;
    case peInvalidFile:
      fprintf( stderr, "The indicated path is not a PTF file.\n" );
      break;
    case peIoError:
      fprintf( stderr, "Some error when attempting to read/write the file.\n" );
      break;
    case peInvalidVersion:
      fprintf( stderr, "The indicated path is of the wrong version.\n" );
      break;
    case pePermissionDenied:
      fprintf( stderr, "The indicated path is not accessible, or attempt to write to read-only file.\n" );
      break;
    case peInvalidArguments:
      fprintf( stderr, "The arguments to the function are wrong.\n" );
      break;
    case peTooBigForMachine:
      fprintf( stderr, "The data in question is too big for the current machine (i e, needs 64 bit memory).\n" );
      break;
    case peReadOutOfBounds:
      fprintf( stderr, "Attempt to read past the end of a record.\n" );
      break;
    case peInvalidSequence:
      fprintf( stderr, "The attempted function is not currently valid.\n" );
      break;
    case peReadOnly:
      fprintf( stderr, "The file is not open for writing.\n" );
      break;
    case peException:
      fprintf( stderr, "C++ exception occurred while executing -- probably out of memory.\n" );
      break;
    case peOutOfMemory:
      fprintf( stderr, "One of those rare occasions where out of memory is not an exception.\n" );
      break;
    case peTypeNotFound:
      fprintf( stderr, "Request for a type that doesn't exist.\n" );
      break;
    case peRecordNotFound:
      fprintf( stderr, "Record is out of range, or has been deleted.\n" );
      break;
    default:
      fprintf( stderr, "Operation completed successfully.\n" );
  }
}

int copyPTFFile( ptf_char const * pSrc, ptf_char const * pDest )
{
  // open the source file
  ptf_file_handle srcFile;
  int status = 0;
  ptf_err ptf_status;

  static char recBuf[ RECBUFSIZE ];

  fprintf( stdout, "Opening source file...\n" );
  if ( PTF_SUCCEEDED( ptf_status = ptf_file_open( pSrc, ptf_open_read, &srcFile) ) ) {
    // create the output file
    ptf_file_handle destFile;
    fprintf( stdout, "Opening destination file...\n" );
    if ( PTF_SUCCEEDED( ptf_status = ptf_file_open( pDest, ptf_open_clobber, &destFile) ) ) {
      // copy the header first - treat it as a special case
      ptf_file_header header;
      memset( &header, 0, sizeof(ptf_file_header) );
      if ( PTF_SUCCEEDED( ptf_status = ptf_file_get_header( srcFile, &header ) ) ) {
        if ( PTF_SUCCEEDED( ptf_file_update_header( destFile, &header ) ) ) {
          fprintf( stdout, "Copied file header.\n" );
        }
        else {
          fprintf( stdout, "*** ptf_file_update_header() failed for destination file.  Header not copied.\n" );
          outputPTFErr( ptf_status );
        }
      }
      else {
        fprintf( stdout, "*** ptf_file_get_header() failed for source file.  Header not copied.\n" );
        outputPTFErr( ptf_status );
      }

      // iterate through the record types in the source file
      ptf_32 numTypes = 0;
      if ( PTF_FAILED( ptf_status = ptf_file_count_types( srcFile, &numTypes ) ) ) {
        fprintf( stdout, "*** ptf_file_count_types() failed.\n" );
        outputPTFErr( ptf_status );
      }
      
      fprintf( stdout, "Source file contains %d record types.\n", numTypes );
      for ( long typeNdx = 0; typeNdx < numTypes; typeNdx++ ) {
        // get the number of records for this record type
        ptf_type curType = ptNoType;
        if ( PTF_FAILED( ptf_status = ptf_file_get_type_at( srcFile, typeNdx, &curType ) ) ) {
          fprintf( stdout, "*** ptf_file_get_type_at() failed.\n" );
          outputPTFErr( ptf_status );
        }

        ptf_32 numRec = 0;
        if ( PTF_FAILED( ptf_status = ptf_file_count_type_records( srcFile, curType, &numRec ) ) ) {
          fprintf( stdout, "*** ptf_file_count_type_records() failed.\n" );
          outputPTFErr( ptf_status );
        }
        else {
          fprintf( stdout, "  Record type %x contains %d records.\n", curType, numRec );
          ptf_record_ref rref;
          if( PTF_FAILED( ptf_status = ptf_file_allocate_records( destFile, curType, numRec-1, &rref ) ) ) {
            fprintf( stdout, "*** ptf_file_allocate_records() failed.\n" );
            outputPTFErr( ptf_status );
          }
        }

        // API won't let you copy records of type "empty"
        if ( ptEmptyRecord == curType ) {
           fprintf( stdout, "  - Skipping 'empty' records in source file...\n" );
           continue;
         }

        // iterate through the records for this record type
        for ( long curRec = 1; curRec < numRec; curRec++ ) {
          ptf_record_ref srcRef;
          memset(&srcRef, 0, sizeof(srcRef));
          srcRef.index = curRec;
          srcRef.type = curType;

          // NOTE: records can be arbitrarily large; need to copy data in chunks
          ptf_64 recSize = 0;
          if ( PTF_FAILED( ptf_status = ptf_file_get_record_size( srcFile, &srcRef, &recSize ) ) ) {
            //  This will fail if there are deleted recors in the file
            //  I will leave the record in the output file in the deleted state it's in 
            //  when first created.
            continue;
          }

          //  Start a new record in the destination file. Note that the record_ref is the 
          //  same -- this is a principal tenet of PTF file processing.
          if ( PTF_FAILED( ptf_status = ptf_file_replace_record( destFile, &srcRef ) ) ) {
            fprintf( stdout, "*** ptf_file_new_record() failed.\n" );
            outputPTFErr( ptf_status );
          }

          long bytesCopied = 0;
          long bytesRemaining = recSize;
          bool status = true;
          while ( bytesCopied < recSize ) {
            // read the next chunk of data from the source record
            long bytesToCopy = ( bytesRemaining < RECBUFSIZE ) ? bytesRemaining : RECBUFSIZE;

            memset( recBuf, 0xff, RECBUFSIZE );
            if ( PTF_FAILED( ptf_status = ptf_file_read_record( srcFile, &srcRef, bytesCopied, recBuf, bytesToCopy ) ) ) {
              fprintf( stdout, "*** ptf_file_read_record() failed.\n" );
              outputPTFErr( ptf_status );
            }

            // write the chunk to the destination record
            if ( PTF_FAILED( ptf_status = ptf_file_write_record( destFile, recBuf, bytesToCopy ) ) ) {
              fprintf( stdout, "*** ptf_file_write_record() failed.\n" );
              outputPTFErr( ptf_status );
            }
            
            bytesCopied += bytesToCopy;
            bytesRemaining -= bytesToCopy;
          }
          // complete the destination record
          if ( PTF_SUCCEEDED( ptf_status = ptf_file_commit_record( destFile, NULL ) ) ) {
            fprintf( stdout, "  - [%d] Copied record with %d bytes.\n", curRec, recSize );
          }
          else {
            fprintf( stdout, "*** ptf_file_commit_record() failed.\n" );
            outputPTFErr( ptf_status );
          }
        }
        if ( PTF_SUCCEEDED( ptf_status = ptf_file_flush( destFile ) ) ) {
          fprintf( stdout, "  Flushed records of type %x.\n", curType );
       }
        else {
          fprintf( stdout, "*** ptf_file_flush() failed.\n" );
          outputPTFErr( ptf_status );
        }
      }

      if ( PTF_SUCCEEDED( ptf_status = ptf_file_close( destFile ) ) ) {
         fprintf( stdout, "Closed destination file.\n" );
      }
      else {
        fprintf( stdout, "*** ptf_file_close() failed for destination file.\n" );
        outputPTFErr( ptf_status );
      }
     }
    else {
      outputPTFErr( ptf_status );
      fprintf( stdout, "Cannot open destination file for CREATE\n" );
      status = 1;
    }

    if ( PTF_SUCCEEDED( ptf_file_close( srcFile ) ) ) {
        fprintf( stdout, "Closed destination file.\n" );
    }
    else {
      fprintf( stdout, "*** ptf_file_close() failed for source file.\n" );
      outputPTFErr( ptf_status );
    }
  }
  else {
    outputPTFErr( ptf_status );
    fprintf( stdout, "Cannot open source file for READ\n" );
    status = 1;
  }

  return status;
}

int main(int argc, char * argv[])
{
  // usage: ptf_copy source-filename destination-filename
  if ( argc != 3 ) {
    printUsage();
    return 1;
  }

  // make PTF strings out of source and destination filenames
  ptf_char srcFile[ PTF_PATHSIZE ];
  if ( !ptf_str_from_char( argv[ 1 ], srcFile, PTF_PATHSIZE ) ) {
    fprintf( stdout, "Invalid source filename %s\n", argv[ 1 ] );
    return 1;
  }
  
  ptf_char destFile[ PTF_PATHSIZE ];
  if ( !ptf_str_from_char( argv[ 2 ], destFile, PTF_PATHSIZE ) ) {
    fprintf( stdout, "Invalid destination filename %s.\n", argv[ 2 ] );
    return 1;
  }

  fprintf( stdout, "Copying PTF file '%s' to '%s'.\n", argv[ 1 ], argv[ 2 ] );
  int status = copyPTFFile( srcFile, destFile );

	return status;
}

