
#if !defined(ptf_mesh_h)
//! \internal header guard
#define ptf_mesh_h

#if !defined(ptf_file_h)
#include "ptf_file.h"
#endif
#if !defined(ptf_types_h)
#include "ptf_types.h"
#endif
//! \addtogroup PtfGeometry The PTF Geometry Library

//! All mesh data instances is identified by an abstract ptf_mesh_handle.
//! These instances can exist separate from any ptf_file instance, and it
//! is up to the programmer to manage the lifetime of these objects. The 
//! API does provide functions to read and write these instances from/to 
//! PTF files.
typedef class ptf_mesh_t *ptf_mesh_handle;

//! Subset information is retrieved from the mesh in terms of this record.
struct ptf_subset_information {
  ptf_32                      type;           //!< What kind of vertex information.
  ptf_32                      size;           //!< Size of pointed-at data.
  void const                 *data;           //!< Pointer to data.
};

//! Create a mesh instance. You need an instance to be able to read data 
//! into it, or add data and write out to a file.
//! \param o_mesh The mesh instance handle will be stored here. You must 
//! call ptf_mesh_destroy() on it once you're done with it.
//! \return ptf_err status.
PTF_API ptf_err ptf_mesh_create(ptf_mesh_handle *o_mesh, ptf_char const * name);

//! Deallocate all data allocated by a mesh instance.
//! \param mesh The allocated instance to destroy. It will not be valid 
//! after this call returns.
//! \return ptf_err status.
PTF_API ptf_err ptf_mesh_destroy(ptf_mesh_handle mesh);

//! Read a ptMeshData header from a PTF file and verify the data.
//! \param o_hdr The mesh header to populate with data from the file.
//! \param file The PTF file to read from.
//! \param ref The ptMeshData type record index to read from.
//! \return ptf_err status.
//! \see ptf_mesh_header
PTF_API ptf_err ptf_mesh_read_header_from_file(ptf_mesh_header *o_hdr, 
    ptf_file_handle file, ptf_fixed_ref const &fref);

//! Read a ptMeshData block from a PTF file and populate the mesh instance 
//! handle with the data found therein.
//! \param mesh The mesh handle to populate with data from the file.
//! \param file The PTF file to read from.
//! \param ref The ptMeshData type record index to read from.
//! \return ptf_err status.
//! \see ptf_mesh_header
PTF_API ptf_err ptf_mesh_read_from_file(ptf_mesh_handle mesh, 
    ptf_file_handle file, ptf_fixed_ref const &ref);

//! Verify that a mesh is minimally consistent, and return peSuccess 
//! if so is the case. A minimally consistent mesh has the right amount 
//! of vertex data for the vertex size and number allocated, and has 
//! vertex components that fit within the given vertex size. Additionally, 
//! a minimally consistent mesh has one or more subsets, none of which 
//! has a minimum or maximum index value that goes outside the number of 
//! vertices.
//! \param mesh The mesh to verify.
//! \return ptf_err status; peSuccess if consistent.
PTF_API ptf_err ptf_mesh_check_consistency(ptf_mesh_handle mesh);

//! Given a populated mesh handle, and an existing mesh data record in a 
//! PTF file, replace that record with the data of the populated mesh.
//! The mesh in memory is left intact; this is a pure "write" operation.
//! This will return an error if the mesh isn't minimally consistent.
//! \param mesh The mesh to write to file.
//! \param file The file to receive the mesh data.
//! \param ref The index of the ptMeshData record that will receive the data.
//! \return ptf_err status.
//! \see ptf_mesh_header
PTF_API ptf_err ptf_mesh_update_to_file(ptf_mesh_handle mesh, 
    ptf_file_handle file, ptf_fixed_ref const &ref);

//! Given a populated mesh handle, append a new ptMeshData record to the 
//! given PTF file, and return the index of this new record.
//! This will return an error if the mesh isn't minimally consistent.
//! \param mesh The mesh handle to write to the file.
//! \param file The file to receive the mesh handle.
//! \param o_ref The index of the mesh record will be returned here.
//! \return ptf_err status.
//! \see ptf_mesh_header
PTF_API ptf_err ptf_mesh_append_to_file(ptf_mesh_handle mesh, 
    ptf_file_handle file, ptf_fixed_ref *o_ref);

//! Return the name of a mesh instance.
//! \param mesh The allocated (and perhaps read) mesh to get the name for.
//! \param o_name A pointer to the zero-terminated name will be stored here.
//! This pointer is valid until the next manipulation of the mesh only, so 
//! you should copy it out!
//! \return ptf_err status.
PTF_API ptf_err ptf_mesh_get_name(ptf_mesh_handle mesh, ptf_char const **o_name);

//! Set the name of a mesh instance.
//! \param mesh The allocated mesh to set the name for.
//! \param name A pointer to the zero-terminated name to be stored.
//! The data at this pointer will be copied into the mesh, so the pointer 
//! can be deallocated after ptf_mesh_set_name() returns. The maximum size 
//! of a mesh name is sizeof(ptf_name), including the terminating 0.
//! \return ptf_err status.
PTF_API ptf_err ptf_mesh_set_name(ptf_mesh_handle mesh, ptf_char const *name);

//! Given an allocated mesh handle, retrieve the LOD information stored in 
//! the mesh header. This lets you locate other meshes that may represent 
//! the same "thing" but in different resolution.
//! \param mesh The mesh handle to get LOD information about. If there is 
//! no LOD association, an index of 0 will be returned.
//! \param o_ref The LOD table chunk.
//! \param o_feature The "feature size" of the mesh.
//! \return ptf_err status.
PTF_API ptf_err ptf_mesh_get_lod_table(ptf_mesh_handle mesh,
    ptf_fixed_ref *o_ref, float *o_feature);

//! Given an allocated mesh handle, set the LOD table chunk that's associated
//! with this file.
//! \param mesh The mesh handle to set the LOD table information on.
//! \param lod A reference to a LOD chunk, containing the LOD table 
//! that references this mesh and its siblings. The index 0 is not 
//! a valid LOD chunk index. To clear any LOD association, set this 
//! to 0.
//! \param feature The "feature size" of the mesh.
//! \return ptf_err status.
PTF_API ptf_err ptf_mesh_set_lod_table(ptf_mesh_handle mesh,
    ptf_fixed_ref lod, float feature);

//! Given an allocated mesh handle, set the mesh components. This may cause 
//! any previously set vertex data to become invalid, but such data is not 
//! purged until you set new vertex data on the mesh (because it may still 
//! remain valid; there is no ordering dependency between setting the vertex 
//! data and setting the vertex components).
//! \param mesh The mesh to set the component description on.
//! \param comp A pointer to an array of components, describing the vertex 
//! components in order.
//! \param count The number of vertex components.
//! \param alignment The alignment of vertices within the vertex array. This 
//! will also be the alignment of individual data blocks within the file 
//! written out from a mesh handle. A typical value is 4, although 8 is also 
//! useful (though wasteful for components such as 4-byte colors). Many 
//! rendering implementations don't support values that aren't a multiple of 
//! 4.
//! \return ptf_err status.
PTF_API ptf_err ptf_mesh_set_vertex_components(ptf_mesh_handle mesh, 
    ptf_vertex_component_description const *comp, ptf_16 count, ptf_16 alignment);

//! Return a pointer to the array of mesh component descriptions found within 
//! a mesh. This pointer will be valid until the next mutation operation on 
//! the mesh handle. Reading or writing from/to files count as mutation operations, 
//! as do any "set" type functions.
//! \param mesh The mesh to get the components from.
//! \param o_comp A pointer to the component description array will be returned here.
//! \param o_count The number of vertex component descriptions is returned here.
//! \param o_alignment The vertex component alignment (within each vertex) is returned here.
//! \return ptf_err status.
PTF_API ptf_err ptf_mesh_get_vertex_components(ptf_mesh_handle mesh,
    ptf_vertex_component_description const **o_comp, ptf_16 *o_count, ptf_16 *o_alignment);

//! Copy the given data into the given mesh, replacing any previous vertex data. 
//! It is up to the user to ensure that the vertex data and the vertex component 
//! descriptions match up before any output functions are called. You can use 
//! ptf_mesh_check_consistency() to verify that you got it right.
//! \param mesh The mesh to copy the vertex data into.
//! \param vertex_size The size of each individual vertex (including alignment).
//! \param vertex_count The number of vertices in the array. Thus, the size of the 
//! data block will be vertex_size*vertex_count bytes.
//! \return ptf_err status.
PTF_API ptf_err ptf_mesh_set_vertex_data(ptf_mesh_handle mesh, 
    void const * ptr, ptf_16 vertex_size, ptf_32 vertex_count);

//! Return a pointer to the vertex data as stored in the mesh. This pointer will 
//! be valid until any mutation function is called on the mesh. Writing to file 
//! counts as a mutation function, in addition to all set functions.
//! \param mesh The mesh to get the vertex data from.
//! \param o_ptr The vertex data pointer is returned here.
//! \param o_vertex_size The size of each vertex is returned here.
//! \param o_vertex_count The number of vertices is returned here.
//! \return ptf_err status.
PTF_API ptf_err ptf_mesh_get_vertex_data(ptf_mesh_handle mesh,
    void const ** o_ptr, ptf_16 *o_vertex_size, ptf_32 *o_vertex_count);

//! Set the given mesh to use the given material archive for all the materials associated
//! with the mesh's subsets.
//! \param mesh The mesh to associate the archive with.
//! \param archive Reference to the material archive to use.
//! \return ptf_err status.
PTF_API ptf_err ptf_mesh_set_material_archive(ptf_mesh_handle mesh, 
    ptf_remote_ref archive);

//! Return the given material archive used for all the given mesh's materials.
//! \param mesh The mesh the archive is associated with.
//! \param o_archive The reference to the material archive is returned here.
//! \return ptf_err status.
PTF_API ptf_err ptf_mesh_get_material_archive(ptf_mesh_handle mesh, 
    ptf_remote_ref * o_archive);

//! Add a subset to the mesh. A subset contains a material reference (showing how 
//! to render the subset of the mesh), as well as an index list (showing which 
//! vertices to use to render the mesh). Subsets can be of various kinds: 
//! Point lists, Line lists, Triangle lists, Quad lists, and even N-gon lists. 
//! However, all elements of the subset must use the same number of indices.
//! There can be many subsets of a mesh, and subsets can re-use the same vertices.
//! \param mesh The mesh to add the subset to. The subset will always get a subset 
//! index that is the current number of subsets.
//! \param mesh The mesh to add the subset to.
//! \param desc The description of the subset. Note that min_index and max_index 
//! will be re-calculated by this call (although the desc argument won't be changed).
//! \param ptr A pointer to the index list. The type of the indices is described in 
//! the desc parameter. The number of indices is also found in the desc parameter.
//! \return ptf_err status.
PTF_API ptf_err ptf_mesh_add_mesh_subset(ptf_mesh_handle mesh,
    ptf_mesh_subset_description const &desc, void const *ptr);

//! Given a mesh, count the number of subsets.
//! \param mesh The mesh to count subsets on.
//! \param o_count The number of subsets is returned here.
//! \return ptf_err status.
//! \see ptf_mesh_get_mesh_subset().
PTF_API ptf_err ptf_mesh_count_mesh_subsets(ptf_mesh_handle mesh, 
    ptf_32 *o_count);

//! Given a mesh, return a subset description at a specific index.
//! \param mesh The mesh to return subsets from.
//! \param index The index of the subset, between 0 (inclusive) and 
//! ptf_mesh_count_mesh_subsets() (exclusive).
//! \param o_desc The subset description is returned here.
//! \param o_ptr The subset index list pointer is returned here. The format of 
//! the index list is described in the o_desc parameter.
//! \return ptf_err status.
PTF_API ptf_err ptf_mesh_get_mesh_subset(ptf_mesh_handle mesh, 
    ptf_32 index, ptf_mesh_subset_description *o_desc, void const **o_ptr);

//! Remove a mesh subset. This may re-number subsets unpredictably, so there 
//! is no good way to identify subsets in a mesh after a removal operation.
//! \param mesh The mesh to remove a subset from.
//! \param index The index of the subset to remove, from 0 (inclusive) to the 
//! number of subsets (exclusive).
//! \return ptf_err status.
PTF_API ptf_err ptf_mesh_remove_mesh_subset(ptf_mesh_handle mesh, 
    ptf_32 index);

//! Get information about the vertices for a given mesh subset.
//! \param mesh The mesh containing the vertex data.
//! \param index The index of the subset to get information about.
//! \param o_info The information about the vertices returned here. The 
//! pointer returned is valid until the next mesh mutation.
//! \param o_count The number of information records returned.
//! \return ptf_err status.
PTF_API ptf_err ptf_mesh_get_subset_information(ptf_mesh_handle mesh,
  ptf_32 index, ptf_subset_information const **o_info, ptf_16 *o_count);

//! You can set the mesh subset information. The data you supply will be 
//! copied into the mesh. Note that any operation on the mesh that affects 
//! vertex data, vertex indices, subsets, or vertex components, will 
//! invalidate this data, and make the mesh re-calculate some basic 
//! information (throwing away yours). Thus, call this last before writing 
//! the mesh to file, if you want to make sure it's not lost.
//! \param mesh The mesh to set subset information data for.
//! \param index The subset to set information for.
//! \param info The information for this subset (an array of structs).
//! \param count The number of information records for this subset.
//! \return ptf_err status.
PTF_API ptf_err ptf_mesh_set_subset_information(ptf_mesh_handle mesh,
  ptf_32 index, ptf_subset_information const *info, ptf_16 count);

//! Get the "distance" coefficient for a particular bounding plane.  That is, get the "d"
//! coefficient of the formula ax + by + cz + d = 0 for the requested plane.
//! Each mesh has up to six bounding planes (see "MeshMinPlane" enum).  Each plane is aligned 
//! with some axis, so the plane normals are known implicitly.  The only thing that's not known 
//! is the "d" coefficient (of the formula ax + by + cz + d = 0).
//! \param mesh The mesh to be queried.
//! \param plane Bounding plane identifier ((see "MeshMinPlane" enum)
//! \param d The distance coefficient, in meters.
PTF_API ptf_err ptf_mesh_get_edge_plane(ptf_mesh_handle mesh, ptf_8 plane, double * d);

//! Set the "distance" coefficient for a particular bounding plane.  That is, set the "d"
//! coefficient of the formula ax + by + cz + d = 0 for the requested plane.
//! \param mesh The mesh to be modified.
//! \param plane Bounding plane identifier ((see "MeshMinPlane" enum)
//! \param d The distance coefficient, in meters.
PTF_API ptf_err ptf_mesh_set_edge_plane(ptf_mesh_handle mesh, ptf_8 plane, double d);

//! Count the number of edges in the given edge list
//! This function returns peTypeNotFound if there are no edges on the given boundary.
//! \param mesh The mesh to be queried.
//! \param plane The boundary in which to count the edges.
//! \param o_count The number of edges in the list.
PTF_API ptf_err ptf_mesh_count_edges(ptf_mesh_handle mesh, ptf_8 plane, ptf_32 * o_count);

//! Retrieve the "nth" edge from the given boundary.  Returns peInvalidArguments if the requested
//! boundary is invalid or the plane for that boundary is undefined.  Also returns peInvalidArguments
//! if the requested edge index is out of range.
//! \param mesh The mesh to be queried.
//! \param index The boundary edge list to be queried.
//! \param index The index of the edge within the edge list.
//! \param edge The retrieved edge.
PTF_API ptf_err ptf_mesh_get_edge(ptf_mesh_handle mesh, ptf_8 plane, ptf_32 index, ptf_mesh_edge * edge);

//! Append an edge to the given edge list (as designated by the boundary plane).
//! \param plane The edge list to which the edge will be added.
//! \param 
PTF_API ptf_err ptf_mesh_add_edge(ptf_mesh_handle mesh, ptf_8 plane, ptf_mesh_edge * edge);

//! Meshes may be annotated with miscellaneous well-defined kinds of information.
//! Two such pieces are the LOD table, and the collision acceleration for the mesh.
//! These kinds of information are known as "semantic annocations" (or "semantics" 
//! for short), and you can set them, get them, and remove them as you wish. The 
//! actual type of the data for a given semantic is determined by the record_ref.
//! \param mesh The mesh to ask for a specific semantic.
//! \param semantic The semantic to get for the mesh (for example, pmsLodTable or pmsCollision).
//! \param o_semantic_info The additional semantic info (if any) is returned here.
//! \param o_reference A record_ref to the semantic data is returned here.
//! \return peNameNotFound if semantic is not found; peSuccess if semantic found.
PTF_API ptf_err ptf_mesh_get_semantic(ptf_mesh_handle mesh,
    ptf_32 semantic, ptf_32 * o_semantic_info, ptf_record_ref * o_reference);

//! Set the data reference for a given semantic annotation of a mesh.
//! \param mesh The mesh to set or update the semantic reference for.
//! \param semantic The semantic to update, for example pmsLodTable.
//! \param semantic_info Extra info about the semantic, if defined for that kind of semantic.
//! \param reference The data record reference for the semantic.
//! \see ptf_mesh_get_semantic
PTF_API ptf_err ptf_mesh_set_semantic(ptf_mesh_handle mesh,
    ptf_32 semantic, ptf_32 semantic_info, ptf_record_ref const & reference);

//! Remove a semantic annotation that currently exists on the mesh.
//! \param mesh The mesh to remove the semantic from.
//! \param semantic The semantic identifier to remove, for example pmsLodTable.
//! \return peNameNotFound if the semantic is not on the mesh.
PTF_API ptf_err ptf_mesh_remove_semantic(ptf_mesh_handle mesh, ptf_32 semantic);

//! Count the total number of semantics on the mesh.
//! \param mesh The mesh to count semantics on.
//! \param o_count The number of semantic annotations will be returned here.
PTF_API ptf_err ptf_mesh_count_semantics(ptf_mesh_handle mesh, ptf_32 *o_count);

//! List the semantic annotations that exist for a mesh.
//! \param mesh The mesh to list semantics on.
//! \param o_semantics The identifiers for the semantics on the mesh are returned in this array.
//! \param max_count The maximum number of semantic identifiers to return (the number of elements in the array).
PTF_API ptf_err ptf_mesh_get_semantics(ptf_mesh_handle mesh, ptf_32 *o_semantics, ptf_32 max_count);

#endif  //  ptf_mesh_h
