#if !defined(ptf_util_h)
//! \internal header guard
#define ptf_util_h

//! \addtogroup PtfLibrary The PTF File Library

//  I really don't like these warnings, as they are bogus.
//  Turn them back on when compiling for _WIN64
#if !defined(_WIN64)
#pragma warning(disable: 4996)
#endif

#if !defined(ptf_h)
#include "ptf.h"
#endif
#if !defined(ptf_err_h)
#include "ptf_err.h"
#endif

#include <stdio.h>
#include <string.h>

//! Unpack a "packed" vertex component value to a canonical float.
//! \param type The type, as described in ptf_vertex_component_description.
//! \param count The number of elements in the component. 3 for a vertex normal, for example.
//! \param alignment The alignment value used when describing the vertices. Typically 4.
//! \param input A pointer to a pointer to the vertex data.
//! \param output A pointer to at least "count" floats, which will receive the unpacked data 
//! values.
//! \return ptf_err status.
//! \param ptf_type_unpack() will update the *io_input pointer to point after the read data, 
//! thus this function can be used to "walk" a vertex array and unpack all of the data by 
//! calling it repeatedly.
PTF_API ptf_err ptf_type_unpack(ptf_8 type, ptf_8 count, ptf_16 alignment, void const **io_input, float *o_output);

namespace impl {

#if defined(_WIN32) || defined(_WIN64) || defined(PLATFORM_UNIX)
//! Seek a file to a given 64-bit position.
//! \param stream The file to seek
//! \param offset The offset to seek
//! \param whence From where to seek -- 0 == start, 1 == current, 2 == end
//! \return 0 on success
#if _MSC_VER >= 1400
inline int fseek64(FILE * stream, ptf_64 offset, int whence)
{
  return _fseeki64(stream, offset, whence);
}
#else
inline int fseek64(FILE * stream, ptf_64 offset, int whence)
{
  if ((ptf_64)(ptf_32)offset != offset) {
    return -1;
  }
  return fseek(stream, (ptf_32)offset, whence);
}
#endif

//! Read how far into the file the file pointer is, as a 64-bit value.
//! \param stream The file to check.
//! \return The current offset, as a 64-bit value.
#if _MSC_VER >= 1400
inline ptf_64 ftell64(FILE * stream) {
  return _ftelli64(stream);
}
#else
inline ptf_64 ftell64(FILE * stream) {
  //  Because you can't seek to a position greater than 32 bits, 
  //  no need to check the return value for overflow.
  return (ptf_64)ftell(stream);
}
#endif

//! Write some data to a file, using 64-bit size value, at the current file pointer pos.
//! \param buf The data to write.
//! \param size The size of the data to write.
//! \param file The file to write to.
//! \return A ptf_err result; peSuccess if successfully written.
inline ptf_err fwrite64(void const * buf, ptf_64 size, FILE * file)
{
  if ((ptf_64)(size_t)size != size) {
    return peTooBigForMachine;
  }
  if (fwrite(buf, 1, (size_t)size, file) != size) {
    return peIoError;
  }
  return peSuccess;
}

//! Read some data from a file, using 64-bit size value, at the current file pointer pos.
//! \param buf The buffer to read into.
//! \param size The size of the data to read.
//! \param file The file to read from.
//! \return A ptf_err result; peSuccess if successfully read.
inline ptf_err fread64(void * buf, ptf_64 size, FILE * file)
{
  if ((ptf_64)(size_t)size != size) {
    return peTooBigForMachine;
  }
  if (fread(buf, 1, (size_t)size, file) != size) {
    return peIoError;
  }
  return peSuccess;
}

//! Ensure that data in a file is actually on disk.
//! \param file The file to sync to disk.
//! \return peSuccess if successful.
inline ptf_err fsync64(FILE * file) {
  if (fflush(file)) {
    return peIoError;
  }
  return peSuccess;
}
#endif


//! Remove a pointer qualification from a type.
template<typename Q> struct deref_ptr {
  typedef Q type;
};
template<typename T> struct deref_ptr<T *> {
  typedef T type;
};
template<typename T> struct deref_ptr<T const *> {
  typedef T const type;
};


//! Zero-initialize a member variable.
//! Used to avoid having to initialize members in constructors.
template<typename T> struct zero_t {
  public:
    zero_t() : t_(0) {}
    explicit zero_t(T const & t) : t_(t) {}
    operator T &() { return t_; }
    operator T const &() const { return t_; }
    T operator->() { return t_; }
    T const operator->() const { return t_; }
    typename deref_ptr<T>::type & operator*() { return *t_; }
    typename deref_ptr<T>::type const & operator*() const { return *t_; }
    T * operator &() { return &t_; }
    T const * operator &() const { return &t_; }
    zero_t & operator=(T const & t) { t_ = t; return *this; }
    bool operator!() const { return !t_; }
    T t_;
};

//! Do reads/writes on a block of memory.
//! This is useful when doing marshalling to/from binary 
//! serialized representations, like for material archives.
class PTF_API block_t {
  public:
    //! Create a block; it is initially empty.
    block_t() {}
    //! prealloc can be used to ensure that only a single 
    //! memory allocation happens, if you know how much 
    //! data you will need. size of the block still starts 
    //! out at 0.
    //! \param prealloc Number of bytes to reserve.
    explicit block_t(size_t prealloc);
    //! Destroy the block, and invalidate its pointer/data.
    ~block_t();
    //! Copy a block from another block. Block data is 
    //! reference counted, and thus copying is cheap, whereas 
    //! writing may cause a copy to be created.
    block_t(block_t const &);
    //! Copy a block from another block. Block data is 
    //! reference counted, and thus copying is cheap, whereas 
    //! writing may cause a copy to be created.
    block_t & operator=(block_t const &);
    //! Extract a POD data type from the current position 
    //! within the block.
    //! \param t The data to copy from the stream.
    //! return true If the data could be copied; false if not 
    //! enough data in the stream.
    template<typename T> bool get(T &t)
      {
        return get_bytes(&t, sizeof(t));
      }
    //! Put a copy of a POD data type into the stream.
    //! \param t The data to copy into the stream.
    //! \return true If the data could be inserted; false on 
    //! out-of-memory.
    //! \note putting data may invalidate data pointers.
    template<typename T> bool put(T t)
      {
        return put_bytes(&t, sizeof(t));
      }
    //! Retrieve a string from the stream. The string may 
    //! be at most 'size' bytes in size (but can be smaller).
    //! \param str The buffer to receive the string.
    //! \param size The maximum size of the string.
    //! \return true If the string could be transferred.
    bool get(ptf_char * str, size_t size)
      {
        return str_out(str, size);
      }
    //! Raw read of some number of bytes.
    //! \param out The buffer to receive bytes.
    //! \param count The number of bytes to read -- all or nothing.
    //! \return true If the amount of data could be copied.
    bool get_bytes(void *out, size_t count);
    //! Raw write of some number of bytes.
    //! \param out The buffer to copy bytes from.
    //! \param count The number of bytes to write -- all or nothing.
    //! \return true If the amount of data could be copied.
    bool put_bytes(void const *out, size_t count);
    //! Reposition the read/write pointer within the stream. The 
    //! value must address a position within the stream, or at the 
    //! exact end of the stream.
    //! \param pos Where to position the pointer (based on whence).
    //! \param whence 0 to count pos from start; 1 to count pos from 
    //! current pos; 2 to count pos from end of stream.
    bool seek(size_t pos, int whence);
    //! \return Actual size of written data in stream.
    size_t size() { return size_; }
    //! \return File pointer position.
    size_t pos() { return pos_; }
    //! \return Pointer to first byte of data.
    //! \note This data pointer is invalidated by any write operation.
    void const *data() { return (void*)(block_+1); }
    //! Allocate a buffer at the current file pointer, and return a 
    //! pointer to that buffer. Extend the size of the block if it is 
    //! not already big enough to contain the buffer (e g, when the 
    //! file pointer is not at the end).
    //! Your job is to fill this block with data.
    //! \param size How much space to allocate.
    //! \return The pointer to put data into, or 0 on failure.
    //! This pointer will be invalidated by the next put operation.
    void *extend(size_t size);
  private:
    //! \internal block
    zero_t<int *> block_;
    //! \internal physical size
    zero_t<size_t> phys_;
    //! \internal written size
    zero_t<size_t> size_;
    //! \internal file pointer pos
    zero_t<size_t> pos_;
    //! \internal let go of storage
    void release();
    //! \internal grip tighter on storage
    void hold();
    //! \internal allow modification
    void split();
    //! \internal make sure there's space
    bool reserve(size_t);
    //! \internal copy a string into the stream
    bool str_in(ptf_char const * str);
    //! \internal copy a string out from the stream
    bool str_out(ptf_char * str, size_t size);
    //! Disallow getting and putting through pointer.
    template<typename Q> bool get(Q *&);
    //! Disallow getting and putting through pointer.
    template<typename Q> bool put(Q *&);
};

    //! Put a PTF string into the memory stream.
    //! \param str The string to copy into the stream.
    //! \return true If the data could be inserted; false on 
    //! out-of-memory.
    //! \note putting data may invalidate data pointers.
    template<> inline bool block_t::put(ptf_char const *str)
      {
        return str_in(str);
      }

}       // namespace impl

#endif  //  ptf_util_h
