
#include <string.h>
#include <vector>
#include <limits.h>
#include <float.h>
#include "ptf_elevation.h"
#include "ptf_str.h"

// macros to check return code from PTF API functions
#define PTF_FAILED(a) ((a) != peSuccess)
#define PTF_SUCCEEDED(a) ((a) == peSuccess)

// elevation header flags
#define   NULL_VALUE_MASK   0x01

// lookup table that maps "pvt" types to the number of bytes needed for that type.
// "0" entries are not supported for DEM samples
static int pvtToBytes[] = {
  0,                  // pvtInvalid = 0,
  sizeof( ptf_32 ),   // pvtFloat32 = 1,
  0,                  // pvtUInt32 = 2,
  0,                  // pvtInt32 = 3,
  sizeof( ptf_16 ),   // pvtUInt16 = 4,
  sizeof( ptf_16 ),   // pvtInt16 = 5,
  sizeof( ptf_8 ),    // pvtUInt8 = 6,
  sizeof( ptf_8 ),    // pvtInt8 = 7,
  0                   // pvtMaxType,
};

////////////////////////////////////////////////////
// PTF in-memory representation stuff

class ptf_dem_t {

public:
  // manager functions
  ptf_dem_t() : extentX_( 0. ), extentY_( 0. ), xScale_( 1.0 ), yScale_( 1.0 ), zScale_( 1.0 ), 
    sampleSize_( 0 ), projLon_( 0. ), projLat_( 0. ), zOffset_( 0. ), buffer_( NULL )
  {
    memset( &header_, 0, sizeof( header_ ) );
  }

  // validation
  bool validateRecHeader();

  ~ptf_dem_t() {  }

  // member data
  ptf_elev_header         header_;

  float                   xScale_;         // horizontal scale: east-west (?)
  float                   yScale_;         // horizontal scale: north-south (?)
  float                   zScale_;         // vertical scale: up-down
  
  float                   extentX_;        // dataset half width (distance east/west from origin to edge of data)
  float                   extentY_;        // dataset half height (distance north/south from origin to edge of data)

  float                   projLon_;        // projection origin longitude
  float                   projLat_;        // projection origin latitude

  float                   zOffset_;        // offset applied to elevation data
  int                     sampleSize_;     // sample size in bytes - corresponds to header_.sample_type

  std::vector< char >     buffer_;         // data buffer to store the samples

private:
  // hidden managers
  ptf_dem_t( const ptf_dem_t & obj );
  ptf_dem_t &operator=( const ptf_dem_t & obj );
};

bool ptf_dem_t::validateRecHeader( )
{
  // sample_type - make sure this is one of our supported types
  switch( header_.sample_type ) {
    case pvtFloat32:
    case pvtInt16:
    case pvtInt8:
    case pvtUInt16:
    case pvtUInt8:
      break;
    default:
      return false;
  }

  // offset must be at least past the length of the header
  if ( header_.data_offset < sizeof( header_ ) ) {
    return false;
  }

  // must have at least one row and one column
  if ( !( header_.num_cols && header_.num_rows ) ) {
    return false;
  }

  // length should be: #cols * #rows * #bytes per sample
  ptf_32 expectedLen = header_.num_cols * header_.num_rows * pvtToBytes[ header_.sample_type ];
  if ( header_.data_length != expectedLen ) {
    return false;
  }

  return true;
}

///////////////////////////////////////////////////////////////////////////

static bool copySample( ptf_dem_val *pDst, ptf_dem_val const *pSrc, ptf_8 type )
{
  switch( type ) {
    case pvtFloat32:
      memcpy( &pDst->float32, &pSrc->float32, sizeof( ptf_32 ) );
      break;

    case pvtInt16:
      memcpy( &pDst->int16, &pSrc->int16, sizeof( ptf_16 ) );
      break;

    case pvtInt8:
      pDst->int8 = pSrc->int8;
      break;
    default:
      return false;
  }

  return true;
}

static int compSample( ptf_dem_val * op1, ptf_dem_val * op2, ptf_8 type )
{
  int retval = 0;
  switch( type ) {
    case pvtFloat32:
      if ( op1->float32 == op2->float32 )
        retval = 0;
      else if ( op1->float32 > op2->float32 )
        retval = 1;
      else
        retval = -1;
      break;

    case pvtInt16:
      if ( op1->int16 == op2->int16 )
        retval = 0;
      else if ( op1->int16 > op2->int16 )
        retval = 1;
      else
        retval = -1;
      break;

    case pvtInt8:
      if ( op1->int8 == op2->int8 )
        retval = 0;
      else if ( op1->int8 > op2->int8 )
        retval = 1;
      else
        retval = -1;
      break;

    case pvtUInt16:
      if ( op1->uint16 == op2->uint16 )
        retval = 0;
      else if ( op1->uint16 > op2->uint16 )
        retval = 1;
      else
        retval = -1;
      break;

    case pvtUInt8:
      if ( op1->uint8 == op2->uint8 )
        retval = 0;
      else if ( op1->uint8 > op2->uint8 )
        retval = 1;
      else
        retval = -1;
      break;
  }

  return retval;
}

///////////////////////////////////////////////////////////////////////////

PTF_API ptf_err ptf_dem_create( ptf_dem_handle *o_dem, ptf_char const * name )
{
  try {
    size_t len = strlen( (char const *) name );
    if (len >= sizeof( (*o_dem )->header_.name ) ) {
      return peInvalidArguments;
    }
    *o_dem = new ptf_dem_t();
    memcpy( (*o_dem)->header_.name, name, len + 1);

    return peSuccess;
  }
  catch(...) {
    return peException;
  }
}

PTF_API ptf_err ptf_dem_destroy( ptf_dem_handle dem )
{
  try {
    delete dem;
    return peSuccess;
  }
  catch(...) {
    return peException;
  }
}

////////////////////////////////////////////////////
// PTF I/O stuff

static void saveHeader( ptf_dem_handle dem )
// copy floating point data from dem structure to ptf record header - done just prior to writing the header
{
  // NOTE: relies on the fact that all the floating point entries in the ptf header are contiguous
  memcpy( &dem->header_.x_scale, &dem->xScale_, sizeof( ptf_32 ) * 8 );
}

static void loadHeader( ptf_dem_handle dem )
// copy floating point data from ptf record header to dem structure - done just after reading the header
{
  memcpy( &dem->xScale_, &dem->header_.x_scale, sizeof( ptf_32 ) * 8 );
}

static void postToDisk( ptf_dem_val const & post, ptf_elev_sample_val & disk )
// convert API's post value to disk equivilent
{
  // NOTE: relies on "memory" and "disk" representations being the same size (one has semantics and one does not)
  memcpy( &disk, &post, sizeof( post ) );
}

static void diskToPost( ptf_elev_sample_val const & disk, ptf_dem_val & post )
// convert API's post value to disk equivilent
{
  // NOTE: relies on "memory" and "disk" representations being the same size (one has semantics and one does not)
  memcpy( &post, &disk, sizeof( disk ) );
}

PTF_API ptf_err ptf_dem_read_from_file( ptf_dem_handle dem, ptf_file_handle file, ptf_fixed_ref const &ref )
{
  try {
    ptf_record_ref rref;
    memset( &rref, 0, sizeof( rref ) );
    rref.type = ptElevationData;
    rref.index = ref.index;

    // read the header from the file
    ptf_64 length = 0;
    ptf_err ptf_status;
    if ( PTF_FAILED( ptf_status = ptf_file_read_record( file, &rref, 0, &dem->header_, sizeof( dem->header_ ) ) ) ) {
      return ptf_status;
    }
    // guard against evil doers...
    if ( !dem->validateRecHeader( ) ) {
      return peInvalidFile;
    }

    loadHeader( dem );

    // allocate a buffer from header params
    dem->sampleSize_ = pvtToBytes[ dem->header_.sample_type ];
    dem->buffer_.resize( dem->header_.data_length );

    // read the buffer from the file
    // NOTE: assume we've chopped the data such that no tile is too big to fit into memory
    if ( PTF_FAILED( ptf_status = ptf_file_read_record( file, &rref, dem->header_.data_offset, (void *) &dem->buffer_[ 0 ], dem->header_.data_length ) ) ) {
      return ptf_status;
    }
  }
  catch(...) {
    return peException;
  }
  return peSuccess;
}

PTF_API ptf_err ptf_dem_update_to_file( ptf_dem_handle dem, ptf_file_handle file, ptf_fixed_ref const &ref )
{
  try {
    // set the "current record" in the PTF file
    ptf_record_ref rRef;
    memset( &rRef, 0, sizeof( rRef ) );
    rRef.type = ptElevationData;
    rRef.index = ref.index;

    ptf_err ptf_status;
    if ( PTF_FAILED( ptf_status = ptf_file_replace_record( file, &rRef ) ) ) {
      return ptf_status;
    }

    // write out the header
    dem->header_.data_offset = sizeof( dem->header_ );
    dem->header_.data_length = (ptf_32) dem->buffer_.size();
    saveHeader( dem );
    if ( PTF_FAILED( ptf_status = ptf_file_write_record( file, &dem->header_, sizeof( dem->header_ ) ) ) ) {
      return ptf_status;
    }

    // write out the data
    if ( dem->buffer_.size() > 0 &&
      PTF_FAILED( ptf_status = ptf_file_write_record( file, (void *) &dem->buffer_[ 0 ], dem->header_.data_length ) ) ) 
    {
      return ptf_status;
    }

    // commit record to disk
    ptf_record_ref actualRef;
    memset( &actualRef, 0, sizeof( actualRef ) );
    if ( PTF_FAILED( ptf_status = ptf_file_commit_record( file, &actualRef ) ) ) {
      return ptf_status;
    }

#ifdef _DEBUG
    // verify that the record ref we get back from commit has the expected index
    if ( !( actualRef.type == ptElevationData && actualRef.index == ref.index ) ) {
      return peIoError;
    }
#endif

    return peSuccess;
  }
  catch(...) {
    return peException;
  }
}


PTF_API ptf_err ptf_dem_append_to_file( ptf_dem_handle dem, ptf_file_handle file, ptf_fixed_ref *o_ref )
{
  try {
    // add a new raster record to the PTF file
    ptf_err ptf_status;
    if ( PTF_FAILED( ptf_status = ptf_file_new_record( file, ptElevationData ) ) ) {
      return ptf_status;
    }

    // write out the header
    dem->header_.data_offset = sizeof( dem->header_ );
    dem->header_.data_length = (ptf_32) dem->buffer_.size();
    saveHeader( dem );
    if ( PTF_FAILED( ptf_status = ptf_file_write_record( file, &dem->header_, sizeof( dem->header_ ) ) ) ) {
      return ptf_status;
    }

    // write out the data
    if ( dem->buffer_.size() > 0 &&
      PTF_FAILED( ptf_status = ptf_file_write_record( file, (void *) &dem->buffer_[ 0 ], dem->header_.data_length ) ) ) 
    {
      return ptf_status;
    }

    // commit record to disk
    ptf_record_ref rRef;
    memset( &rRef, 0, sizeof( rRef ) );
    if ( PTF_FAILED( ptf_status = ptf_file_commit_record( file, &rRef ) ) ) {
      return ptf_status;
    }

    if ( o_ref ) {
      o_ref->index = rRef.index;
    }
    return peSuccess;
  }
  catch(...) {
    return peException;
  }
}


////////////////////////////////////////////////////
// header accessor stuff

PTF_API ptf_err ptf_dem_get_name( ptf_dem_handle dem, ptf_name name )
{
  try {
    ptf_strncpy( name, PTF_NAMESIZE, dem->header_.name );
    return peSuccess;
  }
  catch(...) {
    return peException;
  }
}

PTF_API ptf_err ptf_dem_set_col_row_type( ptf_dem_handle dem, ptf_16 cols, ptf_16 rows, ptf_8 type )
{
  try {
    int tmpSize = pvtToBytes[ type ];
    if ( !( tmpSize && rows && cols ) ) {
      return peInvalidArguments;
    }

    dem->header_.num_rows = rows;
    dem->header_.num_cols = cols;

    dem->header_.sample_type = type;
    dem->sampleSize_ = tmpSize;

    // resize the buffer
    dem->buffer_.resize( dem->header_.num_rows * dem->header_.num_cols * dem->sampleSize_ );

    return peSuccess;
  }
  catch(...) {
    return peException;
  }
}

PTF_API ptf_err ptf_dem_get_col_row( ptf_dem_handle dem, ptf_16 * cols, ptf_16 * rows )
{
  try {
    if ( rows ) {
      *rows = dem->header_.num_rows;
    }
    if ( cols ) {
      *cols = dem->header_.num_cols;
    }
    return peSuccess;
  }
  catch(...) {
    return peException;
  }
}

PTF_API ptf_err ptf_dem_set_extents( ptf_dem_handle dem, float half_width, float half_height )
{
  try {
    if ( half_width < 0. || half_height < 0. ) {
      return peInvalidArguments;
    }

    dem->extentX_ = half_width;
    dem->extentY_ = half_height;
    return peSuccess;
  }
  catch(...) {
    return peException;
  }
}

PTF_API ptf_err ptf_dem_get_extents( ptf_dem_handle dem, float * half_width, float * half_height )
{
  try {
    if ( half_width ) {
      *half_width = dem->extentX_;
    }
    if ( half_height ) {
      *half_height = dem->extentY_;
    }
    return peSuccess;
  }
  catch(...) {
    return peException;
  }
}

PTF_API ptf_err ptf_dem_set_proj_org( ptf_dem_handle dem, float lon, float lat )
{
  try {
    dem->projLon_ = lon;
    dem->projLat_ = lat;
    return peSuccess;
  }
  catch(...) {
    return peException;
  }
}

PTF_API ptf_err ptf_dem_get_proj_org( ptf_dem_handle dem, float * lon, float * lat )
{
  try {
    if ( lon ) {
      *lon = dem->projLon_;
    }
    if ( lat ) {
      *lat = dem->projLat_;
    }
    return peSuccess;
  }
  catch(...) {
    return peException;
  }
}

PTF_API ptf_err ptf_dem_get_min_max( ptf_dem_handle dem, ptf_dem_val * min, ptf_dem_val * max )
{
  try {
    size_t bufSize = dem->buffer_.size();
    if ( 0 == bufSize ) {
      return peInvalidArguments;
    }

    ptf_dem_val rmin;
    ptf_dem_val rmax;
    switch( dem->header_.sample_type ) {
      case pvtFloat32:
        rmin.float32 = FLT_MAX;
        rmax.float32 = FLT_MIN;
        break;
      case pvtInt16:
        rmin.int16 = SHRT_MAX;
        rmax.int16 = SHRT_MIN;
        break;
      case pvtInt8:
        rmin.int8 = CHAR_MAX;
        rmax.int8 = CHAR_MIN;
        break;
      case pvtUInt8:
        rmin.uint8 = UCHAR_MAX;
        rmax.uint8 = 0;
      case pvtUInt16:
        rmin.uint16 = USHRT_MAX;
        rmax.uint16 = 0;
    }

    // iterate samples 2..n
    int index = dem->sampleSize_;
    while ( index < bufSize ) {
      // skip NULL samples
      if ( compSample( (ptf_dem_val *) &dem->buffer_[ index ], (ptf_dem_val *) &dem->header_.null_data_val, dem->header_.sample_type ) ) {
        if ( -1 == compSample( (ptf_dem_val *) &dem->buffer_[ index ], &rmin, dem->header_.sample_type ) ) {
          copySample( &rmin, (ptf_dem_val *) &dem->buffer_[index], dem->header_.sample_type );
        }

        if ( 1 == compSample( (ptf_dem_val *) &dem->buffer_[ index ], &rmax, dem->header_.sample_type ) ) {
          copySample( &rmax, (ptf_dem_val *) &dem->buffer_[index], dem->header_.sample_type );
        }
      }
      index += dem->sampleSize_;
    }

    if ( min ) {
      memcpy( min, &rmin, sizeof( ptf_dem_val ) );
    }
    if ( max ) {
      memcpy( max, &rmax, sizeof( ptf_dem_val ) );
    }
    return peSuccess;
  }
  catch(...) {
    return peException;
  }
}

PTF_API ptf_err ptf_dem_get_sample_type( ptf_dem_handle dem, ptf_8 * type )
{
  if ( !type ) {
    return peInvalidArguments;
  }

  try {
    if ( 0 == type ) {
      return peInvalidArguments;
    }

    *type = dem->header_.sample_type;
    return peSuccess;
  }
  catch(...) {
    return peException;
  }
}

PTF_API ptf_err ptf_dem_set_null_data( ptf_dem_handle dem, ptf_dem_val sentinel )
{
  try {
    postToDisk( sentinel, dem->header_.null_data_val );
    dem->header_.flags |= NULL_VALUE_MASK;
    return peSuccess;
  }
  catch(...) {
    return peException;
  }
}

PTF_API ptf_err ptf_dem_get_null_data( ptf_dem_handle dem, bool * result, ptf_dem_val * sentinel )
{
  try {
    if ( 0 == sentinel ) {
      return peInvalidArguments;
    }
    
    if ( dem->header_.flags & NULL_VALUE_MASK ) {
      diskToPost( dem->header_.null_data_val, *sentinel );
      *result = true;
    }
    else {
      *result = false;
    }

    return peSuccess;
  }
  catch(...) {
    return peException;
  }
}

PTF_API ptf_err ptf_dem_set_scale( ptf_dem_handle dem, float x_scale, float y_scale, float z_scale )
{
  try {
    dem->xScale_ = x_scale;
    dem->yScale_ = y_scale;
    dem->zScale_ = z_scale;
    return peSuccess;
  }
  catch(...) {
    return peException;
  }
}

PTF_API ptf_err ptf_dem_get_scale( ptf_dem_handle dem, float * x_scale, float * y_scale, float * z_scale )
{
  try {
    if ( x_scale ) {
      *x_scale = dem->xScale_;
    }
    if ( y_scale ) {
      *y_scale = dem->yScale_;
    }
    if ( z_scale ) {
      *z_scale = dem->zScale_;
    }
    return peSuccess;
  }
  catch(...) {
    return peException;
  }
}

PTF_API ptf_err ptf_dem_set_data_offset( ptf_dem_handle dem, float z_offset )
{
  try {
    dem->zOffset_ = z_offset;
    return peSuccess;
  }
  catch(...) {
    return peException;
  }
}

PTF_API ptf_err ptf_dem_get_data_offset( ptf_dem_handle dem, float * z_offset )
{
  try {
    if ( z_offset ) {
      *z_offset = dem->zOffset_;
    }
    return peSuccess;
  }
  catch(...) {
    return peException;
  }
}

PTF_API ptf_err ptf_dem_set_sample_val( ptf_dem_handle dem, ptf_16 col, ptf_16 row, ptf_dem_val sample )
{
  try {
#ifdef _DEBUG
    // make sure sample address is in bounds
    if ( row >= dem->header_.num_rows || col >= dem->header_.num_cols ) {
       return peInvalidArguments;
    }

    if ( !( dem->header_.num_rows && dem->header_.num_cols && dem->header_.sample_type ) ) {
       return peInvalidArguments;
    }
#endif

    // if we don't have a buffer, then build it
    if ( 0 == dem->buffer_.size() ) {
      dem->buffer_.resize( dem->header_.num_rows * dem->header_.num_cols * dem->sampleSize_ );
    }
    
    int index = ( ( row * dem->header_.num_cols ) + col ) * dem->sampleSize_;
    copySample( (ptf_dem_val *) &dem->buffer_[ index ], &sample, dem->header_.sample_type );
    return peSuccess;
  }
  catch(...) {
    return peException;
  }
}

PTF_API ptf_err ptf_dem_get_sample_val( ptf_dem_handle dem, ptf_16 col, ptf_16 row, ptf_dem_val * sample )
{
  try {
    if ( 0 == sample ) {
      return peInvalidArguments;
    }

    // verify that we've got enough info set to have a buffer (if not, then bail)
    if ( 0 == dem->buffer_.size() ) {
      return peInvalidArguments;
    }

#ifdef _DEBUG
    // make sure sample address is in bounds
    if ( row >= dem->header_.num_rows || col >= dem->header_.num_cols ) {
       return peInvalidArguments;
    }
#endif

    // look up sample in array and set it
    int index = ( ( row * dem->header_.num_cols ) + col ) * dem->sampleSize_;
    copySample( sample, (ptf_dem_val *) &dem->buffer_[ index ], dem->header_.sample_type );
    return peSuccess;
  }
  catch(...) {
    return peException;
  }
}

PTF_API ptf_err ptf_dem_set_sample_buffer( ptf_dem_handle dem, void * buffer, ptf_16 startCol, ptf_16 startRow, 
                                          ptf_16 width, ptf_16 height, ptf_16 buf_stride )
{
  try {
#ifdef _DEBUG
    // make sure sample address is in bounds
    if ( ( startRow + height ) > dem->header_.num_rows || ( startCol + width ) > dem->header_.num_cols ) {
       return peInvalidArguments;
    }

    if ( !( dem->header_.num_rows && dem->header_.num_cols && dem->header_.sample_type ) ) {
       return peInvalidArguments;
    }
#endif

    // if we don't have a buffer, then build it
    if ( 0 == dem->buffer_.size() ) {
      dem->buffer_.resize( dem->header_.num_rows * dem->header_.num_cols * dem->sampleSize_ );
    }
    
    // calc start index of the input buffer within the DEM buffer
    int demIndex = ( ( startRow * dem->header_.num_cols ) + startCol ) * dem->sampleSize_;
    // see if we can copy the buffer in one shot (width == stride of the DEM)
    if ( 0 == startCol && width == dem->header_.num_cols ) {
      int bytesToCopy = width * height * dem->sampleSize_;
      memcpy( &dem->buffer_[ demIndex ], buffer, bytesToCopy );
    }
    else {
      // need to copy one row at a time
      int bytesToCopy = width * dem->sampleSize_;
      int bufIndex = 0;
      int demStride = dem->header_.num_cols * dem->sampleSize_;
      char *buf = (char *) buffer;
      for ( int r = 0; r < height; r++ ) {
        memcpy( &dem->buffer_[ demIndex ], &buf[ bufIndex ], bytesToCopy );
        bufIndex += buf_stride;
        demIndex += demStride;
      }
    }

    return peSuccess;
  }
  catch(...) {
    return peException;
  }
}

PTF_API ptf_err ptf_dem_get_sample_buffer( ptf_dem_handle dem, void * buffer, ptf_16 startCol, ptf_16 startRow, 
                                          ptf_16 width, ptf_16 height, ptf_16 buf_stride )
{
  try {
#ifdef _DEBUG
    if ( !dem ) {
       return peInvalidArguments;
    }

    // make sure sample address is in bounds
    if ( ( startRow + height ) > dem->header_.num_rows || ( startCol + width ) > dem->header_.num_cols ) {
       return peInvalidArguments;
    }

    if ( !( dem->header_.num_rows && dem->header_.num_cols && dem->header_.sample_type ) ) {
       return peInvalidArguments;
    }
#endif

    // if we don't have a buffer, then bail
    if ( 0 == dem->buffer_.size() ) {
      return peInvalidArguments;
   }
    
    // calc start index of the input buffer within the DEM buffer
    int demIndex = ( ( startRow * dem->header_.num_cols ) + startCol ) * dem->sampleSize_;
    // see if we can copy the buffer in one shot (width == stride of the DEM)
    if ( 0 == startCol && width == dem->header_.num_cols ) {
      int bytesToCopy = width * height * dem->sampleSize_;
      memcpy( buffer, &dem->buffer_[ demIndex ], bytesToCopy );
    }
    else {
      // need to copy one row at a time
      int bytesToCopy = width * dem->sampleSize_;
      int bufIndex = 0;
      int demStride = dem->header_.num_cols * dem->sampleSize_;
      char *buf = (char *) buffer;
      for ( int r = 0; r < height; r++ ) {
        memcpy( &buf[ bufIndex ], &dem->buffer_[ demIndex ], bytesToCopy );
        bufIndex += buf_stride;
        demIndex += demStride;
      }
    }

    return peSuccess;
  }
  catch(...) {
    return peException;
  }
}
