
#include "ptf_mesh.h"
#include "ptf_octree.h"

#include <string.h>
#include <assert.h>
#include <vector>
#include <deque>
#include <stdexcept>


struct ptf_vertex_type_desc {
  size_t size;
};
static ptf_vertex_type_desc type_descs[] = {
  { 0 }, { 4 }, { 4 }, { 4 }, { 2 }, { 2 }, { 1 }, { 1 },
};

static ptf_vertex_type_desc const *get_component_desc(ptf_8 type)
{
  if (type >= sizeof(type_descs)/sizeof(type_descs[0])) {
    return 0;
  }
  return &type_descs[type];
}

//  Some limitations imposed by the implementation.
enum {
  MaxVertexComponentCount = 0x80,
  MaxVertexStride = 0x400,
  MaxSubsetCount = 0x1000,
  MaxVertexCount = 0x200000,
  MaxIndexCount = 0x200000,
};

// plane "enum" for undefined plane.  for internal use only.
#define  MeshUndefinedPlane 255


class ptf_edge_t {
public:
  ptf_edge_t()
  {
    memset( &edge, 0, sizeof(ptf_mesh_edge) );
  }

  ptf_edge_t( ptf_mesh_edge const & obj ) : edge( obj )
  {
  }

  ptf_edge_t & operator=( ptf_edge_t const & obj )
  {
    memcpy( &edge, &obj.edge, sizeof(ptf_mesh_edge) );
    return *this;
  }

  ptf_edge_t( ptf_edge_t const & obj )
  {
    *this = obj;
  }

  // member data
  ptf_mesh_edge           edge;
};


class ptf_edge_list_t {
public:

  ptf_edge_list_t( ptf_8 plane = MeshUndefinedPlane ) : plane_index( plane ), plane_d( 0 ) { }

  ptf_edge_list_t & operator=( ptf_edge_list_t const & obj )
  {
    plane_d = obj.plane_d;
    plane_index = obj.plane_index;
    edges.insert( edges.end(), obj.edges.begin(), obj.edges.end() );

    return (*this);
  }

  ptf_edge_list_t( ptf_edge_list_t const & obj )
  {
    *this = obj;
  }

  // member data
  ptf_8                         plane_index;
  ptf_32                        plane_d;        // plane's "d" coefficient - normal is implicit in plane_index
  std::vector<ptf_edge_t>       edges;
};


class ptf_mesh_t {
  public:
    ptf_32                                        sig;
    std::vector<ptf_vertex_component_description> components;
    std::vector<char>                             vertices;
    std::vector<ptf_mesh_subset_description>      subsets;
    std::deque<std::vector<char> >                indices;
    std::vector<ptf_semantic_ref>                 semantics_;
    ptf_mesh_header                               header;
    std::vector<ptf_edge_list_t>                  edge_lists;

    //  clear_subset_information() clears the following three
    std::vector<ptf_32>                           subset_info_index;
    std::vector<char>                             vertex_info_data;
    std::vector<ptf_subset_information_record>    vertex_info;
    std::vector<ptf_subset_information>           returned_vertex_info;

    ptf_mesh_t()
      {
        memset(&header, 0, sizeof(header));
        sig = 'msh ';
        edge_lists.resize( MeshMaxPlane );
      }
    ~ptf_mesh_t()
      {
        sig = 0;
      }

    bool validate() { return sig == 'msh '; }
    void clear()
      {
        this->~ptf_mesh_t();
        new(this) ptf_mesh_t();
      }

    void calc_subset_information();
    void clear_subset_information();
    void calc_returned_subset_information();
    void compact_subset_information();
    ptf_err get_subset_information(ptf_32 index, ptf_subset_information const **o_info, ptf_16 *o_count);
    ptf_err set_subset_information(ptf_32 index, ptf_subset_information const *info, ptf_16 count);

    template<typename T> void accumulate(char const *src, char *dst, T &t);
};

ptf_err ptf_mesh_create(ptf_mesh_handle *o_mesh, ptf_char const *name)
{
  try {
    size_t len = strlen((char const *)name);
    if (len >= sizeof((*o_mesh)->header.name)) {
      return peInvalidArguments;
    }
    *o_mesh = new ptf_mesh_t();
    memcpy((*o_mesh)->header.name, name, len+1);
    return peSuccess;
  }
  catch(...) {
    return peException;
  }
}

ptf_err ptf_mesh_destroy(ptf_mesh_handle mesh)
{
  try {
    if (!mesh->validate()) return peInvalidArguments;
    delete mesh;
    return peSuccess;
  }
  catch(...) {
    return peException;
  }
}

ptf_err ptf_mesh_read_header_from_file(ptf_mesh_header *o_hdr, 
    ptf_file_handle file, ptf_fixed_ref const &fref)
{
  try {
    ptf_record_ref ref;
    memset(&ref, 0, sizeof(ref));
    ref.type = ptMeshData;
    ref.index = fref.index;

    //  read header and verify values
    ptf_err err = ptf_file_read_record(file, &ref, 0, o_hdr, sizeof(*o_hdr));
    ptf_32 datapos = o_hdr->data_offset;
    assert(!(datapos & (o_hdr->file_alignment-1)));
    if (err == peSuccess) {
      if (o_hdr->data_offset == 0 || o_hdr->file_alignment == 0) {
        return peInvalidFile;
      }
      if (o_hdr->file_alignment & (o_hdr->file_alignment - 1)) {
        return peInvalidFile;
      }
      if (o_hdr->data_offset & (o_hdr->file_alignment-1)) {
        return peInvalidFile;
      }
      if (o_hdr->vertex_count < 1 || o_hdr->vertex_count > MaxVertexCount) {
        return peInvalidFile;
      }
      if (o_hdr->vertex_stride < 1 || o_hdr->vertex_stride > MaxVertexStride) {
        return peInvalidFile;
      }

      if (o_hdr->num_components == 0) {
        return peInvalidFile;
      }
      if (o_hdr->num_components >= MaxVertexComponentCount) {
        return peInvalidFile;
      }
      if (o_hdr->num_subsets == 0) {
        return peInvalidFile;
      }
      if (o_hdr->num_subsets > MaxSubsetCount) {
        return peInvalidFile;
      }
      if (o_hdr->num_semantics > 0x7fff) {
        return peInvalidFile;
      }
    }

    return err;
  }
  catch(...) {
    return peException;
  }
}

ptf_err ptf_mesh_read_from_file(ptf_mesh_handle mesh, 
    ptf_file_handle file, ptf_fixed_ref const &fref)
{
  try {
    if (!mesh->validate()) return peInvalidArguments;
    ptf_record_ref ref;
    memset(&ref, 0, sizeof(ref));
    ref.type = ptMeshData;
    ref.index = fref.index;

    //  figure out record size
    ptf_64 record_size = 0;
    ptf_err err = ptf_file_get_record_size(file, &ref, &record_size);
    if (err != peSuccess) return err;
    mesh->clear_subset_information();

    //  read header and verify values
    ptf_mesh_header hdr;
    err = ptf_mesh_read_header_from_file(&hdr, file, fref);
    ptf_32 cnt = 0;
    ptf_32 datapos = hdr.data_offset;
    assert(!(datapos & (hdr.file_alignment-1)));
    if (err == peSuccess) {
      err = ptf_mesh_count_mesh_subsets(mesh, &cnt);
    }
    if (err == peSuccess) {
      //  prepare mesh data vectors
      mesh->clear();
      mesh->components.resize(hdr.num_components);
      mesh->subsets.resize(hdr.num_subsets);
      mesh->indices.resize(hdr.num_subsets);
      memcpy(&mesh->header, &hdr, sizeof(hdr));

      //  read semantics
      for (ptf_32 i = 0; i < hdr.num_semantics && err == peSuccess; ++i) {
        ptf_semantic_ref sref;
        err = ptf_file_read_record(file, &ref, datapos, &sref, sizeof(sref));
        if (err == peSuccess) {
          datapos += sizeof(sref);
          mesh->semantics_.push_back(sref);
        }
        else {
          break;
        }
      }
      assert(!(datapos & (hdr.file_alignment-1)));
    }
    if (err == peSuccess) {
      //  read vertex component descriptions
      err = ptf_file_read_record(file, &ref, datapos, &mesh->components[0], 
          sizeof(ptf_vertex_component_description) * hdr.num_components);
    }
    //  read subsets
    if (err == peSuccess) {
      datapos = (datapos + sizeof(ptf_vertex_component_description) 
              * hdr.num_components + hdr.file_alignment - 1)
          & -hdr.file_alignment;
      err = ptf_file_read_record(file, &ref, datapos, &mesh->subsets[0], 
          sizeof(ptf_mesh_subset_description) * hdr.num_subsets);
    }

    //  read vertex data
    if (err == peSuccess) {
      datapos = (datapos + sizeof(ptf_mesh_subset_description) * hdr.num_subsets + 
              hdr.file_alignment - 1) & -hdr.file_alignment;
      mesh->vertices.resize(hdr.vertex_stride * hdr.vertex_count);
      err = ptf_file_read_record(file, &ref, datapos, &mesh->vertices[0], 
          hdr.vertex_count * hdr.vertex_stride);
      datapos = (datapos + hdr.vertex_count * hdr.vertex_stride);
      //  because vertex stride is already aligned
      assert(!(datapos & (hdr.file_alignment-1)));
    }
    if (err == peSuccess) {
      //  read subset index arrays
      for (size_t i = 0; i < mesh->subsets.size(); ++i) {
        size_t iSize = 0;
        switch (mesh->subsets[i].index_type) {
          case pvtUInt32:
            iSize = 4;
            break;
          case pvtUInt16:
            iSize = 2;
            break;
          default:
            return peInvalidFile;
        }
        if (mesh->subsets[i].primitive_type == 0) {
          return peInvalidFile;
        }
        if (mesh->subsets[i].num_indices < 1 ||
            ((mesh->subsets[i].num_indices / mesh->subsets[i].primitive_type) * 
              mesh->subsets[i].primitive_type) != mesh->subsets[i].num_indices) {
          return peInvalidFile;
        }
        mesh->indices[i].resize(iSize * mesh->subsets[i].num_indices);
        err = ptf_file_read_record(file, &ref, datapos, &mesh->indices[i][0],
            iSize * mesh->subsets[i].num_indices);
        datapos += (ptf_32)(iSize * mesh->subsets[i].num_indices + hdr.file_alignment - 1)
            & -hdr.file_alignment;
        assert(!(datapos & (hdr.file_alignment-1)));
      }
    }

    //   Read the edge list descriptions
    if ( hdr.num_edge_lists ) {
      std::vector<ptf_mesh_edge_list_description> raw_lists;
      raw_lists.resize( hdr.num_edge_lists );
      err = ptf_file_read_record(file, &ref, datapos, &raw_lists[0], 
          sizeof( ptf_mesh_edge_list_description ) * hdr.num_edge_lists );
      datapos += (sizeof( ptf_mesh_edge_list_description ) * hdr.num_edge_lists + hdr.file_alignment - 1)
          & -hdr.file_alignment;

      // read the edges for each list
      for ( int i = 0; i < hdr.num_edge_lists; i++ ) {
        ptf_8 plane = raw_lists[i].plane_index;
        mesh->edge_lists[plane].plane_index = plane;
        mesh->edge_lists[plane].plane_d = raw_lists[i].plane_d;

        std::vector<ptf_mesh_edge> raw_edges;
        raw_edges.resize( raw_lists[i].num_edges );
        err = ptf_file_read_record(file, &ref, datapos, &raw_edges[0], sizeof(ptf_mesh_edge) * 
            raw_lists[i].num_edges );
        datapos += (sizeof( ptf_mesh_edge ) * raw_lists[i].num_edges + hdr.file_alignment - 1)
            & -hdr.file_alignment;

        // copy the raw edges
        mesh->edge_lists[plane].edges.resize( raw_lists[i].num_edges );
        for ( int j = 0; j < raw_lists[i].num_edges; j++ ) {
          mesh->edge_lists[plane].edges[j].edge = raw_edges[j];
        }
      }
    }

    if (hdr.num_subsets) {
      //  Compatibility with draft spec 1: if there is no subset information 
      //  data written, then re-generate it instead of reading it.
      if (datapos + sizeof(ptf_subset_information_header) < record_size) {
        //  do I have the data? Then demarshal it
        ptf_subset_information_header psih;
        err = ptf_file_read_record(file, &ref, datapos, &psih, sizeof(psih));
        if (err != peSuccess) return err;
        datapos += (sizeof(ptf_subset_information_header) + hdr.file_alignment - 1)
            & -hdr.file_alignment;
        mesh->subset_info_index.resize(hdr.num_subsets);
        mesh->vertex_info.resize(psih.num_records);
        mesh->vertex_info_data.resize(psih.data_size);
        err = ptf_file_read_record(file, &ref, datapos, &mesh->subset_info_index[0], 
            sizeof(ptf_32) * mesh->subset_info_index.size());
        if (err != peSuccess) return err;
        datapos += (ptf_32)((sizeof(size_t) * mesh->subset_info_index.size() + hdr.file_alignment - 1)
            & -hdr.file_alignment);
        if (mesh->vertex_info.size()) {
          err = ptf_file_read_record(file, &ref, datapos, &mesh->vertex_info[0], 
              sizeof(ptf_subset_information_record) * mesh->vertex_info.size());
          if (err != peSuccess) return err;
          datapos += (ptf_32)((sizeof(ptf_subset_information_record) * mesh->vertex_info.size() + hdr.file_alignment - 1)
              & -hdr.file_alignment);
        }
        if (mesh->vertex_info_data.size() > 0) {
          err = ptf_file_read_record(file, &ref, datapos, &mesh->vertex_info_data[0],
              mesh->vertex_info_data.size());
          if (err != peSuccess) return err;
          datapos += (ptf_32)((mesh->vertex_info_data.size() + hdr.file_alignment - 1)
              & -hdr.file_alignment);
        }
      }
      else {
        //  else calculate the information
        mesh->calc_subset_information();
      }
    }

    return err;
  }
  catch(...) {
    return peException;
  }
}

ptf_err ptf_mesh_check_consistency(ptf_mesh_handle mesh)
{
  try {
    if (!mesh->validate()) return peInvalidArguments;
    if (mesh->header.file_alignment == 0) {
      mesh->header.file_alignment = 1;
    }
    //  check that the mesh isn't empty
    if (mesh->vertices.size() == 0) return peInvalidArguments;
    if (mesh->components.size() == 0) return peInvalidArguments;
    if (mesh->indices.size() == 0) return peInvalidArguments;
    if (mesh->subsets.size() == 0) return peInvalidArguments;

    //  check some basic properties of the vertex components
    size_t mask = -mesh->header.file_alignment;
    size_t add = mesh->header.file_alignment - 1;
    if (mesh->header.vertex_stride & add) {
      return peInvalidArguments;
    }
    size_t stride = 0;
    for (size_t i = 0; i < mesh->components.size(); ++i) {
      ptf_vertex_type_desc const *vtd = get_component_desc(mesh->components[i].type);
      if (!vtd || vtd->size == 0) {
        return peInvalidArguments;
      }
      if (mesh->components[i].type_count < 1 || mesh->components[i].type_count > 4) {
        return peInvalidArguments;
      }
      stride += (mesh->components[i].type_count * vtd->size + add) & mask;
    }

    //  check the vertex array
    if (stride > mesh->header.vertex_stride) {
      return peInvalidArguments;
    }
    if (mesh->header.vertex_stride * mesh->header.vertex_count != mesh->vertices.size()) {
      return peInvalidArguments;
    }

    // check the subsets
    if (mesh->subsets.size() != mesh->indices.size()) {
      return peInvalidArguments;
    }
    for (size_t i = 0; i < mesh->subsets.size(); ++i) {
      if (mesh->subsets[i].min_index > mesh->subsets[i].max_index) {
        return peInvalidArguments;
      }
      if (mesh->subsets[i].max_index >= mesh->header.vertex_count) {
        return peInvalidArguments;
      }
      if (mesh->subsets[i].num_indices != 
          (mesh->subsets[i].num_indices / mesh->subsets[i].primitive_type) * mesh->subsets[i].primitive_type) {
        return peInvalidArguments;
      }
      if (!mesh->subsets[i].num_indices) {
        return peInvalidArguments;
      }
      size_t iSize = 0;
      switch (mesh->subsets[i].index_type) {
        case pvtUInt16:
          iSize = 2;
          break;
        case pvtUInt32:
          iSize = 4;
          break;
      }
      if (iSize == 0) {
        return peInvalidArguments;
      }
      if (mesh->subsets[i].num_indices * iSize != mesh->indices[i].size()) {
        return peInvalidArguments;
      }
      //  It would probably be faster to hoist the vector indexors 
      //  out to before the loop, and the size if() as well.
      for (size_t j = 0; j < mesh->subsets[i].num_indices; ++j) {
        if (iSize == 2) {
          ptf_16 ix = *(ptf_16 *)&mesh->indices[i][j * iSize];
          if (ix < mesh->subsets[i].min_index) {
            return peInvalidArguments;
          }
          if (ix > mesh->subsets[i].max_index) {
            return peInvalidArguments;
          }
        }
        else {
          ptf_32 ix = *(ptf_32 *)&mesh->indices[i][j * iSize];
          if (ix < mesh->subsets[i].min_index) {
            return peInvalidArguments;
          }
          if (ix > mesh->subsets[i].max_index) {
            return peInvalidArguments;
          }
        }
      }
    }
    return peSuccess;
  }
  catch(...) {
    return peException;
  }
}

static ptf_err pad_file(ptf_file_handle file, void const * pad, ptf_32 written, ptf_32 alignment)
{
  if (!(written & (alignment-1))) {
    return peSuccess;
  }
  ptf_32 towrite = alignment - (written & (alignment-1));
  return ptf_file_write_record(file, pad, towrite);
}

static ptf_err write_to_current_record(
    ptf_file_handle file, ptf_mesh_handle mesh)
{
  std::vector<char> pad;
  pad.resize(mesh->header.file_alignment);
  if (pad.size() > 0) {
    memset(&pad[0], 0, mesh->header.file_alignment);
  }

  //  setup header
  mesh->header.spare1 = 0;
  mesh->header.num_components = (ptf_8)mesh->components.size();
  mesh->header.num_subsets = (ptf_16)mesh->subsets.size();
  mesh->header.num_semantics = (ptf_16)mesh->semantics_.size();
  mesh->header.data_offset = (sizeof(mesh->header) + mesh->header.file_alignment-1)
      & -mesh->header.file_alignment;
  // count the number of defined edge lists
  mesh->header.num_edge_lists = 0;
  for ( int i = MeshMinPlane; i < MeshMaxPlane; i++ ) {
    if ( mesh->edge_lists[i].plane_index != MeshUndefinedPlane ) {
      mesh->header.num_edge_lists++;
    }
  }

  //  write and pad header
  ptf_err err = ptf_file_write_record(file, &mesh->header, sizeof(mesh->header));
  if (err != peSuccess) return err;
  if (pad.size() > 0) {
    err = pad_file(file, &pad[0], sizeof(mesh->header), mesh->header.file_alignment);
    if (err != peSuccess) return err;
  }
  //  write and pad semantics_
  if (mesh->semantics_.size() > 0) {
    err = ptf_file_write_record(file, &mesh->semantics_[0], (ptf_32)
      sizeof(mesh->semantics_[0])*mesh->header.num_semantics);
    if (err != peSuccess) return err;
    if (pad.size() > 0) {
      err = pad_file(file, &pad[0], sizeof(mesh->semantics_[0])*mesh->header.num_semantics, 
        mesh->header.file_alignment);
      if (err != peSuccess) return err;
    }
  }
  //  write and pad components
  if (mesh->components.size() > 0) {
    err = ptf_file_write_record(file, &mesh->components[0], (ptf_32)
      sizeof(mesh->components[0])*mesh->header.num_components);
    if (err != peSuccess) return err;
    if (pad.size() > 0) {
      err = pad_file(file, &pad[0], sizeof(mesh->components[0])*mesh->header.num_components, 
        mesh->header.file_alignment);
      if (err != peSuccess) return err;
    }
  }
  //  write and pad subsets
  if (mesh->subsets.size() > 0) {
    err = ptf_file_write_record(file, &mesh->subsets[0], (ptf_32)
      sizeof(mesh->subsets[0])*mesh->header.num_subsets);
    if (err != peSuccess) return err;
    if (pad.size() > 0) {
      err = pad_file(file, &pad[0], sizeof(mesh->subsets[0])*mesh->header.num_subsets, 
        mesh->header.file_alignment);
      if (err != peSuccess) return err;
    }
  }
  //  write and pad vertex data
  if (mesh->vertices.size() > 0) {
    err = ptf_file_write_record(file, &mesh->vertices[0], (ptf_32)
      mesh->header.vertex_stride * mesh->header.vertex_count);
    if (err != peSuccess) return err;
    if (pad.size() > 0) {
      err = pad_file(file, &pad[0], mesh->header.vertex_stride * mesh->header.vertex_count, 
        mesh->header.file_alignment);
      if (err != peSuccess) return err;
    }
  }
  //  write and pad index data
  for (size_t i = 0; i < mesh->subsets.size(); ++i) {
    ptf_32 iSize = (mesh->subsets[i].index_type <= pvtInt32 ? 4 : 2);
    err = ptf_file_write_record(file, &mesh->indices[i][0], (ptf_32)
        mesh->subsets[i].num_indices * iSize);
    if (err != peSuccess) return err;
    if (pad.size() > 0) {
      err = pad_file(file, &pad[0], mesh->subsets[i].num_indices * iSize, 
        mesh->header.file_alignment);
      if (err != peSuccess) return err;
    }
  }

  // write out edge lists (if any)
  if ( mesh->header.num_edge_lists ) {
    // compose "raw" edge list descriptors and write them out
    for ( int i = MeshMinPlane; i < MeshMaxPlane; i++ ) {
      if ( mesh->edge_lists[i].plane_index != MeshUndefinedPlane ) {
        ptf_mesh_edge_list_description edge_descr;
        memset( &edge_descr, 0, sizeof(edge_descr) );
        edge_descr.plane_index = mesh->edge_lists[i].plane_index;
        edge_descr.plane_d = mesh->edge_lists[i].plane_d;
        edge_descr.num_edges = (ptf_32) mesh->edge_lists[i].edges.size();

        err = ptf_file_write_record(file, &edge_descr, (ptf_32) sizeof(ptf_mesh_edge_list_description) );
        if (err != peSuccess) return err;
      }
    }
    if (pad.size() > 0) {
      err = pad_file(file, &pad[0], (ptf_32) mesh->edge_lists.size() * sizeof(ptf_mesh_edge_list_description), 
        mesh->header.file_alignment);
      if (err != peSuccess) return err;
    }
    // write all the edge lists
    ptf_32 edge_bytes = 0;
    for ( int i = MeshMinPlane; i < MeshMaxPlane; i++ ) {
      if ( mesh->edge_lists[i].plane_index != MeshUndefinedPlane ) {
        ptf_32 num_edges = (ptf_32) mesh->edge_lists[i].edges.size();
        std::vector<ptf_mesh_edge> raw_edges;
        raw_edges.reserve( num_edges );
        for ( int j = 0; j < num_edges; j++ ) {
          raw_edges.push_back( mesh->edge_lists[i].edges[j].edge );
        }

        err = ptf_file_write_record(file, &raw_edges[0], (ptf_32) sizeof(ptf_mesh_edge) * num_edges );
        if (err != peSuccess) return err;
        edge_bytes += (ptf_32) sizeof(ptf_mesh_edge) * num_edges;
      }
    }
    if (pad.size() > 0) {
      err = pad_file(file, &pad[0], edge_bytes, mesh->header.file_alignment);
      if (err != peSuccess) return err;
    }
  }

  //  write and pad subset information
  if (mesh->subset_info_index.size() != mesh->subsets.size()) {
    mesh->clear_subset_information();
    mesh->calc_subset_information();
  }
  else {
    mesh->compact_subset_information();
  }
  //  write header
  ptf_subset_information_header psih;
  memset(&psih, 0, sizeof(psih));
  psih.num_records = (ptf_32)mesh->vertex_info.size();
  psih.data_size = (ptf_32)mesh->vertex_info_data.size();
  err = ptf_file_write_record(file, &psih, (ptf_32)sizeof(psih));
  if (err != peSuccess) return err;
  //  pad
  if (pad.size() > 0) {
    err = pad_file(file, &pad[0], (ptf_32)sizeof(psih), 
      mesh->header.file_alignment);
    if (err != peSuccess) return err;
  }
  if (mesh->subset_info_index.size() > 0) {
    //  write index array
    err = ptf_file_write_record(file, &mesh->subset_info_index[0], 
        (ptf_32)(sizeof(ptf_32) * mesh->subsets.size()));
    if (err != peSuccess) return err;
    //  pad
    if (pad.size() > 0) {
      err = pad_file(file, &pad[0], (ptf_32)(sizeof(ptf_32) * mesh->subsets.size()), 
        mesh->header.file_alignment);
      if (err != peSuccess) return err;
    }
  }
  if (mesh->vertex_info.size() > 0) {
    //  write info array
    err = ptf_file_write_record(file, &mesh->vertex_info[0], (ptf_32)
        (sizeof(ptf_subset_information_record) * mesh->vertex_info.size()));
    if (err != peSuccess) return err;
    //  pad
    if (pad.size() > 0) {
      err = pad_file(file, &pad[0], (ptf_32)
        (sizeof(ptf_subset_information_record) * mesh->vertex_info.size()), 
        mesh->header.file_alignment);
      if (err != peSuccess) return err;
    }
  }
  if (mesh->vertex_info_data.size() > 0) {
    //  write data array
    err = ptf_file_write_record(file, &mesh->vertex_info_data[0],
        (ptf_32)mesh->vertex_info_data.size());
    if (err != peSuccess) return err;
    //  pad
    if (pad.size() > 0) {
      err = pad_file(file, &pad[0], (ptf_32)mesh->vertex_info_data.size(), 
        mesh->header.file_alignment);
      if (err != peSuccess) return err;
    }
  }

  //  declare victory!
  return peSuccess;
}

ptf_err ptf_mesh_update_to_file(ptf_mesh_handle mesh, 
    ptf_file_handle file, ptf_fixed_ref const &ref)
{
  try {
    if (!mesh->validate()) return peInvalidArguments;
    ptf_record_ref rref;
    memset(&rref, 0, sizeof(ref));
    rref.type = ptMeshData;
    rref.index = ref.index;
    ptf_err err = ptf_mesh_check_consistency(mesh);
    if (err != peSuccess) return err;
    err = ptf_file_replace_record(file, &rref);
    if (err != peSuccess) return err;
    err = write_to_current_record(file, mesh);
    if (err != peSuccess) return err;
    err = ptf_file_commit_record(file, &rref);
    return err;
  }
  catch(...) {
    return peException;
  }
}

ptf_err ptf_mesh_append_to_file(ptf_mesh_handle mesh, 
    ptf_file_handle file, ptf_fixed_ref *o_ref)
{
  try {
    if (!mesh->validate()) return peInvalidArguments;
    ptf_record_ref rref;
    memset(&rref, 0, sizeof(rref));
    ptf_err err = ptf_mesh_check_consistency(mesh);
    if (err != peSuccess) return err;
    err = ptf_file_new_record(file, ptMeshData);
    if (err != peSuccess) return err;
    err = write_to_current_record(file, mesh);
    if (err != peSuccess) return err;
    err = ptf_file_commit_record(file, &rref);
    o_ref->index = rref.index;
    return err;
  }
  catch(...) {
    return peException;
  }
}

ptf_err ptf_mesh_get_name(ptf_mesh_handle mesh, ptf_char const **o_name)
{
  try {
    if (!mesh->validate()) return peInvalidArguments;
    *o_name = mesh->header.name;
    return peSuccess;
  }
  catch(...) {
    return peException;
  }
}

ptf_err ptf_mesh_set_name(ptf_mesh_handle mesh, ptf_char const *name)
{
  try {
    if (!mesh->validate()) return peInvalidArguments;
    if (strlen((char const *)name) >= sizeof(mesh->header.name))
      return peInvalidArguments;
    memcpy(mesh->header.name, name, strlen((char const *)name)+1);
    return peSuccess;
  }
  catch(...) {
    return peException;
  }
}

ptf_err ptf_mesh_get_lod_table(ptf_mesh_handle mesh,
    ptf_fixed_ref *o_ref, float *o_feature)
{
  try {
    if (!mesh->validate()) return peInvalidArguments;
    ptf_32 info;
    ptf_record_ref rref;
    ptf_err err = ptf_mesh_get_semantic(mesh, pmsLodTable, &info, &rref);
    if (err != peSuccess) rref.index = 0;
    o_ref->index = rref.index;
    *o_feature = mesh->header.feature_size;
    return peSuccess;
  }
  catch (...) {
    return peException;
  }
}

ptf_err ptf_mesh_set_lod_table(ptf_mesh_handle mesh,
    ptf_fixed_ref lod, float feature)
{
  try {
    if (!mesh->validate()) return peInvalidArguments;
    if( lod.index != 0 ) {
      ptf_record_ref rref;
      rref.file = 0;
      rref.index = lod.index;
      rref.type = ptLodData;
      ptf_err err = ptf_mesh_set_semantic(mesh, pmsLodTable, 0, rref);
      if (err != peSuccess) return err;
    }
    mesh->header.feature_size = feature;
    return peSuccess;
  }
  catch (...) {
    return peException;
  }
}

ptf_err ptf_mesh_set_vertex_components(ptf_mesh_handle mesh, 
    ptf_vertex_component_description const *comp, ptf_16 count, ptf_16 alignment)
{
  try {
    if (!mesh->validate()) return peInvalidArguments;
    if (!alignment || (alignment & (alignment - 1)) || (alignment & 3)) {
      return peInvalidArguments;
    }
    //  Impose some arbitrary upper limit on number of components
    if (count >= MaxVertexComponentCount) {
      return peTooBigForMachine;
    }
    mesh->clear_subset_information();
    mesh->components.resize(count);
    for (size_t i = 0; i < count; ++i) {
      if (comp[i].type >= pvtMaxType || comp[i].type == pvtInvalid) {
        return peInvalidArguments;
      }
      if (comp[i].usage >= pvuMaxUsage) {
        return peInvalidArguments;
      }
      if (comp[i].type_count > 4 || comp[i].type_count == 0) {
        return peInvalidArguments;
      }
      if (comp[i].usage_index > 8) {
        return peInvalidArguments;
      }
      if (comp[i].usage == pvuNormal && comp[i].type_count != 3) {
        return peInvalidArguments;
      }
      //  alpha, BGR, and BGRA are all approved color kinds
      if (comp[i].usage == pvuColor && comp[i].type_count == 2) {
        return peInvalidArguments;
      }
      mesh->components[i] = comp[i];
    }
    mesh->header.file_alignment = alignment;
    mesh->header.num_components = count;
    return peSuccess;
  }
  catch(...) {
    return peException;
  }
}

ptf_err ptf_mesh_get_vertex_components(ptf_mesh_handle mesh,
    ptf_vertex_component_description const **o_comp, ptf_16 *o_count, ptf_16 *o_alignment)
{
  try {
    if (!mesh->validate()) return peInvalidArguments;
    if (!mesh->components.size()) {
      return peInvalidArguments;
    }
    *o_comp = &mesh->components[0];
    *o_count = (ptf_16)mesh->components.size();
    *o_alignment = mesh->header.file_alignment;
    return peSuccess;
  }
  catch(...) {
    return peException;
  }
}

ptf_err ptf_mesh_set_vertex_data(ptf_mesh_handle mesh, 
    void const * ptr, ptf_16 vertex_size, ptf_32 vertex_count)
{
  try {
    if (!mesh->validate()) return peInvalidArguments;
    if (vertex_size < 1 || vertex_size > MaxVertexStride) {
      return peInvalidArguments;
    }
    if (vertex_count < 1 || vertex_count > MaxVertexCount) {
      return peInvalidArguments;
    }
    mesh->clear_subset_information();
    mesh->vertices.resize(vertex_size * vertex_count);
    memcpy(&mesh->vertices[0], ptr, vertex_size * vertex_count);
    mesh->header.vertex_stride = vertex_size;
    mesh->header.vertex_count = vertex_count;
    return peSuccess;
  }
  catch(...) {
    return peException;
  }
}

ptf_err ptf_mesh_get_vertex_data(ptf_mesh_handle mesh,
    void const ** o_ptr, ptf_16 *o_vertex_size, ptf_32 *o_vertex_count)
{
  try {
    if (!mesh->validate()) return peInvalidArguments;
    if (!mesh->vertices.size()) {
      return peInvalidArguments;
    }
    *o_ptr = &mesh->vertices[0];
    *o_vertex_size = mesh->header.vertex_stride;
    *o_vertex_count = mesh->header.vertex_count;
    return peSuccess;
  }
  catch(...) {
    return peException;
  }
}

PTF_API ptf_err ptf_mesh_set_material_archive(ptf_mesh_handle mesh, 
    ptf_remote_ref archive)
{
  try {
    if (!mesh->validate()) return peInvalidArguments;
    mesh->header.material_archive = archive;
    return peSuccess;
  }
  catch(...) {
    return peException;
  }
}

PTF_API ptf_err ptf_mesh_get_material_archive(ptf_mesh_handle mesh, 
    ptf_remote_ref * o_archive)
{
  try {
    if (!mesh->validate()) return peInvalidArguments;
    *o_archive = mesh->header.material_archive;
    return peSuccess;
  }
  catch(...) {
    return peException;
  }
}

ptf_err ptf_mesh_add_mesh_subset(ptf_mesh_handle mesh,
    ptf_mesh_subset_description const &desc, void const *ptr)
{
  try {
    size_t iSize = 2;
    if (!mesh->validate()) return peInvalidArguments;
    if (desc.min_index > desc.max_index) {
      return peInvalidArguments;
    }
    if (desc.index_type != pvtUInt32 && desc.index_type != pvtUInt16) {
      return peInvalidArguments;
    }
    if (desc.index_type == pvtUInt32) {
      iSize = 4;
    }
    if (desc.num_indices == 0) {
      return peInvalidArguments;
    }
    if (desc.primitive_type == 0) {
      return peTooBigForMachine;
    }
    if ((desc.num_indices / desc.primitive_type) * desc.primitive_type != desc.num_indices) {
      return peInvalidArguments;
    }
    if (desc.num_indices > MaxIndexCount) {
      return peTooBigForMachine;
    }
    //  arbitrarily limit the number of subsets to 4,000
    if (mesh->subsets.size() > MaxSubsetCount) {
      return peInvalidArguments;
    }
    mesh->clear_subset_information();
    mesh->subsets.push_back(desc);
    mesh->indices.push_back(std::vector<char>());
    mesh->indices.back().resize(iSize * desc.num_indices);
    memcpy(&mesh->indices.back()[0], ptr, iSize * desc.num_indices);
    return peSuccess;
  }
  catch(...) {
    return peException;
  }
}

ptf_err ptf_mesh_count_mesh_subsets(ptf_mesh_handle mesh, 
    ptf_32 *o_count)
{
  try {
    if (!mesh->validate()) return peInvalidArguments;
    *o_count = (ptf_32)mesh->subsets.size();
    return peSuccess;
  }
  catch(...) {
    return peException;
  }
}

ptf_err ptf_mesh_get_mesh_subset(ptf_mesh_handle mesh, 
    ptf_32 index, ptf_mesh_subset_description *o_desc, void const **o_ptr)
{
  try {
    if (!mesh->validate()) return peInvalidArguments;
    if (index >= mesh->subsets.size()) {
      return peInvalidArguments;
    }
    *o_desc = mesh->subsets[index];
    *o_ptr = &mesh->indices[index][0];
    return peSuccess;
  }
  catch(...) {
    return peException;
  }
}

ptf_err ptf_mesh_remove_mesh_subset(ptf_mesh_handle mesh, 
    ptf_32 index)
{
  try {
    if (!mesh->validate()) return peInvalidArguments;
    if (index >= mesh->subsets.size()) {
      return peInvalidArguments;
    }
    if (index < mesh->subsets.size()-1) {
      mesh->subsets[index] = mesh->subsets.back();
      mesh->indices[index] = mesh->indices.back();
    }
    mesh->clear_subset_information();
    mesh->subsets.pop_back();
    mesh->indices.pop_back();
    return peSuccess;
  }
  catch(...) {
    return peException;
  }
}

ptf_err ptf_mesh_get_subset_information(ptf_mesh_handle mesh,
  ptf_32 index, ptf_subset_information const **o_info, ptf_16 *o_count)
{
  try {
    if (!mesh->validate()) return peInvalidArguments;
    if (index >= mesh->subsets.size()) {
      return peInvalidArguments;
    }
    return mesh->get_subset_information(index, o_info, o_count);
  }
  catch(...) {
    return peException;
  }
}

ptf_err ptf_mesh_set_subset_information(ptf_mesh_handle mesh,
  ptf_32 index, ptf_subset_information const *info, ptf_16 count)
{
  try {
    if (!mesh->validate()) return peInvalidArguments;
    if (index >= mesh->subsets.size()) {
      return peInvalidArguments;
    }
    return mesh->set_subset_information(index, info, count);
  }
  catch(...) {
    return peException;
  }
}


/////////////////////////////////////////////////////////////////

ptf_err ptf_mesh_get_edge_plane(ptf_mesh_handle mesh, ptf_8 plane, double * d)
{
  try {
    if ( ( plane < MeshMinPlane ) || ( plane >= MeshMaxPlane ) ) return peInvalidArguments;
    // can only access an edge list after its plane has been defined (either explicitly, or via read)
    if ( mesh->edge_lists[plane].plane_index == MeshUndefinedPlane ) return peInvalidArguments;

    return ptf_octreecoord_to_meters( mesh->edge_lists[plane].plane_d, d );
  }
  catch(...) {
    return peException;
  }
}

ptf_err ptf_mesh_set_edge_plane(ptf_mesh_handle mesh, ptf_8 plane, double d)
{
  try {
    if ( ( plane < MeshMinPlane ) || ( plane >= MeshMaxPlane ) ) return peInvalidArguments;
    
    ptf_32 ptf_d = 0;
    ptf_err err = ptf_meters_to_octreecoord( d, &ptf_d );
    if ( err != peSuccess ) {
      return err;
    }
    
    mesh->edge_lists[plane].plane_index = plane;
    mesh->edge_lists[plane].plane_d = ptf_d;
    return peSuccess;
  }
  catch(...) {
    return peException;
  }
}

ptf_err ptf_mesh_get_edge(ptf_mesh_handle mesh, ptf_8 plane, ptf_32 index, ptf_mesh_edge * edge)
{
  try {
    if ( ( plane < MeshMinPlane ) || ( plane >= MeshMaxPlane ) ) return peInvalidArguments;
    if ( mesh->edge_lists[plane].plane_index == MeshUndefinedPlane ) return peInvalidArguments;
    if ( index >= mesh->edge_lists[plane].edges.size() ) return peInvalidArguments;

    *edge = mesh->edge_lists[plane].edges[index].edge;
    return peSuccess;
  }
  catch(...) {
    return peException;
  }
}

ptf_err ptf_mesh_add_edge(ptf_mesh_handle mesh, ptf_8 plane, ptf_mesh_edge * edge)
{
  try {
    if ( ( plane < MeshMinPlane ) || ( plane >= MeshMaxPlane ) ) return peInvalidArguments;
    // must define the edge plane before adding edges
    if ( mesh->edge_lists[plane].plane_index == MeshUndefinedPlane ) return peInvalidArguments;

    ptf_edge_t new_edge( *edge );
    mesh->edge_lists[plane].edges.push_back( new_edge );

    return peSuccess;
  }
  catch(...) {
    return peException;
  }
}

ptf_err ptf_mesh_count_edges(ptf_mesh_handle mesh, ptf_8 plane, ptf_32 * o_count)
{
  try {
    if ( ( plane < MeshMinPlane ) || ( plane >= MeshMaxPlane ) ) return peInvalidArguments;
    if ( mesh->edge_lists[plane].plane_index == MeshUndefinedPlane ) return peTypeNotFound;

    *o_count = (ptf_32) mesh->edge_lists[plane].edges.size();
    return peSuccess;
  }
  catch(...) {
    return peException;
  }
}

ptf_err ptf_mesh_get_semantic(ptf_mesh_handle mesh,
    ptf_32 semantic, ptf_32 * o_semantic_info, ptf_record_ref * o_reference)
{
  try {
    if (!mesh->validate()) return peInvalidArguments;
    ptf_32 index, end = (ptf_32)mesh->semantics_.size();
    for (index = 0; index < end; ++index) {
      if (mesh->semantics_[index].semantic == semantic) {
        break;
      }
    }
    if (index >= mesh->semantics_.size()) {
      return peNameNotFound;
    }
    *o_semantic_info = mesh->semantics_[index].semantic_info;
    *o_reference = mesh->semantics_[index].reference;
    return peSuccess;
  }
  catch (...) {
    return peException;
  }
}

ptf_err ptf_mesh_set_semantic(ptf_mesh_handle mesh,
    ptf_32 semantic, ptf_32 semantic_info, ptf_record_ref const & reference)
{
  try {
    if (!mesh->validate()) return peInvalidArguments;
    ptf_32 index, end = (ptf_32)mesh->semantics_.size();
    for (index = 0; index < end; ++index) {
      if (mesh->semantics_[index].semantic == semantic) {
        mesh->semantics_[index].semantic_info = semantic_info;
        mesh->semantics_[index].reference = reference;
      }
    }
    if (index >= mesh->semantics_.size()) {
      ptf_semantic_ref new_semantic;
      new_semantic.semantic = semantic;
      new_semantic.semantic_info = semantic_info;
      new_semantic.reference = reference;
      mesh->semantics_.push_back( new_semantic );
    }
    return peSuccess;
  }
  catch (...) {
    return peException;
  }
}

ptf_err ptf_mesh_remove_semantic(ptf_mesh_handle mesh, ptf_32 semantic)
{
  try {
    if (!mesh->validate()) return peInvalidArguments;
    ptf_32 index, end = (ptf_32)mesh->semantics_.size();
    for (index = 0; index < end; ++index) {
      if (mesh->semantics_[index].semantic == semantic) {
        mesh->semantics_.erase(mesh->semantics_.begin() + index);
        return peSuccess;
      }
    }
    return peNameNotFound;
  }
  catch (...) {
    return peException;
  }
}

ptf_err ptf_mesh_count_semantics(ptf_mesh_handle mesh, ptf_32 *o_count)
{
  try {
    if (!mesh->validate()) return peInvalidArguments;
    *o_count = (ptf_32)mesh->semantics_.size();
    return peSuccess;
  }
  catch (...) {
    return peException;
  }
}

ptf_err ptf_mesh_get_semantics(ptf_mesh_handle mesh, ptf_32 *o_semantics, ptf_32 max_count)
{
  try {
    if (!mesh->validate()) return peInvalidArguments;
    for (ptf_32 i = 0, n = (ptf_32)mesh->semantics_.size(); i != n; ++i) {
      if (i == max_count) { //  reached max size of output array
        return peSuccess;
      }
      o_semantics[i] = mesh->semantics_[i].semantic;
    }
    return peSuccess;
  }
  catch (...) {
    return peException;
  }
}

namespace {
  template<typename Oper> struct AccumFunc {
    void operator()(char const *src, char *dst, size_t align, ptf_vertex_component_description const *descs, ptf_16 nDescs)
      {
        for (int i = 0; i < nDescs; ++i) {
          size_t s = 0;
          for (int j = 0; j < descs[i].type_count; ++j) {
            size_t q = 0;
            switch (descs[i].type & 0xf) {
              case pvtFloat32:
                q = 4;
                o((float const *)src, (float *)dst);
                break;
              case pvtInt32:
                q = 4;
                o((int const *)src, (int *)dst);
                break;
              case pvtUInt32:
                q = 4;
                o((unsigned int const *)src, (unsigned int *)dst);
                break;
              case pvtInt16:
                q = 2;
                o((short const *)src, (short *)dst);
                break;
              case pvtUInt16:
                q = 2;
                o((unsigned short const *)src, (unsigned short *)dst);
                break;
              case pvtInt8:
                q = 1;
                o((char const *)src, (char *)dst);
                break;
              case pvtUInt8:
                q = 1;
                o((unsigned char const *)src, (unsigned char *)dst);
                break;
              default:
                throw std::runtime_error("Unknown vertex format in Mesh -- should not happen!");
            }
            src += q;
            dst += q;
            s += q;
          }
          s = ((s + (align-1)) & -align) - s;
          src += s;
          dst += s;
        }
      }
    Oper o;
  };

  struct MinData {
    template<typename T> void operator()(T const *a, T *b) {
      *b = std::min(*a, *b);
    }
  };

  struct MaxData {
    template<typename T> void operator()(T const *a, T *b) {
      *b = std::max(*a, *b);
    }
  };

  AccumFunc<MinData> minFunc;
  AccumFunc<MaxData> maxFunc;
};

template<typename T> void ptf_mesh_t::accumulate(char const *src, char *dst, T &t) {
  t(src, dst, header.file_alignment, &components[0], (ptf_16)components.size());
}

void ptf_mesh_t::calc_subset_information()
{
  //  for each vertex in the subset, for each component, accumulate
  for (size_t ssix = 0; ssix < subsets.size(); ++ssix) {
    ptf_mesh_subset_description pmsd = subsets[ssix];
    std::vector<char> vertMin, vertMax;
    vertMin.resize(header.vertex_stride);
    vertMax.resize(header.vertex_stride);
    if (pmsd.index_type == pvtUInt16) {
      unsigned short const *idcs = (unsigned short const *)&indices[ssix][0];
      memcpy(&vertMin[0], &this->vertices[idcs[0]*(size_t)header.vertex_stride], header.vertex_stride);
      memcpy(&vertMax[0], &this->vertices[idcs[0]*(size_t)header.vertex_stride], header.vertex_stride);
      for (ptf_32 i = 1; i < pmsd.num_indices; ++i) {
        unsigned short ix = idcs[i];
        accumulate(&this->vertices[ix*(size_t)header.vertex_stride], &vertMin[0], minFunc);
        accumulate(&this->vertices[ix*(size_t)header.vertex_stride], &vertMax[0], maxFunc);
      }
    }
    else {
      unsigned int const *idcs = (unsigned int const *)&indices[ssix][0];
      memcpy(&vertMin[0], &this->vertices[idcs[0]*(size_t)header.vertex_stride], header.vertex_stride);
      memcpy(&vertMax[0], &this->vertices[idcs[0]*(size_t)header.vertex_stride], header.vertex_stride);
      for (ptf_32 i = 1; i < pmsd.num_indices; ++i) {
        unsigned int ix = idcs[i];
        accumulate(&this->vertices[ix*(size_t)header.vertex_stride], &vertMin[0], minFunc);
        accumulate(&this->vertices[ix*(size_t)header.vertex_stride], &vertMax[0], maxFunc);
      }
    }
    //  remember calculated information
    subset_info_index.push_back((ptf_32)vertex_info.size());
    size_t size = vertex_info_data.size();
    vertex_info_data.resize(size + vertMin.size() + vertMax.size());
    memcpy(&vertex_info_data[size], &vertMin[0], vertMin.size());
    memcpy(&vertex_info_data[size + vertMin.size()], &vertMax[0], vertMax.size());
    ptf_subset_information_record pvi;
    memset(&pvi, 0, sizeof(pvi));
    pvi.type = pviSubsetInfoMin;
    pvi.size = (ptf_32)vertMin.size();
    pvi.offset = (ptf_32)size;
    vertex_info.push_back(pvi);
    pvi.type = pviSubsetInfoMax;
    pvi.size = (ptf_32)vertMax.size();
    pvi.offset = (ptf_32)(size + vertMin.size());
    vertex_info.push_back(pvi);
  }
  //  I'll pay attention to the returned information when and if someone asks for it.
  returned_vertex_info.clear();
}

void ptf_mesh_t::clear_subset_information()
{
  subset_info_index.clear();
  vertex_info_data.clear();
  vertex_info.clear();
  returned_vertex_info.clear();
}

void ptf_mesh_t::calc_returned_subset_information() {
  returned_vertex_info.resize(vertex_info.size());
  for (size_t i = 0; i < vertex_info.size(); ++i) {
    returned_vertex_info[i].type = vertex_info[i].type;
    returned_vertex_info[i].size = vertex_info[i].size;
    returned_vertex_info[i].data = (vertex_info_data.size() ? 
        &vertex_info_data[0] + vertex_info[i].offset : (char *)0);
    assert(returned_vertex_info[i].data != 0 || returned_vertex_info[i].size == 0);
  }
}

void ptf_mesh_t::compact_subset_information() {
  //  The "data" part of the subset may be non-compact. Create a new vector to make sure it isn't.
  size_t newSize = 0;
  for (size_t i = 0; i < vertex_info.size(); ++i) {
    newSize += (vertex_info[i].size + 3) & -4;
  }
  std::vector<char> nu;
  nu.resize(newSize);
  size_t offset = 0;
  for (size_t i = 0; i < vertex_info.size(); ++i) {
    memcpy(&nu[offset], &vertex_info_data[vertex_info[i].offset], vertex_info[i].size);
    vertex_info[i].offset = (ptf_32)offset;
    offset += (vertex_info[i].size + 3) & -4;
  }
  nu.swap(vertex_info_data);
}

ptf_err ptf_mesh_t::get_subset_information(ptf_32 index, ptf_subset_information const **o_info, ptf_16 *o_count)
{
  if (vertex_info.empty()) {
    calc_subset_information();
  }
  if (returned_vertex_info.size() == 0) {
    calc_returned_subset_information();
  }
  size_t offset = subset_info_index[index];
  size_t top = (index == subset_info_index.size()-1) ? vertex_info.size() : subset_info_index[index+1];
  if (top == offset) {
    *o_info = 0;
    *o_count = 0;
  }
  else {
    *o_info = &returned_vertex_info[offset];
    *o_count = (ptf_16)(top - offset);
  }
  return peSuccess;
}


ptf_err ptf_mesh_t::set_subset_information(ptf_32 index, ptf_subset_information const *info, ptf_16 count)
{
  //  Make sure I regenerate return data when/if it's asked for again.
  returned_vertex_info.clear();
  //  This is an intricate operation, because of the 
  //  implicit array storage method.
  if (subset_info_index.empty()) {
    //  Make sure there's some structure to edit, if none has been generated yet.
    //  All zeros means that each subset has zero records, which will work OK in the 
    //  algorithm below (records and data is empty).
    subset_info_index.insert(subset_info_index.begin(), subsets.size(), 0);
  }
  assert(subset_info_index.size() > index);
  //  Update indices
  size_t oldRecordBase = subset_info_index[index];
  size_t oldRecordEnd = (index == subset_info_index.size()-1) ? 
      vertex_info.size() : subset_info_index[index+1];
  for (size_t i = index+1; i < subset_info_index.size(); ++i) {
    if (subset_info_index[i] >= oldRecordEnd) {
      subset_info_index[i] += (ptf_32)((size_t)count - (oldRecordEnd - oldRecordBase));
    }
  }
  //  Remove existing records
  vertex_info.erase(vertex_info.begin() + oldRecordBase, vertex_info.begin() + oldRecordEnd);
  //  Add new records
  ptf_subset_information_record psir;
  memset(&psir, 0, sizeof(psir));
  vertex_info.insert(vertex_info.begin() + oldRecordBase, count, psir);
  //  Push the data at the end of the data array.
  size_t newOffset = vertex_info_data.size();
  size_t newSize = 0;
  for (ptf_16 i = 0; i < count; ++i) {
    vertex_info[oldRecordBase + i].type = info[i].type;
    vertex_info[oldRecordBase + i].size = info[i].size;
    //  note that the data is inserted in the loop below
    vertex_info[oldRecordBase + i].offset = (ptf_32)(newOffset + newSize);
    newSize += (info[i].size + 3) & -4;
  }
  //  I add the new data at the end, even if there is old data in there somewhere
  if (newSize > 0) {
    //  Now that I know how much to resize the vector, resize it
    vertex_info_data.insert(vertex_info_data.end(), newSize, 0);
    newSize = 0;
    for (ptf_16 i = 0; i < count; ++i) {
      //  note that I assign .offset in the loop above!
      memcpy(&vertex_info_data[newOffset + newSize], info[i].data, info[i].size);
      newSize += (info[i].size + 3) & -4;
    }
  }
  return peSuccess;
}

