
#include "ptf_util.h"
#include "ptf_types.h"

using namespace impl;

template<typename T> inline ptf_err unpack(void const **io_input, float *o_output, ptf_8 count, ptf_16 alignment)
{
  T const *t = (T const *)*io_input;
  for (ptf_8 i = 0; i < count; ++i) {
    o_output[i] = t[i];
  }
  size_t s = (sizeof(T) * count + (alignment-1)) & -alignment;
  *io_input = (char const *)*io_input + s;
  return peSuccess;
}

template<typename T, int divBy> inline ptf_err unpack_scale(void const **io_input, float *o_output, ptf_8 count, ptf_16 alignment)
{
  T const *t = (T const *)*io_input;
  for (ptf_8 i = 0; i < count; ++i) {
    //  Use an actual division to make sure that -32767 / -32767 comes out as 1.
    //  Multiplying with the reciprocal doesn't guarantee that, but division does.
    o_output[i] = float(t[i]) / float(divBy);
  }
  size_t s = (sizeof(T) * count + (alignment-1)) & -alignment;
  *io_input = (char const *)*io_input + s;
  return peSuccess;
}

template<typename T, int divBy> inline ptf_err unpack_contract(void const **io_input, float *o_output, ptf_8 count, ptf_16 alignment)
{
  //  make sure we can't compile with a bad divBy
  typedef int compile_time_assert_power_of_two[!(divBy & (divBy-1))];
  //  I know that divBy is a power of two. Thus the reciprocal is exact in that special case.
  static float const recip = 1.0f / divBy;
  T const *t = (T const *)*io_input;
  for (ptf_8 i = 0; i < count; ++i) {
    o_output[i] = float(t[i]) * recip;
  }
  size_t s = (sizeof(T) * count + (alignment-1)) & -alignment;
  *io_input = (char const *)*io_input + s;
  return peSuccess;
}

template<typename T, int mulBy> inline ptf_err unpack_expand(void const **io_input, float *o_output, ptf_8 count, ptf_16 alignment)
{
  T const *t = (T const *)*io_input;
  for (ptf_8 i = 0; i < count; ++i) {
    o_output[i] = float(t[i]) * mulBy;
  }
  size_t s = (sizeof(T) * count + (alignment-1)) & -alignment;
  *io_input = (char const *)*io_input + s;
  return peSuccess;
}

ptf_err ptf_type_unpack(ptf_8 type, ptf_8 count, ptf_16 alignment, void const **io_input, float *o_output)
{
  if (!alignment || (alignment & (alignment-1))) {
    return peInvalidArguments;
  }

  switch (type) {

    case pvtFloat32:
      return unpack<float>(io_input, o_output, count, alignment);
    case pvtUInt32:
      return unpack<unsigned int>(io_input, o_output, count, alignment);
    case pvtInt32:
      return unpack<int>(io_input, o_output, count, alignment);

    case pvtUInt16:
      return unpack_scale<unsigned short, 65535>(io_input, o_output, count, alignment);
    case pvtInt16:
      return unpack_scale<short, 32767>(io_input, o_output, count, alignment);
    case pvtUInt8:
      return unpack_scale<unsigned char, 255>(io_input, o_output, count, alignment);
    case pvtInt8:
      return unpack_scale<char, 127>(io_input, o_output, count, alignment);

    case pvtUInt16_4:
      return unpack_contract<unsigned short, 4096>(io_input, o_output, count, alignment);
    case pvtInt16_4:
      return unpack_contract<short, 4096>(io_input, o_output, count, alignment);
    case pvtUInt8_4:
      return unpack_contract<unsigned char, 16>(io_input, o_output, count, alignment);
    case pvtInt8_4:
      return unpack_contract<char, 16>(io_input, o_output, count, alignment);

    case pvtUInt16_8:
      return unpack_contract<unsigned short, 256>(io_input, o_output, count, alignment);
    case pvtInt16_8:
      return unpack_contract<short, 256>(io_input, o_output, count, alignment);
    case pvtUInt8_8:
      return unpack<unsigned char>(io_input, o_output, count, alignment);
    case pvtInt8_8:
      return unpack<char>(io_input, o_output, count, alignment);

    case pvtUInt16_12:
      return unpack_contract<unsigned short, 16>(io_input, o_output, count, alignment);
    case pvtInt16_12:
      return unpack_contract<short, 16>(io_input, o_output, count, alignment);
    case pvtUInt8_12:
      return unpack_expand<unsigned char, 16>(io_input, o_output, count, alignment);
    case pvtInt8_12:
      return unpack_expand<char, 16>(io_input, o_output, count, alignment);

    case pvtUInt16_16:
      return unpack<unsigned short>(io_input, o_output, count, alignment);
    case pvtInt16_16:
      return unpack<short>(io_input, o_output, count, alignment);
    case pvtUInt8_16:
      return unpack_expand<unsigned char, 256>(io_input, o_output, count, alignment);
    case pvtInt8_16:
      return unpack_expand<char, 256>(io_input, o_output, count, alignment);

    default:
      return peInvalidArguments;
  }
}

/*
  private:
    //! \internal block
    zero_t<int *> block_;
    //! \internal physical size
    zero_t<size_t> phys_;
    //! \internal written size
    zero_t<size_t> size_;
    //! \internal file pointer pos
    zero_t<size_t> pos_;
*/

block_t::block_t(size_t prealloc)
{
  block_ = new int[((prealloc+3)>>2)+1];
  block_[0] = 1;
  phys_ = (prealloc+3) & -4;
}

block_t::~block_t()
{
  release();
}

block_t::block_t(block_t const & o)
{
  block_ = o.block_;
  phys_ = o.phys_;
  size_ = o.size_;
  pos_ = o.pos_;
  hold();
}

block_t & block_t::operator=(block_t const & o)
{
  if (block_ == o.block_) return *this;
  release();
  block_ = o.block_;
  phys_ = o.phys_;
  size_ = o.size_;
  pos_ = o.pos_;
  hold();
  return *this;
}

bool block_t::get_bytes(void * out, size_t count)
{
  if (count > size_-pos_) {
    return false;
  }
  memcpy(out, ((char *)(block_+1)) + pos_, count);
  pos_ += count;
  return true;
}

bool block_t::put_bytes(void const * out, size_t count)
{
  if (!reserve(pos_ + count)) return false;
  memcpy(((char *)(block_+1)) + pos_, out, count);
  pos_ += count;
  if (size_ < pos_) size_ = pos_;
  return true;
}

void block_t::release()
{
  if (block_) {
    if (block_[0]-- == 1) {
      delete[] block_.t_;
    }
  }
  block_ = 0;
}

void block_t::hold()
{
  if (block_) {
    block_[0]++;
  }
}

void block_t::split()
{
  //  if I'm not alone, make sure that I let go of the data
  if (block_[0] > 1) {
    int * p = block_;
    release();
    block_ = new int[(phys_>>2)+1];
    memcpy(block_+1, p+1, size_);
    block_[0] = 1;
  }
}

bool block_t::reserve(size_t s)
{
  if (phys_ < s) {
    if (s < phys_ + 256) {
      s = phys_ + 256;
    }
    if (s < (size_t)(phys_ * 1.3)) {
      s = (size_t)(phys_ * 1.3);
    }
    s = (s + 3) & -4;
    int * p = new int[(s>>2)+1];
    memcpy(p+1, block_+1, size_);
    p[0] = 1;
    release();
    block_ = p;
    phys_ = s;
  }
  else {
    split();
  }
  return true;
}

bool block_t::str_in(ptf_char const * str)
{
  size_t s = strlen((char const *)str);
  return put_bytes(str, s+1);
}

bool block_t::str_out(ptf_char * str, size_t size)
{
  *str = 0;
  size_t p = pos_;
  char const * src = (char *)(block_+1);
  while (p < size_) {
    if (src[p++] == 0) {
      goto done;
    }
  }
  //  no termination
  return false;
done:
  if (p-pos_ > size) {
    return false;
  }
  memcpy(str, src+pos_, p-pos_);
  pos_ = p;
  return true;
}

bool block_t::seek(size_t pos, int whence)
{
  switch (whence) {
    case 0:
      break;
    case 1:
      pos += pos_;
      break;
    case 2:
      pos += size_;
      break;
  }
  if (pos > size_) {
    return false;
  }
  pos_ = pos;
  return true;
}

void *block_t::extend(size_t size)
{
  if (!reserve(size + pos_)) {
    return 0;
  }
  pos_ += size;
  if (size_ < pos_) {
    size_ = pos_;
  }
  return ((char *)(block_+1)) + pos_ - size;
}
